#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "fitsio.h"
#include "fitsdebug.h"

	 
int fd_create_file(fd_file *file,    /* structure holding fits file */
		   char *filename,   /* filename of file to create  */ 
		   int tabletype,    /* ASCII_TBL or BINARY_TBL     */
		   char *tablename,  /* Name of the table           */
		   int *status,      /* Status flag                 */
		   int flush,        /* Flush flag                  */
		   int supersede,    /* Supersede flag              */
		   int ncol,         /* # of columns in the table   */
		   ...) {            /* Rest of argument 
					- column format
					- column name
					- column unit               */

  /* open a new fitsfile and assign it the name filename.
     

  */

  va_list ap;               /* points to each unnamed arg in turn */

  char tmp_filename[MAXSTRLEN];
  fitsfile *fptr;
  /* required keywords */
  int simple = TRUE;
  short bitpix = 8;
  int extend = TRUE;
  /* keywords required for the Table HDU */
  int i;

  /* First things first: check if all input is within the allowed
     range */

  /* check if supersede flag has valid value, if not return error */
  if (!fd_set_supersede(file,supersede)) {
    *status = FD_STAT_SUPERSEDE;
    return FD_STAT_SUPERSEDE;
  }
  /* check if flush flag has valid value, if not return error */
  if (!fd_set_flush(file,flush)) {
    *status = FD_STAT_FLUSH;
    return FD_STAT_FLUSH;
  }

  
  /* If supersede is turned on replace any existing files with a  */
  /* brand new file. Cfitsio solves this by puting a "!" in front */
  /* of the file name.                                            */
  if (supersede == FD_SUPERSEDE_ON) {
    /* add an exclamation mark in front of the filename so that cfitsio
       knows to delete any existing files */
    strcpy(tmp_filename,"!");
    strcat(tmp_filename,filename);
  } else {
    strcpy(tmp_filename,filename);
  }

  /* Create the fits file                                              */
  /* 1: if supersede_off and the filename already exists: create a new */
  /*    file and tell the used you did so. Open this new file as the   */
  /*    debug file. The new file will have a number appended to the    */
  /*    original file name preceded by a ",". Thus if the file is      */
  /*    called 'fits.fits' the new file will be 'fits.fits,1'.         */
  if ((supersede == FD_SUPERSEDE_OFF) &&
      fd_file_exist(tmp_filename)) {
    fd_append_extension(tmp_filename);
    filename = tmp_filename;
  }

  /* 2: create the new fits file */
  if (fits_create_file(&fptr, tmp_filename, status)) {
    return *status; 
  } else {
    fd_set_fptr(file,fptr);
    fd_set_filename(file,filename);
    fd_set_hdu(file,1);
    fd_set_file_status(file,FD_FILE_OPEN);
  }

  /* 3: supply the few required keywords without which the fits file  */
  /*    will not work (SIMPLE = T, BITPIX = 8, NAXIS                  */
  if (fits_write_key(fptr, TLOGICAL, "SIMPLE", &simple,
		     "file does conform to FITS standard", status))
    return *status;
  if (fits_write_key(fptr, TSHORT, "BITPIX", &bitpix, 
		     "number of bits per data pixel", status))
    return *status;
  if (fits_write_key_null(fptr, "NAXIS", "number of data axes", status))
    return *status;

  /* 4: Tell fits file that extensions are ok */
  if (fits_write_key(fptr, TLOGICAL, "EXTEND", &extend,
		     "FITS dataset may contain extensions", status))
    return *status;

  /* 5: append creation date and a comment line to the fits file */
  if (fits_write_date(fptr, status)) return *status;
  if (fd_comment_no_flush(file, FD_DEFAULT_COMMENT, status)) 
    return *status;

  /* 6: Start setting up all necessary info to create a binary or */
  /*    ascii table to be included in the filts file              */
  if (!fd_set_tabletype(file,tabletype)) {
    *status = FD_STAT_TABLE_TYPE;
    return FD_STAT_TABLE_TYPE;
  } 
  fd_set_tablename(file,tablename);


  if (tabletype == BINARY_TBL) {
    
    /* WE ARE DEALING WITH A BINARY TABLE */

    /* reteive information on the table column names, formats, and units */
    /* we use the stdarg variable argument routines                      */
    /* 1: first set number of columns in the table                       */
    
    fd_set_ncol(file,ncol);
    
    /* 2: Now assign memory and fill the arrays                          */
    file->typecode = mem_malloc_safe(ncol*sizeof(typecodeS));
    file->tform = mem_malloc_safe(ncol*sizeof(char *));
    file->ttype = mem_malloc_safe(ncol*sizeof(char *));
    file->tunit = mem_malloc_safe(ncol*sizeof(char *));
    va_start(ap, ncol);
    for (i = 0; i < fd_get_ncol(file); i++) {
      
      fd_set_tformi(file,i,va_arg(ap, char *));
      fd_set_ttypei(file,i,va_arg(ap, char *));
      fd_set_tuniti(file,i,va_arg(ap, char *));
      
    }
    va_end(ap);
    

    fd_set_nrow(file,0);      /* file is empty when we create it */
    
    
    if (fits_create_tbl(file->fptr, file->tabletype, 0, file->ncol,
			file->ttype, file->tform, file->tunit, file->tablename,
			status)) {
      return *status;
    } else {
      fd_set_hdu(file,fd_get_hdu(file)+1);
    }
    

  } else if (tabletype == ASCII_TBL) {



    /* WE ARE DEALING WITH AN ASCII TABLE NOW */

    /* reteive information on the table column names, formats, and units */
    /* we use the stdarg variable argument routines                      */
    /* 1: first set number of columns in the table                       */
    
    fd_set_ncol(file,ncol);
    
    /* 2: Now assign memory and fill the arrays                          */
    file->typecode = mem_malloc_safe(ncol*sizeof(typecodeS));
    file->tform = mem_malloc_safe(ncol*sizeof(char *));
    file->ttype = mem_malloc_safe(ncol*sizeof(char *));
    file->tunit = mem_malloc_safe(ncol*sizeof(char *));
    va_start(ap, ncol);
    for (i = 0; i < fd_get_ncol(file); i++) {
      
      fd_set_tformi(file,i,va_arg(ap, char *));
      fd_set_ttypei(file,i,va_arg(ap, char *));
      fd_set_tuniti(file,i,va_arg(ap, char *));
      
    }
    va_end(ap);
    

    fd_set_nrow(file,0);      /* file is empty when we create it */
    
    
    if (fits_create_tbl(file->fptr, file->tabletype, 0, file->ncol,
			file->ttype, file->tform, file->tunit, file->tablename,
			status)) {
      return *status;
    } else {
      fd_set_hdu(file,fd_get_hdu(file)+1);
    }
    
  }


  /* Finally: before returning check the flush flag and handle
     the request accordingly                                   */
  
  if (fd_flush_close(file))
    msg_message(3,"-->Error in fd_flush_close"); 
    
  return FD_STAT_ALLOK;

}


int fd_flush_close(fd_file *file) {

  /* Depending on the flush flag set by the user close the fitsfile */
  /* and later open it when more writing is needed, or flush the    */
  /* buffer and keep the file open.                                 */

  int status = 0;
  int flush = file->flush;    /* fd_get_flush(file); */

  if (file->fptr != NULL) {
    if (flush == FD_FLUSH_ON) {
      if (fits_flush_file(file->fptr, &status)) return status;
    } else if (flush == FD_FLUSH_OFF) {
      fd_set_file_status(file,FD_FILE_CLOSED);
      if (fits_close_file(file->fptr, &status)) return status;
    } else {
      msg_message(3,"Flush flag has incorrect value (%d)\n",flush);
      return FD_STAT_FLUSH;
    }
    return FD_STAT_ALLOK;
  } else {
    msg_message(3,"file->fptr == NULL\n");
    return FD_STAT_FILE_NOT_OPEN;
  }
  
}

int fd_flush_open(fd_file *file) {

  /* Depending on the flush flag set by the user open the fitsfile  */
  /* for more writing                                               */

  int status = 0;
  int hdutype = file->tabletype;   /* fd_get_tabletype(file); */
  int flush = file->flush;         /* fd_get_flush(file); */

  if (flush == FD_FLUSH_OFF) {
      if (fits_open_file(&file->fptr, 
			 file->filename,
			 READWRITE,
			 &status)) return status;
      fd_set_file_status(file,FD_FILE_OPEN);
      /* move to the correct HDU */
      fits_movabs_hdu(file->fptr, fd_get_hdu(file), &hdutype, &status);
  } else if (file->fptr == NULL || 
	     fd_get_file_status(file) != FD_FILE_OPEN) {
    return FD_STAT_FILE_NOT_OPEN;
  }

  return FD_STAT_ALLOK;
  
}


int fd_close_file(fd_file *file, int *status) {

  /* Close the fits file associated with the debuging */
  /* structure                                        */

  if (fd_get_file_status(file) == FD_FILE_OPEN) {
    if (fits_close_file(file->fptr, status)) {
      return *status;
    } else {
      fd_set_file_status(file, FD_FILE_CLOSED);
      *status = FD_STAT_ALLOK;
      return *status;
    }
  } 

  *status = FD_STAT_ALLOK;
  return *status;

}

int fd_exit_file(fd_file *file, int *status) {

  /* Free allocated memory and check if file is closed */

  int i;

  /* a precaution */
  if (file->fptr != NULL) 
    fd_close_file(file, status); 

  /* free memory */

  if (fd_get_ncol(file) != -1) {
    for (i = 0; i < fd_get_ncol(file); i++) {
      if (file->typecode[i].raw != NULL) free(file->typecode[i].raw);
      if (file->typecode[i].type_code != NULL) free(file->typecode[i].type_code);
      if (file->tform[i] != NULL) free(file->tform[i]);
      if (file->ttype[i] != NULL)  free(file->ttype[i]); 
      if (file->tunit[i] != NULL)  free(file->tunit[i]); 
    }
    if (file->typecode != NULL) free(file->typecode);
    if (file->tform != NULL) free(file->tform);
    if (file->ttype != NULL) free(file->ttype);
    if (file->tunit != NULL) free(file->tunit);
  }

  return *status;

}

void fd_print_fd_file(fd_file *file) {

  /* print a description of a FD_FILE structure to the */
  /* screen for easy debuging                          */

  printf("\n\nDescription of FD_FILE:\n\n");
  printf(" File name      : %s\n", fd_get_filename(file));
  printf(" Table name     : %s\n", fd_get_tablename(file));
  printf(" TFORM          : "); fd_print_tform(file);
  printf(" TTYPE          : "); fd_print_ttype(file);
  printf(" TUNIT          : "); fd_print_tunit(file);
  printf(" HDU #          : %d\n", fd_get_hdu(file));
  printf(" Extension type : %d\n", fd_get_tabletype(file));
  printf(" # col          : %d\n", fd_get_ncol(file));
  printf(" # row          : %ld\n", fd_get_nrow(file));
  printf(" Flush flag     : %d\n", fd_get_flush(file));
  printf(" Supersede flag : %d\n", fd_get_supersede(file));
  printf(" File status    : %d\n", fd_get_file_status(file));
  printf("\n");

  return;

}

void fd_print_tform(fd_file *file) {
  
  int ncol = fd_get_ncol(file);
  int i;

  printf("(");
  for (i = 0; i < ncol; i++) {
    printf("%s",file->tform[i]); 
    if (i != ncol-1) printf(",");
  }
  printf(")\n");

}

void fd_print_ttype(fd_file *file) {
  
  int ncol = fd_get_ncol(file);
  int i;

  printf("(");
  for (i = 0; i < ncol; i++) {
    printf("%s",file->ttype[i]); 
    if (i != ncol-1) printf(",");
  }
  printf(")\n");

}

void fd_print_tunit(fd_file *file) {
  
  int ncol = fd_get_ncol(file);
  int i;

  printf("(");
  for (i = 0; i < ncol; i++) {
    printf("%s",file->tunit[i]); 
    if (i != ncol-1) printf(",");
  }
  printf(")\n");

}
