#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>

#define ERG_KEV 1.60219e-9

void linear_approx(double *x, double *y, double *e, int *m, double *u, 
		   double *v, double *w, int *k, int *ip);
/* x: x axis, y: y axis, e: error, m: Nbin
   u: x axis, v: y axis, w: unused,k: Nout-1, ip: i1*i2*i3.
   i1: -1, e is an array.  +1, e is a value
   i2: 2 if result must be within error.  1 if not.
   i3: 1 if not continuous.  3 if continuous. 
   ip = +1*2*3 = 6 for this application */

/* Checks results after completion to ensure no negative or too small */
/* values.  Values below minimum positive floating point value are set */
/* to zero automatically */

void compress_energy_spectrum(int Nbin, double *bin_edge, double *emission, 
			      double error, char *type,
			      int *Nval, double **Eval, double **Fluxval) {

  /* Assumes emission is in units of ergs cm^3 / s */
			
  double *x, *y, *e;
  double *u, *v, *w;
  double dE;
  int m, k, ip;
  int iBin;

  ip = -6;
  m  = Nbin;
  x  = (double *) malloc(m*sizeof(double));
  y  = (double *) malloc(m*sizeof(double));
  e  = (double *) malloc(m*sizeof(double));
  u  = (double *) malloc(m*sizeof(double));
  v  = (double *) malloc(m*sizeof(double));
  w  = NULL;
  k  = 0;

  for (iBin=0;iBin<m;iBin++) {
    x[iBin] = (bin_edge[iBin] + bin_edge[iBin+1])/2.0;
    dE      = (bin_edge[iBin+1] - bin_edge[iBin]);
    y[iBin] = (emission[iBin]/(x[iBin]*ERG_KEV)); /* photons cm^3 / s */
    y[iBin] = y[iBin]/dE;  /* photons cm^3 / s / keV */
    e[iBin] = fabs(error*y[iBin]);
  }

  /* Make the call */
  linear_approx(x, y, e, &m, u, v, w, &k, &ip);

  if (k == 0) {
    printf("compress_continuum: Failed on %s\n",type);
    abort();
    /* Use backup method--no compression */
    k = m-1;
    for (iBin=0;iBin<m;iBin++) {
      u[iBin] = x[iBin];
      v[iBin] = y[iBin];
    }
  }

  *Nval   = k+1;
  *Eval    = (double *) malloc((*Nval)*sizeof(double));
  *Fluxval = (double *) malloc((*Nval)*sizeof(double));
  for (iBin=0; iBin<*Nval; iBin++) {
    (*Eval)[iBin] = u[iBin];
    if (v[iBin] > FLT_MIN) {
      (*Fluxval)[iBin] = v[iBin];
    } else {
      (*Fluxval)[iBin] = 0.0;
    }
  }
  free(x);
  free(y);
  free(e);
  free(u);
  free(v);
  
  /* 
  if (debug->do_compress) {
    fprintf(debug->compress,"compress_continuum: %s: In : %d, Out : %d\n", 
	    type, m, *Nval);
  }
  */
}

void compress_photon_spectrum(double *Energy, double *flux, int *Nbin, 
			      double error) {
			
  /* Takes a spectrum flux[i] in units of photon cm^3 / s / bin, with bin edges
     E[i] and E[i+1] and Nbin bins, and returns the compressed (via linear
     interpolates) result with tolerance given by error.  The compression
     is done IN PLACE */

  double *x, *y, *e;
  double *u, *v, *w;
  double dE;
  int m, k, ip;
  int iBin;

  ip = -6;
  m  = *Nbin;
  x  = (double *) malloc(m*sizeof(double));
  y  = (double *) malloc(m*sizeof(double));
  e  = (double *) malloc(m*sizeof(double));
  u  = (double *) malloc(m*sizeof(double));
  v  = (double *) malloc(m*sizeof(double));
  w  = NULL;
  k  = 0;

  for (iBin=0;iBin<m;iBin++) {
    x[iBin] = (Energy[iBin] + Energy[iBin+1])/2.0;
    dE      = (Energy[iBin+1] - Energy[iBin]);
    y[iBin] = flux[iBin];  /* photons cm^3 / s */
    y[iBin] = y[iBin]/dE;  /* photons cm^3 / s / keV */
    e[iBin] = fabs(error*y[iBin]);
  }

  /* Make the call */
  linear_approx(x, y, e, &m, u, v, w, &k, &ip);

  if (k == 0) {
    printf("compress_flux: Failed.\n");
    /* Use backup method--no compression */
    *Nbin = m;
    for (iBin=0;iBin<m;iBin++) {
      Energy[iBin] = x[iBin];
      flux[iBin]   = y[iBin];
    }
  } else {
    *Nbin = k+1;
    for (iBin=0; iBin<*Nbin; iBin++) {
      Energy[iBin] = u[iBin];
      if (v[iBin] > FLT_MIN) {
	flux[iBin] = v[iBin];
      } else {
	flux[iBin] = 0.0;
      }
    }
  }

  free(x);
  free(y);
  free(e);
  free(u);
  free(v);

  /* printf("compress_photon_spectrum: In : %d, Out : %d\n", m, *Nbin); */
}
