#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "apec.h"
#include "apec_funcdefs.h"
#include "messages.h"

void apec_nei(struct PARAMETERS *params, struct BIN_SPEC spectrum,
	      ElementP *Atomic_params, struct ION_BAL_TABLE *ionbal, 
	      struct DEBUG_FILES debug) {

  struct ELEMENT_ION_STATE *state;
  struct LEVEL_RATES *level_rates=NULL;
  struct TRANS_RATES trans_rates; 
  struct EMISSION_LINES *line=NULL;
  struct TWOPHOT_DATA twophot;
  int iEl, irmJ, iRow, nlev, iIonState;
  int Z, rmJ;
  double Te, Ne;
  double *Ion_pop;       /* These relate to the AR data */
  double *Lev_pop=NULL;  /* level population */
  double nfree;
  
  /* Ensure we don't run through this again in CIE mode */
  params->Ndens = 0;
  params->Ntemp = 0;
  params->do_xspectable = FALSE; /* Can't do XSPEC Table model yet */
  
  /* Now do an NEI run */

  /* Loop through the each row of the ionization balance. */
  for (iRow = 0; iRow < ionbal->Nrows; iRow++) {
    Ne = ionbal->dens[iRow];
    
    /* The electron temperature is given in the file */
    Te = ionbal->T[iRow];

    message("apec_nei","T = %e, n = %e",Te, Ne);

    /* Loop through each element to perform various tasks. */
    for (iEl = 0; iEl < params->N_Elements; iEl++) {

      state = Atomic_params[iEl];
      Z = params->list_Z[iEl];
      
      /* Print out the numerical values in structure arrays */
      print_table(&debug, Atomic_params[iEl]);

      Ion_pop = (double *) malloc_safe((Z+1)*sizeof(double));
	
      for(iIonState=0;iIonState<(Z+1);iIonState++) {
	Ion_pop[iIonState] = (ionbal->X[iRow])[ionbal->indx[Z] + iIonState];
      }

      level_rates = (struct LEVEL_RATES *)
	malloc_safe((Z+1)*sizeof(struct LEVEL_RATES));

      /* Loop through the ionization states to get the recombination rates,*/
      /* and therefore be able to distribute them to the upper levels */
      
      /* Table-driven ionization/recombination */
      for (rmJ = 1; rmJ < Z+1; rmJ++) {

	allocate_level_rates(state[rmJ-1].ion_inf->nlev,
			     &(level_rates[rmJ-1]));
	
	/* Calculate ionization rates */
	/*
	calc_ioniz_rate(Te, Ne, &debug, params,&(state[irmJ-1]),
			&(level_rates[irmJ-1].qioniz_total));
	*/
	level_rates[rmJ-1].qioniz_total = 0.0;

	/* We may have recombination to upper levels, and so */
	/* we will need to know the population of the next highest */
	/* ionization state. */
	/* Assign the population for the next ionization state. */
	
	/* Assume recomb_rates from ionization rates and Ion_pops */
	if (rmJ-1 < Z ) {
	  assume_recomb_rates(Ion_pop[rmJ-1],Ion_pop[rmJ],
			      level_rates[rmJ-1].qioniz_total,
			      &(level_rates[rmJ-1].qrecomb_total),
			      &(level_rates[rmJ-1].ionc));
	}
      }
	
      /* ********************************************************* */
      /*          Calculate level population and emission          */
      /* ********************************************************* */

      for (irmJ=0; irmJ < params->list_NrmJ[iEl]; irmJ++)
	alloc_spectrum(Z, params->list_rmJ[iEl][irmJ], &spectrum);

      for (irmJ=0; irmJ < params->list_NrmJ[iEl]; irmJ++) {

	rmJ = params->list_rmJ[iEl][irmJ];
	nlev=state[rmJ-1].ion_inf->nlev;

	if ((nlev > 0) && (Ion_pop[rmJ-1] > MIN_ION_POP)) {
	  
	  /* Allocate space for Lev_pop */
	  if (params->do_lines) {
	    Lev_pop = (double *) malloc_safe(nlev*sizeof(double));
	    calc_trans_rates(Te, Te, Ne, &debug,&(state[rmJ-1]),&trans_rates);
	  }

	  if (params->do_rrc) {
	    calc_rad_rec_cont(Te, &debug, Z, rmJ, &(state[rmJ]),
			      &(state[rmJ-1]), Ion_pop[rmJ],
			      level_rates[rmJ-1].qrecomb, params, 
			      &spectrum);
	  }
	  if (params->do_sat) {
	    calc_satellite(Te, Ne, &debug, &(state[rmJ]),
			   &(state[rmJ-1]),params,Ion_pop[rmJ],
			   level_rates[rmJ-1].qsatel,&line);
	  }
	  /* Set up rate matrix and inversion. Return Lev_pop with */
	  /* total pop of each ion normalized to one. */
	  
	  if (params->do_lines) {
	    solve_rate_eqns(&debug,Ne,&trans_rates,&(level_rates[rmJ-1]),
			    Lev_pop);
	  }
	  
	  /* Calculate line emission. */
	  if (params->do_lines) {
	    nfree = calc_free_particles(params);
	    calc_emission(Te, Ne, &debug,&(state[rmJ-1]),params,
			  Ion_pop[rmJ-1],Lev_pop,nfree,&twophot, &line);
	    
	    calc_two_phot(&debug, &(state[rmJ-1]), &twophot, &spectrum);
	    
	    bin_lines(Z, rmJ, params, line, &spectrum);
	    free(Lev_pop);
	  }
	}
	if ((params->do_brems)&&(Ion_pop[rmJ-1] > MIN_ION_POP)&&(rmJ>1)) {
	  /* Calculate Bremsstrahlung */
	  bremsstrahlung(Z, rmJ, Te, Ion_pop, params, &spectrum);
	}
      }
      
      free(Ion_pop);
      dealloc_level_rates(Z, level_rates);
    }  /* Close the Element loop */
    
    fits_output(params, Te, Ne, iRow, 0, &line, &spectrum);
    clear_spectrum(params->N_Elements, params->list_Z, &spectrum);
    clear_lines(&line);
    
  } /* Close the iRow loop */

}
