#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <fitsio.h>

#include "messages.h"
#include "apec.h"
#include "apec_macros.h"
#include "datatypes.h"

double cheby_series(double x, double *cs, int n);
void fits_error(int status);
void init_brems_hummer(char *filemap_file);
void load_filemap(char *filemap_file, struct FILEMAP **filemap);
int get_filename_from_map(struct FILEMAP *filemap, int Z, int J,
			  int datatype, char **filename);

static double *gaunt_D=NULL;
#define NUM_I 11
#define NUM_J 8

void brems_gaunt_hummer(double *E, double T, int Z, int N, double *g_ff) {

  /* We are passed E (keV), T (K), and Z.  We then calculate */
  /*   u = hv/kT, and gamma2 = Z^2 RYDBERG/kT  */
  /* The gaunt factor is calculated from a Chebyshev expansion to the */
  /* data, using the gaunt_D variable and some lookups. */
  
  double u, gamma2;
  double x_u, x_g;
  double c_j[NUM_J];
  int j, iBin;

  if (gaunt_D == NULL) {
     errmess("brems_hummer","Brems_hummer not initialized");
  }

  gamma2 = Z*Z*RYDBERG/(KBOLTZ*T);
  if ((gamma2 < 1.e-3) || (gamma2 > 1.e3)) {
    if (Z < 10) /* Only if from a low-Z element */
      message("brems_hummer","Warning, gamma^2 = %e is out of range.",gamma2);
    for (iBin=0;iBin<N;iBin++) 
      g_ff[iBin] = 1;  /* Approximate, but at least reasonable */ 
    return; 
  }

  for (iBin=0;iBin<N;iBin++) {

    u = E[iBin]/(KBOLTZ*T);

    if (u < 1.e-4) {
      message("brems_hummer","Warning, u = %e is out of range.",u);
      g_ff[iBin] = 1;  /* Approximate, but at least reasonable */ 
    } else {
      if (u > 31.6227766) {/* Approximate, but brems is suppressed by e^-u */
	g_ff[iBin] = 1.;   /* anyway, so getting the gaunt factor wrong by a */
      } else {             /* little bit (maybe 10x?) isn't so bad. */

	x_u = (2*log10(u) + 2.5)/5.5;
	x_g = log10(gamma2)/3.;

	for (j=0; j < NUM_J; j++) {
	  c_j[j] = cheby_series(x_g, &(gaunt_D[j*NUM_I]), NUM_I);
	}
	
	g_ff[iBin] = cheby_series(x_u, c_j, NUM_J);
	/* message("brems_hummer","Z=%d, T=%e, u=%e, g_ff=%e",
	   Z, T, log10(u), g_ff[iBin]); */
      }
    }
  }
}

void init_brems_hummer(char *filemap_file) { 

  int status=0, hdunum, hdutype, anynul, casesen;
  int col_i, col_j, col_d;
  int i;
  int *data_i, *data_j;
  float *gaunt_D_fl;
  char column[MAXSTRLEN], comment[MAXSTRLEN];
  fitsfile *fptr;
  long num_rows, firstrow, firstelem;
  void *nulval;
  struct FILEMAP *filemap;
  char *fn_brems_hummer;

  if (gaunt_D != NULL) return;  /* Already initialized. */
  load_filemap(filemap_file, &filemap);
  if (get_filename_from_map(filemap, 0, 0, HUMMER_TYPE, &fn_brems_hummer)) {
    errmess("get_abund","Cannot find Hummer brems file in filemap %s",
	    filemap_file);
  }
  if (fits_open_file(&fptr, fn_brems_hummer, READONLY, &status)) {
    message("brems_hummer","Error opening fits file %s",fn_brems_hummer);
    fits_error(status);
  }

  hdunum = 2;  /* Skip the (non-existant) image frame */
  if (fits_movabs_hdu(fptr, hdunum, &hdutype, &status)) {
    message("brems_hummer","Error opening data array in fits file %s",
	    fn_brems_hummer);
    fits_error(status);
  }
  
  casesen = TRUE;
  fits_get_colname(fptr, casesen, GAUNT_HUMMER_IVAL, column, &col_i, &status);
  fits_get_colname(fptr, casesen, GAUNT_HUMMER_JVAL, column, &col_j, &status);
  fits_get_colname(fptr, casesen, GAUNT_HUMMER_DVAL, column, &col_d, &status);
  if (status!=0) {
    message("brems_hummer","Error getting data column names in fits file %s",
	    fn_brems_hummer);
    fits_error(status);
  }

  if (fits_read_key_lng(fptr,"NAXIS2", &num_rows, comment, &status)) {
    message("get_abund","Error getting dimensions in fits file.");
    fits_error(status);
  }

  /* message("brems_hummer","num_rows = %d",num_rows); */

  gaunt_D = (double *) malloc((size_t) num_rows*sizeof(double));
  gaunt_D_fl = (float *) malloc((size_t) num_rows*sizeof(float));
  data_i = (int *) malloc((size_t) num_rows*sizeof(float));
  data_j = (int *) malloc((size_t) num_rows*sizeof(float));

  nulval = NULL;
  firstrow = 1;
  firstelem = 1;

  fits_read_col(fptr, TINT, col_i, firstrow, firstelem, num_rows,
		nulval, data_i, &anynul, &status);
  fits_read_col(fptr, TINT, col_j, firstrow, firstelem, num_rows,
		nulval, data_j, &anynul, &status);
  fits_read_col(fptr, TFLOAT, col_d, firstrow, firstelem, num_rows,
		nulval, gaunt_D_fl, &anynul, &status);

  for (i=0;i<num_rows;i++) gaunt_D[i] = (double) gaunt_D_fl[i];

  if (status!=0) { 
    message("brems_hummer","Error getting column data in fits file %s.",
	    fn_brems_hummer);
    fits_error(status);
  }

  if (fits_close_file(fptr, &status)) {
    message("brems_hummer","Error closing fits file %s.",fn_brems_hummer);
    fits_error(status);
  }

  /* message("brems_hummer","Got the bremsstrahlung gaunt factor data"); */
  free(data_j);
  free(data_i);
  free(gaunt_D_fl); 
}

double cheby_series(double x, double *cs, int n) {

/* 
   c     april 1977 version.  w. fullerton, c3, los alamos scientific lab.
   c
   c evaluate the n-term chebyshev series cs at x.  adapted from
   c r. broucke, algorithm 446, c.a.c.m., 16, 254 (1973).  also see fox
   c and parker, chebyshev polys in numerical analysis, oxford press, p.56.
   c
   c             input arguments --
   c x      value at which the series is to be evaluated.
   c cs     array of n terms of a chebyshev series.  in eval-
   c        uating cs, only half the first coef is summed.
   c n      number of terms in array cs.
   c
   
   Ported to C, 10-DEC-97 by Randall Smith, SAO 
   */

  double b2=0., b1=0., b0=0., twox;
  int i, ni;

  if ((n < 1)||(n > 1000)) {
    errmess("cheby_series","Illegal Number of terms = %d",n);
  }
  
  if ((x < -1.1)||(x > 1.1)) {
    message("cheby_series","x=%e outside legal range of -1,+1",x);
  }
  
  twox = 2.*x;
  for (i=0;i<n;i++) {
    b2 = b1;
    b1 = b0;
    ni = n  - i - 1;
    b0 = twox*b1 - b2 + cs[ni];
  }

  return 0.5 * (b0-b2);
}
