#include <stdlib.h>
#include <math.h>

#include "apec.h"
#include "messages.h"

/*! \file calc_emission.c */ 

void write_debug(struct DEBUG_GENERIC *debug, int ascii, int fits, 
		 char* fmt, ...);
double calc_free_particles(double abund[]);

/*! \brief Calculate emission lines and two photon emission based on level populations 
 *  \param debug Debugging output files [optional; set to NULL to ignore]
 *  \param abund abundance vector with all elemental abundances relative to Hydrogen (needed to calculate free particles per nucleon) 
 *  \param Ion_pop Ion abundance for the ion of interest
 *  \param Lev_pop[] Level population vector for the entire ion
 *  \param state Atomic data for ion of interest
 *  \param Te Electron temperature (in Kelvin) 
 *  \param Ne Electron density (in cm^-3)
 *  \param twophot Two photon output structure; set to NULL to ignore
 *  \param line Line output structure; set to NULL to ignore
 * The code steps through the state database and calculates all the emission
 * line information, specifically:
 * lambda = wavelength (A) of line 
 * epsilon = atomic factor [photons cm+3 s-1] 
 * Intensity (obs at earth) [photons cm-2 s-1]
    = epsilon * Emission Measure/ (4 pi R^2) 
 * Plambda = Line emission in [ergs cm^3 s-1] 
 */

void calc_emission(struct DEBUG_FILES *debug,
		   double abund[], double Ion_pop, double *Lev_pop,
		   struct ELEMENT_ION_STATE *state,
		   double Te, double Ne, 
		   struct TWOPHOT_DATA *twophot,
		   struct EMISSION_LINES **line) {



  int rmJ;
  int ll, lu;
  int nl, do_twophot;
  long max_nwv, nt;
  int Z, Nel;

  double chi;
  double lambda, lambda_err, epsilon, Plambda;
  double A;
  double NH, NetoNH;

  struct EMISSION_LINES *lptr=NULL;
  
  Z = state->ion_inf->Z;
  rmJ = state->ion_inf->rmJ;
  Nel = Z - rmJ + 1;
  max_nwv = state->ion_inf->nwv;
  if (twophot != NULL) twophot->exists = FALSE;

  chi =  calc_free_particles(abund);
  NetoNH = (chi-1)*((abund[1]+abund[2])/abund[1]);
  NH = Ne/NetoNH;

  if (line != NULL) lptr = *line;

  /* Convert Lev_pop, currently normalized to an ionization state
     population of 1.0, to a population density */

  for (nl = 0; nl < state->ion_inf->nlev; nl++) {
    Lev_pop[nl] = Lev_pop[nl] * Ion_pop * abund[Z] * NH;
  }
  
  /* Calculate the line emission  */

  for (nt = 0; nt < max_nwv; nt++) {
    do_twophot = FALSE;
    ll = state->linerad_dat[nt].lev_lower;
    lu = state->linerad_dat[nt].lev_upper;
    lambda = state->linerad_dat[nt].lambda;
    lambda_err = state->linerad_dat[nt].lambda_err;
    A = state->linerad_dat[nt].A;

    /* Check to see if this is a two-photon decay.  There are only two
       important cases: Hydrogenic ions with 2s -> 1s, or He-like ions
       for the 1s2s ^1S -> 1s^2 ^1S.  The He-like ^3S state can give a 
       two-photon decay, but only 1/10,000 of the time, so we ignore it. */
    
    if (twophot != NULL) {
      if ((ll==1)&&((Nel == 1)||(Nel==2))) {/* Hydrogenic or Helium-like ion */
	if ((state->lev_dat[lu-1].N == 2)&&  /* connecting to ground state, */
	    (state->lev_dat[lu-1].L == 0)&&  /* with the correct upper level */
	    (state->lev_dat[lu-1].deg == state->lev_dat[ll-1].deg)) { 
	  do_twophot = TRUE;
	  twophot->exists = TRUE;
	  twophot->Z = Z;
	  twophot->rmJ = state->ion_inf->rmJ;
	  twophot->lambda = lambda;  
	  twophot->lev_pop = Lev_pop[lu-1]/(NH*Ne);
	  twophot->A = A;
	  if (debug != NULL) {
	    if (debug->do_emiss) {
	      write_debug(&debug->emiss, debug->ascii, debug->fits,
			  "%2d %2d %7.2f %7.2f %2d %2d %8.2f %8.2e %8.2e\n",
			  Z,rmJ,log10(Ne),log10(Te),lu,ll,lambda,
			  twophot->lev_pop, 
			  twophot->lev_pop*(HC_IN_ERG_A/lambda));
	    }
	  }
	}
      }
    } 

    if ((line != NULL) && (!do_twophot)) {
      
      if (lambda > 0) {
	if (lptr == NULL) {
	  *line = (struct EMISSION_LINES *)
	    malloc_safe(sizeof(struct EMISSION_LINES));
	  lptr = *line;
	} else {
	  while (lptr->next != NULL) lptr = lptr->next;
	  lptr->next = (struct EMISSION_LINES *)
	    malloc_safe(sizeof(struct EMISSION_LINES));
	  lptr = lptr->next;
	}
	lptr->next = NULL;
	
	epsilon = Lev_pop[lu - 1] * A /(Ne * NH) ; /* phot cm^3 / s */
	Plambda = epsilon * (HC_IN_ERG_A/lambda);  /*  erg cm^3 / s */
	
	lptr->epsilon = Lev_pop[lu - 1] * A /(Ne * NH) ; /* phot cm^3 / s */
	lptr->Plambda = epsilon * (HC_IN_ERG_A/lambda); /*  erg cm^3 / s */
	lptr->lambda  = lambda;  /* Angstroms */
	lptr->lambda_err=lambda_err;  /* Angstroms */
	lptr->Z = Z;
	lptr->rmJ = rmJ;
	lptr->ul= lu;
	lptr->ll= ll;
	lptr->type = BOUNDBOUND_LINE;
	if (debug != NULL) {
	  if ((epsilon > MIN_EPSILON) && (debug->do_emiss)) {
	    write_debug(&debug->emiss, debug->ascii, debug->fits,
			"%2d %2d %7.2f %7.2f %2d %2d %8.2f %8.2e %8.2e\n",
			Z,rmJ,log10(Ne),log10(Te),lu,ll,lambda,epsilon,
			Plambda);
	  }
	}
      } else {
	/* message("calc_emission","Bad line wavelength") */
      }
    }
  }
  if (debug != NULL) {
    if (debug->do_emiss && debug->ascii) fflush(debug->emiss.ascii);
  }
}
