#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "apec.h"
#include "messages.h"

#ifndef M_E
#define M_E 2.7182818284590452354
#endif

/*! \file calc_trans_rates.c */

void  printrates(FILE *qtrans, int rmJ, int nlev, double Te, double Ne, 
		 struct TRANS_RATES *trans_rates);
void  fits_printrates(fd_file *new_qtrans, int rmJ, int nlev, 
		     double Te, double Ne, struct TRANS_RATES *trans_rates);
void calc_maxwell_rates(int coll_type, double min_T, double max_T, 
			double Tarr[], double om[], double dE, double T,
			int Z, double degl, double degu, double *exc_rate, 
			double *dex_rate);
char *roman_element(int Z, int J);
void init_trans_rates(int maxsize, struct TRANS_RATES *trans_rates);
/* double photoexcitation(double lambda, double wt_ratio, 
   double wv, double Aji); */

void init_trans_rates(int maxsize, struct TRANS_RATES *trans_rates) {

  if (trans_rates->maxNlev < maxsize) {
    if (trans_rates->maxNlev != 0) {
      free(trans_rates->qexc);
      free(trans_rates->qp);
      free(trans_rates->Arad);
    }
    /* Allocate space for the largest possible set of transition rates */
    trans_rates->qexc= (double *) malloc_safe(maxsize*maxsize*sizeof(double));
    trans_rates->qp  = (double *) malloc_safe(maxsize*maxsize*sizeof(double));
    trans_rates->Arad= (double *) malloc_safe(maxsize*maxsize*sizeof(double));
    trans_rates->maxNlev = maxsize;
  }
}

/*! \brief Calculate atomic rate coefficients for (de)excitation. 
 *  \param debug Debugging output files [optional; set to NULL to ignore]
 *  \param state Atomic data for ion of interest
 *  \param Te Electron temperature (in Kelvin) 
 *  \param Tp Proton temperature (in Kelvin) 
 *  \param Ne Electron density (in cm^-3)
 *  \param trans_rates Output structure holding all transition rates 
 *  Note that only two-body processes are considered; electron densities
 *  above ~1e16 cm^-3 will not lead to accurate results.
 */


void calc_trans_rates(struct DEBUG_FILES *debug, 
		      struct ELEMENT_ION_STATE *state,
		      double Te, double Tp, double Ne, 
		      struct TRANS_RATES *trans_rates) {

  int i;
  long int ii;
  int j;
  int ll;        /* lower level of transition */
  int lu;        /* upper level of transition */
  long int nexc;
  int nlev;
  int npro;
  long int nwv;

  double degl;   /* statistical weight of lower level */
  double degu;   /* statistical weight of upper level */
  double delE;   /* energy difference */
  double El;     /* energy of lower level */
  double Eu;     /* energy of upper level */
  double logTe;
  double exc_rate, dex_rate;

  struct COLLTRNS_DATA *ptr;

  if (state == NULL) 
    errmess("calc_trans_rates","NULL ion data pointer");
  if (trans_rates == NULL) 
    errmess("calc_trans_rates","NULL transition rate pointer");

  trans_rates->Z = state->ion_inf->Z;
  trans_rates->rmJ = state->ion_inf->rmJ;
  trans_rates->Te = Te;
  trans_rates->Tp = Tp;

  nlev = state->ion_inf->nlev;
  nexc = state->ion_inf->nexc;
  npro = state->ion_inf->npro;
  nwv = state->ion_inf->nwv;

  trans_rates->nlev = nlev;

  /* Storage space for transition rates should be already allocated in  */
  /* init_trans_rates call, but check here. */
  if (nlev > trans_rates->maxNlev) init_trans_rates(nlev, trans_rates);

  /* Initialize the relevant elements of qexc */
  for (i=0; i < nlev; i++) {
    for (j=0; j < nlev; j++) {
      trans_rates->qexc[nlev*i+j] = 0.0;
      trans_rates->qp[nlev*i+j] = 0.0;
      trans_rates->Arad[nlev*i+j] = 0.0;
    }
  }

  logTe = log10(Te);

  /* loop through the number of excitation transitions */
  for (ii=0; ii < nexc+npro; ii++) {
    
    if (ii < nexc) {
      ptr = &(state->colltrns_dat[ii]);
    } else {
      ptr = &(state->protrns_dat[ii-nexc]);
    }
    ll = ptr->lev_lower;
    lu = ptr->lev_upper;
    if ((ll > nlev)||(lu > nlev)||(ll <= 0)||(lu <= 0)) {
      message("calc_trans_rates","Problem with %s, ll=%d, lu=%d",
	      roman_element(trans_rates->Z, trans_rates->rmJ-1), ll, lu);
      if (ii < nexc) {
	errmess("calc_trans_rates","in the %d th electron excitation",ii);
      } else {
	errmess("calc_trans_rates","in the %d th proton excitation",ii-nexc);
      }
    }
    El = state->lev_dat[ll-1].energy;
    Eu = state->lev_dat[lu-1].energy;
    degl = state->lev_dat[ll-1].deg;
    degu = state->lev_dat[lu-1].deg;
    
    delE = Eu - El;  /* in eV */
    if (delE == 0.0) delE = 1.e-8;
    if (delE < 0.0) { /* Swap these levels */
      if (debug != NULL) {
	if (debug->do_qtrans && debug->ascii) {
	  fprintf(debug->qtrans.ascii,
		  "calc_trans_rates: Z, rmJ, dE, l1, l2: %d %d %e %d %d\n",
		  state->ion_inf->Z, state->ion_inf->rmJ, delE, ll, lu);
	}
	if (debug->do_qtrans && debug->fits) {
	  fd_write_row(&debug->qtrans.fits, 
		       state->ion_inf->Z, state->ion_inf->rmJ, delE, ll, lu,
		       0.0, 0.0, 0, 0, 0, 0.0, 0.0, 0.0, 0.0);
	}
      }
      ptr->lev_lower = lu;  /* This is legal, since Omega_ij = Omega_ji */
      ptr->lev_upper = ll;  /* and Upsilon_ij = Upsilon_ji */
      ll = ptr->lev_lower;
      lu = ptr->lev_upper;
      El = state->lev_dat[ll-1].energy;
      Eu = state->lev_dat[lu-1].energy;
      degl = state->lev_dat[ll-1].deg;
      degu = state->lev_dat[lu-1].deg;
      delE = Eu - El;  /* in eV */
    }
    
    if (ii < nexc) {
      calc_maxwell_rates(ptr->coll_type, ptr->min_T, ptr->max_T,ptr->T,ptr->om,
			 delE/1.e3, Te, state->ion_inf->Z, degl, degu, 
			 &exc_rate, &dex_rate);
    } else {
      calc_maxwell_rates(ptr->coll_type, ptr->min_T, ptr->max_T,ptr->T,ptr->om,
			 delE/1.e3, Tp, state->ion_inf->Z, degl, degu, 
			 &exc_rate, &dex_rate);
    }

    trans_rates->qexc[(ll-1)*nlev+lu-1] += Ne*exc_rate;
    trans_rates->qexc[(lu-1)*nlev+ll-1] += Ne*dex_rate;
  }

  for (ii=0; ii < nwv; ii++) {
    lu = state->linerad_dat[ii].lev_upper;
    ll = state->linerad_dat[ii].lev_lower;
    trans_rates->qexc[nlev*(lu-1)+ll-1] += state->linerad_dat[ii].A;
    trans_rates->Arad[nlev*(lu-1)+ll-1] = state->linerad_dat[ii].A;

    /* now add bji to qexc[ll][lu] if appropriate */

    /* 
       if (lambda > 5000. && lu < 4) {
       bji=photoexcitation(lambda,degu/degl,chi*Te,state->linerad_dat[ii].A);
       trans_rates->qexc[nlev*(ll-1)+lu-1] += bji;
       }
     */
  }
  if (debug != NULL) {
    if (debug->do_qtrans) {
      if (debug->ascii) {
	printrates(debug->qtrans.ascii, state->ion_inf->rmJ, nlev, Te, 
		   Ne,trans_rates);
	fflush(debug->qtrans.ascii);
      }
      if (debug->fits) 
	fits_printrates(&debug->qtrans.fits, state->ion_inf->rmJ, nlev, Te, 
			Ne,trans_rates);
    }
  }
}

/* *********************************** */
/*         Diagnostic printout         */
/* *********************************** */

void  printrates(FILE *qtrans, int rmJ, int nlev, double Te, double Ne, 
		 struct TRANS_RATES *trans_rates) {
  int i, j;
  
  fprintf(qtrans," logNe = %8.2f logTe = %8.2f rmJ = %2d\n",
	  log10(Ne),log10(Te),rmJ);
  fprintf(qtrans,
	  " ll   lu   q[ll][lu]   q[lu][ll]   qp[ll][lu]    Arad[lu][ll]\n");

  for (i=0; i < nlev-1; i++) {
    for (j=i+1; j < nlev; j++) {
      if ((trans_rates->qexc[nlev*i+j] != 0.0) ||
	  (trans_rates->qexc[nlev*j+i] != 0.0)) {
	fprintf(qtrans,"%3d%5d %11.3g %11.3g %11.3g %11.3g\n",i+1,j+1,
		trans_rates->qexc[nlev*i+j],trans_rates->qexc[nlev*j+i],
		trans_rates->qp[nlev*i+j],trans_rates->Arad[nlev*j+i]);
      }
    }
  }

}

void  fits_printrates(fd_file *new_qtrans, int rmJ, int nlev, 
		     double Te, double Ne, struct TRANS_RATES *trans_rates) {
  int i, j;
  
  for (i=0; i < nlev-1; i++) {
    for (j=i+1; j < nlev; j++) {
      if ((trans_rates->qexc[nlev*i+j] != 0.0) ||
	  (trans_rates->qexc[nlev*j+i] != 0.0)) {
	fd_write_row(new_qtrans, 0, 0, 0.0, 0, 0, log10(Ne),log10(Te),rmJ,
		     i+1,j+1,
		     trans_rates->qexc[nlev*i+j],trans_rates->qexc[nlev*j+i],
		     trans_rates->qp[nlev*i+j],trans_rates->Arad[nlev*j+i]);
      }
    }
  }
}
