#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <fitsio.h>

#include "apec.h"

#define MAXSTRLEN 1024
#define BITPIX 8

char *roman_element(int Z, int J);
char *element_name(int Z);
void fits_printerror(char *process, int status);

void fits_write_input(fitsfile *fptr, struct PARAMETERS *params) {

  float val_fl;
  int iAtom, Z, irmJ, rmJ;
  int chdu;
  int status=0;

  /* Ensure we're at the correct HDU for outputing this data */
  /* First, save the current HDU */
  fits_get_hdu_num(fptr, &chdu);
  if (chdu != 1) {
    if (fits_movabs_hdu(fptr, 1, NULL, &status)) {
      fits_printerror("moving to primary HDU",status);
    }
  }
			
  fits_write_comment(fptr, "*** BEGIN INITIALIZATION VALUES ***", &status);

  /* ************************************ */
  /* Write out the input file information */
  /* ************************************ */

  fits_write_comment(fptr, "Input Files", &status);

  if (params->ionbal_file[0] != '\0') {
    if (fits_write_key(fptr, TSTRING, "SIONBAL", params->ionbal_file,
		       "Ionization balance file name", &status)) 
      fits_printerror("writing SIONBAL keyword",status);
  }

  if (params->filemap_file[0] != '\0') {
    if (fits_write_key(fptr, TSTRING, "SFILEMAP", params->filemap_file,
		       "Filemap file name", &status)) 
      fits_printerror("writing SFILEMAP keyword",status);
  }

  if (params->abund_source[0] != '\0') {
    if (fits_write_key(fptr, TSTRING, "SABUND_SOURCE", params->abund_source,
		       "Abundance source", &status)) 
      fits_printerror("writing SABUND_SOURCE keyword",status);
  }

  /* ************************************* */
  /* Write out the output file information */
  /* ************************************* */

  fits_write_comment(fptr, "Output Files", &status);

  if (params->nei_output) {
    if (fits_update_key_null(fptr,"NEI_OUTPUT","Output per-ion emission",
			     &status))
      fits_printerror("writing NEI_OUTPUT keyword",status);
  }

  if (params->do_line) {
    if (fits_update_key_null(fptr,"OUTPUT_LINE","Output line list file",
			     &status))
      fits_printerror("writing OUTPUT_LINE keyword",status);
    if (fits_write_key(fptr, TSTRING, "SMODELFILE", params->fn_linespec,
		       "Output file head", &status)) 
      fits_printerror("writing SMODELFILE keyword",status);
  }

  if (params->do_coco) {
    if (fits_update_key_null(fptr,"OUTPUT_COCO",
			     "Output compressed continuum file",
			     &status)) 
      fits_printerror("writing OUTPUT_COCO keyword",status);
    if (fits_write_key(fptr, TSTRING, "SMODELFILE", params->fn_cocospec,
		       "Output file head", &status)) 
      fits_printerror("writing SMODELFILE keyword",status);
  }

  if (params->do_comp) {
    if (fits_update_key_null(fptr,"OUTPUT_COMP",
			     "Output per-ion compressed continuum file",
			     &status)) 
      fits_printerror("writing OUTPUT_COMP keyword",status);
    if (fits_write_key(fptr, TSTRING, "SMODELFILE", params->fn_compspec,
		       "Compressed output file head", &status))
      fits_printerror("writing SMODELFILE keyword",status);
  }

  if (params->do_xspectable) {
    if (fits_write_key(fptr, TSTRING, "SMODELFILE", params->fn_xspectable,
		       "XSPEC Table file name", &status))
      fits_printerror("writing SMODELFILE keyword",status);
  }
  
  if (params->do_bincont) {
    if (fits_update_key_null(fptr,"OUTPUT_BINCONT", 
			     "Output binned continuum file", &status)) 
      fits_printerror("writing OUTPUT_BINCONT keyword",status);
    if (fits_write_key(fptr, TSTRING, "SMODELFILE", params->fn_bincontspec,
		       "Binned continuum file head", &status))
      fits_printerror("writing SMODELFILE keyword",status);
  }

  if ((params->modelname)[0] != '\0'){
    if (fits_write_key(fptr, TSTRING, "SMODELNAME", params->modelname,
		       "XSPEC Model Name", &status))
      fits_printerror("writing SMODELNAME keyword",status);
  }

  /* ***************************************** */
  /* Write out the physical processes included */
  /* ***************************************** */

  fits_write_comment(fptr, "Physical Processes", &status);

  if (params->do_rrc) {
    if (fits_update_key_null(fptr,"DO_RRC",
			     "Include Radiative Recombination Continuum", 
			     &status))
      fits_printerror("writing DO_RRC keyword",status);
  } else {
    if (fits_update_key_null(fptr, "NO_RRC", 
			     "No Radiative Recombination Continuum", 
			     &status))
      fits_printerror("writing NO_RRC keyword",status);
  }

  if (params->do_sat) {
    if (fits_update_key_null(fptr,"DO_SATEL",
			     "Include Dielectronic Satellite Lines", &status))
      fits_printerror("writing DO_SATEL keyword",status);
  } else {
    if (fits_update_key_null(fptr, "NO_SATEL", 
			     "No Dielectronic Satellite Lines", &status))
      fits_printerror("writing NO_SATEL keyword",status);
  }

  if (params->do_lines) {
    if (fits_update_key_null(fptr,"DO_LINES", "Include Emission Lines", 
			     &status))
      fits_printerror("writing DO_LINES keyword",status);
  } else {
    if (fits_update_key_null(fptr, "NO_LINES", "No Emission Lines", &status))
      fits_printerror("writing NO_LINES keyword",status);
  }

  if (params->do_twoph) {
    if (fits_update_key_null(fptr,"DO_TWOPH", "Include Two Photon Continuum", 
			     &status))
      fits_printerror("writing DO_TWOPH keyword",status);
  } else {
    if (fits_update_key_null(fptr, "NO_TWOPH",  "No Two Photon Continuum", 
			     &status))
      fits_printerror("writing NO_TWOPH keyword",status);
  }

  if (params->do_brems) {
    if (fits_update_key_null(fptr,"DO_BREMS", 
			     "Include Bremsstrahlung Continuum", &status))
      fits_printerror("writing DO_BREMS keyword",status);
    if (params->Brems_Type == HUMMER) {
      if (fits_write_key(fptr, TSTRING, "SBREMS_TYPE", "Hummer",
			 "Bremsstrahlung type", &status))
	fits_printerror("writing SBREMS_TYPE keyword",status);
    }
    if (params->Brems_Type == KELLOGG) {
      if (fits_write_key(fptr, TSTRING, "SBREMS_TYPE", "Kellogg",
			 "Bremsstrahlung type", &status))
	fits_printerror("writing SBREMS_TYPE keyword",status);
    }
    if (params->Brems_Type == RELATIVISTIC) {
      if (fits_write_key(fptr, TSTRING, "SBREMS_TYPE", "Relativistic",
			 "Bremsstrahlung type", &status))
	fits_printerror("writing SBREMS_TYPE keyword",status);
    }
    if (params->Brems_Type == BREMS_NONE) {
      if (fits_write_key(fptr, TSTRING, "SBREMS_TYPE", "None",
			 "Bremsstrahlung type", &status))
	fits_printerror("writing SBREMS_TYPE keyword",status);
    }
  } else {
    if (fits_update_key_null(fptr, "NO_BREMS","No Bremsstrahlung Continuum",
			     &status))
      fits_printerror("writing NO_BREMS keyword",status);
  }

  /* ****************************************** */
  /* Write out the spectral binning information */
  /* ****************************************** */

  fits_write_comment(fptr, "Spectral Binning", &status);

  switch(params->BinType) {
  case(BIN_LIN_WAVE):
    if (fits_update_key_null(fptr, "BIN_LIN_WAVE", 
			     "Linear wavelength spacing",
			     &status))
      fits_printerror("writing BIN_LIN_WAVE keyword",status);
    break;
  case (BIN_LOG_WAVE):
    if (fits_update_key_null(fptr, "BIN_LOG_WAVE", 
			     "Log. wavelength spacing",
			     &status))
      fits_printerror("writing BIN_LOG_WAVE keyword",status);
    break;
  case(BIN_LIN_ENERGY):
    if (fits_update_key_null(fptr, "BIN_LIN_ENERGY", 
			     "Linear energy spacing",
			     &status))
      fits_printerror("writing BIN_LIN_ENERGY keyword",status);
    break;
  case (BIN_LOG_ENERGY):
    if (fits_update_key_null(fptr, "BIN_LOG_ENERGY", "Log. energy spacing",
			     &status))
      fits_printerror("writing BIN_LOG_ENERGY keyword",status);
    break;
  }

  if ((params->BinType == BIN_LIN_WAVE)||(params->BinType==BIN_LOG_WAVE)) {
  if (fits_write_key(fptr, TINT, "INUM_LAMBDA", &(params->Nbin),
		     "Number of wavelength bins", &status))
    fits_printerror("writing INUM_LAMBDA keyword",status);

  val_fl = (float) params->BinMin;
  if (fits_write_key(fptr, TFLOAT, "DLAMBDA_START", &val_fl,
		     "Starting wavelength bin", &status))
      fits_printerror("writing DLAMBDA_START keyword",status);

  val_fl = (float) params->BinMax;
  if (fits_write_key(fptr, TFLOAT, "DLAMBDA_END", &val_fl,
		     "Final wavelength bin", &status))
      fits_printerror("writing DLAMBDA_END keyword",status);
  }

  if ((params->BinType==BIN_LIN_ENERGY)||
      (params->BinType==BIN_LOG_ENERGY)) {
    if (fits_write_key(fptr, TINT, "INUM_E", &(params->Nbin),
		     "Number of energy bins", &status))
    fits_printerror("writing INUM_E keyword",status);

    val_fl = (float) params->BinMin;
    if (fits_write_key(fptr, TFLOAT, "DE_START", &val_fl,
		       "Starting energy bin", &status))
      fits_printerror("writing DE_START keyword",status);

    val_fl = (float) params->BinMax;
    if (fits_write_key(fptr, TFLOAT, "DE_END", &val_fl,
		       "Final energy bin", &status))
      fits_printerror("writing DE_END keyword",status);
  }

  val_fl = (float) params->min_epsilon;
  if (fits_write_key(fptr, TFLOAT, "DMIN_EPSILON", &val_fl,
		     "[ph cm^3/s] Minimum output line emissivity", &status))
    fits_printerror("writing DMIN_EPSILON keyword",status);
  
  /* ********************************* */
  /* Write out the physical parameters */
  /* ********************************* */

  fits_write_comment(fptr, "Physical Parameters", &status);

  if (params->Dens_linear) {
    if (fits_update_key_null(fptr, "LIN_DENS", "Linear density spacing",
			     &status))
      fits_printerror("writing LIN_DENS keyword",status);
  } else {
    if (fits_update_key_null(fptr, "LOG_DENS", 
			     "Logarithmic density spacing",
			     &status))
      fits_printerror("writing LOG_DENS keyword",status);
  }

  if (fits_write_key(fptr, TINT, "INUM_DENSITIES", &(params->Ndens),
		     "Number of densities", &status))
    fits_printerror("writing INUM_DENSITIES keyword",status);

  val_fl = (float) params->Dens_Start;
  if (fits_write_key(fptr, TFLOAT, "DDENSITY_START", &val_fl,
		     "Starting density", &status))
      fits_printerror("writing DDENSITY_START keyword",status);

  val_fl = (float) params->Dens_Step;
  if (fits_write_key(fptr, TFLOAT, "DDENSITY_STEP", &val_fl,
		     "Density step size", &status))
      fits_printerror("writing DDENSITY_STEP keyword",status);

  if (params->T_linear) {
    if (fits_update_key_null(fptr, "LIN_TEMP", 
			     "Linear temperature spacing",
			     &status))
      fits_printerror("writing LIN_TEMP keyword",status);
  } else {
    if (fits_update_key_null(fptr,"LOG_TEMP",
			     "Logarithmic temperature spacing",
			     &status))
      fits_printerror("writing LOG_TEMP keyword",status);
  }

  if (fits_write_key(fptr, TINT, "INUM_TEMP", &(params->Ntemp),
		     "Number of temperatures", &status))
    fits_printerror("writing INUM_TEMP keyword",status);

  val_fl = (float) params->T_Start;
  if (fits_write_key(fptr, TFLOAT, "DTEMP_START", &val_fl,
		     "Starting temperature", &status))
      fits_printerror("writing DTEMP_START keyword",status);

  val_fl = (float) params->T_Step;
  if (fits_write_key(fptr, TFLOAT, "DTEMP_STEP", &val_fl,
		     "Temperature step size", &status))
      fits_printerror("writing DTEMP_STEP keyword",status);
  
  /* ********************************* */
  /* Write out the atoms/ions included */
  /* ********************************* */

  fits_write_comment(fptr, "Atoms Included", &status);

  for (iAtom=0;iAtom<params->N_Elements;iAtom++) {
    Z   = (params->list_Z)[iAtom];
    if (params->list_NrmJ[iAtom] == Z+1) { /* Do everything */
      if (fits_write_key(fptr, TSTRING, "SATOM", element_name(Z),
			 "Element name", &status))
	fits_printerror("writing SATOM keyword",status);
    } else {
      for (irmJ=0;irmJ<params->list_NrmJ[iAtom];irmJ++) {
	rmJ = params->list_rmJ[iAtom][irmJ];
	if (fits_write_key(fptr, TSTRING, "SATOM", roman_element(Z,rmJ-1),
			   "Element name", &status))
	  fits_printerror("writing SATOM keyword",status);
      }
    }

  }

  if (fits_write_comment(fptr, "*** END INITIALIZATION VALUES ***",&status))
      fits_printerror("writing final comment",status);    

  /* Move back to correct HDU, if needed */
  if (chdu != 1) {
    if (fits_movabs_hdu(fptr, chdu, NULL, &status)) {
      fits_printerror("returning to original HDU",status);
    }
  }

}
