#include <stdlib.h>
#include <stdio.h>

#ifdef USE_PFILE
#include "pfile.h"
#else
#include "ape/ape_error.h"
#include "ape/ape_msg.h"
#include "ape/ape_par.h"
#include "ape/ape_trad.h"

enum {
  PF_UNKNOWN_TYPE,
  PF_BOOLEAN_TYPE,
  PF_DOUBLE_TYPE,
  PF_FILE_TYPE,
  PF_FLOAT_TYPE,
  PF_INTEGER_TYPE,
  PF_STRING_TYPE
};

#include <assert.h>

#endif

#include "messages.h"
#include "apec.h"

struct atomstr {
  int Z;
  int rmJ;
  char File[MAXSTRLEN];
  struct atomstr *next;
};

#define TRUE 1
#define FALSE 0

static char *filemap_file;
static int require_checksum;
static char *ionization;
static int useionbalancetable;
static char *ionbal_file;
static char *abund_source;
static double min_epsilon;
static char *tempunits;
static int T_linear;
static double T_Start;
static double T_Step;
static int Ntemp;
static int Dens_linear;
static double Dens_Start;
static double Dens_Step;
static int Ndens;
static char *grid_units;
static int linear_grid;
static double BinMin;
static double BinMax;
static int Nbin;
static char *inc_atoms;
static char *inc_ions;
static int do_rrc;
static int do_sat;
static int do_lines;
static int do_twoph;
static int do_brems;
static char *brems_type;
static char *modelname;
static char *outputtype;
static int nei_output;
static char *debugformat;
static char *debugfitsformat;
static int deb_do_emiss;
static int deb_do_ionbal;
static int deb_do_table;
static int deb_do_qioniz;
static int deb_do_qrec;
static int deb_do_qtrans;
static int deb_do_rates;
static int deb_do_satel;
static int deb_do_twoph;
static int deb_do_levpop;
static int deb_do_compress;

static Param_Table_Type Control_Parm_Table [] = {
  {"FileMap",             PF_FILE_TYPE,    &filemap_file},
  {"RequireChecksum",     PF_BOOLEAN_TYPE, &require_checksum},
  {"Ionization",          PF_STRING_TYPE,  &ionization},
  {"UseIonBalanceTable",  PF_BOOLEAN_TYPE, &useionbalancetable},
  {"IonBalanceTable",     PF_FILE_TYPE,    &ionbal_file},
  {"Abundances",          PF_STRING_TYPE,  &abund_source},
  {"MinEpsilon",          PF_DOUBLE_TYPE,  &min_epsilon},
  {"TempUnits",           PF_STRING_TYPE,  &tempunits},
  {"LinearTemp",          PF_BOOLEAN_TYPE, &T_linear},
  {"TempStart",           PF_DOUBLE_TYPE,  &T_Start},
  {"TempStep",            PF_DOUBLE_TYPE,  &T_Step},
  {"NumTemp",             PF_INTEGER_TYPE, &Ntemp},
  {"LinearDens",          PF_BOOLEAN_TYPE, &Dens_linear},
  {"DensStart",           PF_DOUBLE_TYPE,  &Dens_Start},
  {"DensStep",            PF_DOUBLE_TYPE,  &Dens_Step},
  {"NumDens",             PF_INTEGER_TYPE, &Ndens},
  {"GridUnits",           PF_STRING_TYPE,  &grid_units},
  {"LinearGrid",          PF_BOOLEAN_TYPE, &linear_grid},
  {"GridMinimum",         PF_DOUBLE_TYPE,  &BinMin},
  {"GridMaximum",         PF_DOUBLE_TYPE,  &BinMax},
  {"NumGrid",             PF_INTEGER_TYPE, &Nbin},
  {"IncAtoms",            PF_STRING_TYPE,  &inc_atoms},
  {"IncIons",             PF_STRING_TYPE,  &inc_ions},
  {"RRC",                 PF_BOOLEAN_TYPE, &do_rrc},
  {"DRSatellite",         PF_BOOLEAN_TYPE, &do_sat},
  {"EmissionLines",       PF_BOOLEAN_TYPE, &do_lines},
  {"TwoPhoton",           PF_BOOLEAN_TYPE, &do_twoph},
  {"Bremsstrahlung",      PF_BOOLEAN_TYPE, &do_brems},
  {"BremsType",           PF_STRING_TYPE,  &brems_type},
  {"OutputFileStem",      PF_STRING_TYPE,  &modelname},
  {"OutputType",          PF_STRING_TYPE,  &outputtype},
  {"OutputPerIon",        PF_BOOLEAN_TYPE, &nei_output},
  {"DebugFormat",         PF_STRING_TYPE,  &debugformat},
  {"DebugFitsFormat",     PF_STRING_TYPE,  &debugfitsformat},
  {"DebugEmission",       PF_BOOLEAN_TYPE, &deb_do_emiss},
  {"DebugIonBalance",     PF_BOOLEAN_TYPE, &deb_do_ionbal},
  {"DebugTable",          PF_BOOLEAN_TYPE, &deb_do_table},
  {"DebugIonization",     PF_BOOLEAN_TYPE, &deb_do_qioniz},
  {"DebugRecombination",  PF_BOOLEAN_TYPE, &deb_do_qrec},
  {"DebugTransition",     PF_BOOLEAN_TYPE, &deb_do_qtrans},
  {"DebugRates",          PF_BOOLEAN_TYPE, &deb_do_rates},
  {"DebugSatellite",      PF_BOOLEAN_TYPE, &deb_do_satel},
  {"DebugTwophoton",      PF_BOOLEAN_TYPE, &deb_do_twoph},
  {"DebugLevelPopulation",PF_BOOLEAN_TYPE, &deb_do_levpop},
  {"DebugCompression",    PF_BOOLEAN_TYPE, &deb_do_compress},
  {NULL, 0, NULL} };

int strip_strcmp(const char *s1, const char *s2);
void parse_atom(char *inc_atoms, char *inc_ions, struct atomstr **atoms);
void extract_atom_input(struct PARAMETERS *params, struct atomstr *atoms);
void setup_debug(char *debugformat, char *debugfitsformat,
		 struct DEBUG_FILES *debug);
void expand_env(char *string);

#ifndef USE_PFILE
static int get_parameters(Param_Table_Type *params);
#endif

void apec_par_init(int argc, char *argv[], Param_File_Type *p_input,
  struct PARAMETERS *params, struct DEBUG_FILES *debug) {

#ifdef USE_PFILE
  Param_File_Type *p=NULL;
  char *Parameter_File=NULL;
#else
  int status=0;
  int ii=0;
#endif

  /* ************************** */
  /* Check the input parameters */
  /* ************************** */

#ifdef USE_PFILE
  if (p_input == NULL) {
    if (NULL == (p = pf_parse_cmd_line ("apec.par", NULL, argc, argv))) {
      errmess ("apec_par_init", "apec: Error opening parameter file.\n");
    }
  } else {
    p = p_input;
  }

  if (NULL == (Parameter_File = pf_get_output_filename (p))) {
    pf_close_parameter_file (p);
    errmess ("apec_par_init", "apec: Cannot get output parameter file name.\n");
  }
   
  if (-1 == pf_get_parameters (p, Control_Parm_Table)) {
    errmess ("apec_par_init", "apec: error getting parameters.");
  }
#else
  if (p_input == NULL) {
    /* Make a copy of the arguments, including name of apec par file and
       NULL one-past-last pointer. */
    char **new_argv = (char **) calloc(argc + 2, sizeof(char *));
    
    if (NULL == new_argv) {
      errmess ("apec_par_init", "apec: Unable to allocate array for command line arguments.\n");
    }

    new_argv[0] = "apec";

    for (ii = 0; ii < argc; ++ii) {
      new_argv[ii+1] = argv[ii];
    }

    if (eOK != (status = ape_trad_init (argc+1, new_argv))) {
      free (new_argv);
      errmess ("apec_par_init", "apec: Error opening parameter file.\n");
    } else {
      free (new_argv);
    }
  } else {
    /* The following pfile-specific case is not supported in Ape for the moment. */
    /* p = p_input; */
    assert(0);
  }

  if (eOK == status) {
    if (eOK != (status = get_parameters (Control_Parm_Table))) {
      ape_trad_close (0);
      fprintf ("apec_par_init", "apec: Error getting parameters (Ape error code %d).\n", status);
    } else {
      ape_trad_close (0);
    }
  }
#endif
}

void get_input_data(int argc, char *argv[], Param_File_Type *p_input,
		    struct PARAMETERS *params, 
		    struct DEBUG_FILES *debug) {

  int i, iAtom;
  struct atomstr *atoms=NULL;
  char tmpstr[MAXSTRLEN];

  /* Read input parameters and populate params and debug structures. */
  apec_par_init(argc, argv, p_input, params, debug);

  params->N_Elements = 0;
  params->list_Z = NULL;
  params->list_NrmJ = NULL;
  for (iAtom=0;iAtom<MAXIMUM_Z+1;iAtom++) (params->list_rmJ)[iAtom] = NULL;
  
  strncpy(tmpstr,filemap_file,MAXSTRLEN);
  expand_env(tmpstr);
  strcpy(params->filemap_file, tmpstr);
  params->require_checksum = require_checksum;
  strncpy(tmpstr,ionbal_file,MAXSTRLEN);
  expand_env(tmpstr);
  strcpy(params->ionbal_file, tmpstr);
  strcpy(params->abund_source, abund_source);
  params->min_epsilon = min_epsilon;
  params->T_linear = T_linear;
  params->T_Start=T_Start;
  params->T_Step =T_Step;
  params->Ntemp= Ntemp;
  params->Dens_linear = Dens_linear;
  params->Dens_Start = Dens_Start;
  params->Dens_Step = Dens_Step;
  params->Ndens = Ndens;
  params->BinMin = BinMin;
  params->BinMax = BinMax;
  params->Nbin = Nbin;
  params->do_rrc = do_rrc;
  params->do_sat = do_sat;
  params->do_lines = do_lines;
  params->do_twoph = do_twoph;
  params->do_brems = do_brems;
  strcpy(params->modelname,modelname);
  params->nei_output = nei_output;
  debug->do_emiss   = deb_do_emiss;  
  debug->do_ionbal  = deb_do_ionbal;  
  debug->do_table   = deb_do_table;   
  debug->do_qioniz  = deb_do_qioniz;  
  debug->do_qrec    = deb_do_qrec;    
  debug->do_qtrans  = deb_do_qtrans;  
  debug->do_rates   = deb_do_rates;   
  debug->do_satel   = deb_do_satel;   
  debug->do_twoph   = deb_do_twoph;   
  debug->do_levpop  = deb_do_levpop;  
  debug->do_compress= deb_do_compress;

  /* Setup debug info */
  setup_debug(debugformat, debugfitsformat, debug);  

  /* NEI or CIE calculation */
  if (strip_strcmp(ionization,"NEI")==0) {
    params->nei=TRUE;
  } else if (strip_strcmp(ionization,"CIE")==0) {
    params->nei=FALSE; 
  } else {
    errmess("get_input_data","Ionization type %s not recognized.",ionization);
  }
  
  /* Spectral Binning info */
  
  if (strip_strcmp(grid_units,"angstrom")==0) {
    if (linear_grid) {
      params->BinType = BIN_LIN_WAVE;
    } else {
      params->BinType = BIN_LOG_WAVE;
    }
  } else if (strip_strcmp(grid_units,"keV")==0) {
    if (linear_grid) {
      params->BinType = BIN_LIN_ENERGY;
    } else {
      params->BinType = BIN_LOG_ENERGY;
    }
  } else {
    errmess("get_input_data","Grid units %s not recognized.",grid_units);
  }

  /* Density range info */
  params->density = (double *) malloc(params->Ndens*sizeof(double));
  for (i=0;i<params->Ndens;i++) {
    if (params->Dens_linear) { /* Linear by default */
      (params->density)[i] = params->Dens_Start + i*params->Dens_Step;
    } else {
      (params->density)[i] = pow(10.,params->Dens_Start + i*params->Dens_Step);
    }
  }
  
  /* Temperature range info */
  params->temperature = (double *) malloc(params->Ntemp*sizeof(double));
  for (i=0;i<params->Ntemp;i++) {
    if (params->T_linear) { /* Linear by default */
      params->temperature[i] = params->T_Start + i*params->T_Step;
    } else {
      params->temperature[i] = pow(10.,params->T_Start + i*params->T_Step);
    } 
    if (strip_strcmp(tempunits,"keV")==0) params->temperature[i] /= KBOLTZ;
  }

  /* Ion Balance file */   

  params->use_ionbal_table = FALSE;
  if (useionbalancetable) {
    debug->ionbal_read_or_calc = IONBAL_READ;
    params->use_ionbal_table = TRUE;
  }

  /* Output Model Names */
  
  /* Now, fill in all the other file names from the main stem */
  /* Suffixes will be added by the fits_spec_init routine. */
  strcpy(params->fn_linespec,params->modelname);
  strcpy(params->fn_cocospec,params->modelname);
  strcpy(params->fn_compspec,params->modelname);
  strcpy(params->fn_bincontspec,params->modelname);
  sprintf(params->fn_xspectable,"%s.xspec",params->modelname);
  
  if ((strip_strcmp(outputtype,"full")==0)||
      (strip_strcmp(outputtype,"all")==0)) {
    params->do_line = TRUE;
    params->do_coco = TRUE;
    params->do_comp = TRUE;
    params->do_bincont = TRUE;
    params->do_xspectable = TRUE;
  }
  
  if (strip_strcmp(outputtype,"small")==0) {
    params->do_line = TRUE;
    params->do_coco = TRUE;
    params->do_comp = FALSE;
    params->do_bincont = FALSE;
    params->do_xspectable = FALSE;
  }
  
  /* Bremsstrahlung info */
  if (strip_strcmp(brems_type,"Hummer")==0) 
    {params->Brems_Type = HUMMER; }
  else if (strip_strcmp(brems_type,"Kellogg")==0) 
    {params->Brems_Type = KELLOGG; }
  else if (strip_strcmp(brems_type,"Relativistic")==0) 
    {params->Brems_Type = RELATIVISTIC; }
  else if (strip_strcmp(brems_type,"None")==0) 
    {params->Brems_Type = BREMS_NONE; }
  else {
    errmess("get_input_data","Bremsstrahlung type %s was not found.",
	    brems_type);
  }
  
  /* What atoms/ions to use?   In the case where an entire atom is
     passed to the code, all the ions are set up.  If individual ion states
     are given, a single state for each partial atom is created, filling in 
     only those ions requested. */
  strcpy(params->inc_atoms, inc_atoms);
  strcpy(params->inc_ions, inc_ions);
  parse_atom(inc_atoms, inc_ions, &atoms);
  extract_atom_input(params, atoms);
  
}

int strip_strcmp(const char *s1, const char *s2) {
  
  /* Strips any whitespace from string s1, and compares it to s2. */

  char strip_string[MAXSTRLEN];

  /*  strcpy(strip_string,s1); */
  sscanf(s1,"%s",strip_string);
  return strcmp(strip_string,s2);
}

#ifndef USE_PFILE
static int get_parameters(Param_Table_Type *params) {
  int status=eOK;
  for (; NULL!=params->name && eOK == status; ++params) {
    /* Handle strings a little carefully, as APEC expects them all to be
       MAXSTRLEN long, but ape allocates just enough space. */
    char * tmp_str = 0;
if (0 != tmp_str) fprintf(stderr, "You're wrong, JP!!!!\n");
    switch (params->type) {
      case PF_BOOLEAN_TYPE: {
        int *value = (int *) params->value;
        char bool_value='\0';
        status = ape_trad_query_bool(params->name, &bool_value);
        *value = bool_value;
        break;
      }
      case PF_DOUBLE_TYPE:
        status = ape_trad_query_double(params->name, (double *) params->value);
        break;
      case PF_FILE_TYPE:
        status = ape_trad_query_file_name(params->name, &tmp_str);
        break;
      case PF_FLOAT_TYPE:
        status = ape_trad_query_float(params->name, (float *) params->value);
        break;
      case PF_INTEGER_TYPE:
        status = ape_trad_query_int(params->name, (int *) params->value);
        break;
      case PF_STRING_TYPE:
        /* Try to get string, matching the case of enumerated parameters. */
        status = ape_trad_query_string_case(params->name, &tmp_str, eEnumCase);
        /* If the parameter wasn't enumerated, that's OK, just ignore the error code. */
        if (eRangeNoEnum == status) status = eOK;
        break;
      default:
        status = -1;
    }
    /* If tmp_str was used, copy it over to the actual parameter value. */
    if (eOK == status && 0 != tmp_str) {
      char * real_str = calloc(MAXSTRLEN, sizeof(char));
      if (0 == real_str) status = eDynAllocFailed;
      if (eOK == status) {
        strncpy(real_str, tmp_str, MAXSTRLEN - 1);
        *(char **) params->value = real_str;
      }
      free(tmp_str);
    }
  }
  return status;
}
#endif
