#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <fitsio.h>

#include "messages.h"
#include "apec.h"
#include "datatypes.h"

struct IONREC_TMP {
  struct IONREC_DATA p;
  struct IONREC_LINE_DATA l;
  struct IONREC_TMP *next;
};

void read_fits_ionrec_row(fitsfile *fptr, int row, int Z, struct IONREC_DATA *ptr,
			  struct IONREC_LINE_DATA *line, 
			  int colnum[], char *colnames[]);

unsigned long read_fits_ionrec_data(int require_checksum,
				    int Z, int rmJ, fitsfile *fptr, 
				    int *Nvalues, struct IONREC_DATA **);
unsigned long check_fits_file(fitsfile *fptr);
void fits_printerror(char *process, int status);

unsigned long read_fits_ionrec_data(int require_checksum, 
				    int Z, int rmJ, fitsfile *fptr, 
				    int *Nvalues,
				    struct IONREC_DATA **ion_dat_ptr) {

  /* Input assumptions: 
     fptr is a fits file ptr, indexed to the correct HDU. 
     ion_dat_ptr is an IONREC_DATA pointer for the data in that HDU.
     Output requirements:
     ion_dat should be loaded up with all the appropriate data.  
   */
  
  int Nrows, iCol, iRow, i;
  int status=0;
  char *colnames[15] = {"ION_FINAL","LEVEL_FINAL","LEVEL_INIT",
			"TR_TYPE","TR_INDEX","PAR_TYPE",
			"MIN_TEMP","MAX_TEMP","TEMPERATURE","IONREC_PAR",
			"WAVELEN","WAVE_OBS","WAVE_ERR","BR_RATIO",
			"BR_RAT_ERR"};
  int colnum[15]; /* 15 names listed above */
  int Ncols = 15;
  int Zel;
  char comment[MAXSTRLEN];
  struct IONREC_TMP *top, *ptr;
  struct IONREC_DATA *ion_dat;
  struct IONREC_LINE_DATA *lptr;
  unsigned long datasum = 0;

  /* First, check that the is file OK and return the datasum */
  if (require_checksum) datasum = check_fits_file(fptr);
    
  /* Figure out how many values we have to load. */
  fits_read_key(fptr, TINT, "NAXIS2", &Nrows, comment, &status);
  if (status!=0) {
    fitsmess("read_fits_ionrec_data",status, "Error reading NAXIS2 header value");
    status=0;
  }
  
  /* Figure out what element we're dealing with */
  fits_read_key(fptr, TINT, "ELEMENT", &Zel, comment, &status);
  if (status!=0) {
    fitsmess("read_fits_ionrec_data",status, "Error reading ELEMENT header value");
    status=0;
  }

  if (Nrows == 0) { /* No actual data here. */
    *ion_dat_ptr = NULL;
    *Nvalues = Nrows;
    return datasum; 
  }
  /* Get the column numbers */
  
  for(iCol=0;iCol<Ncols;iCol++) {
    fits_get_colnum(fptr, CASEINSEN, colnames[iCol], &(colnum[iCol]), 
		    &status);
    if (status!=0) {
      fitsmess("read_fits_ionrec_data",status,
	       "Error reading %s column number", colnames[iCol]);
      status=0;
    }
  }

  /* Now, read in the data. */
  top = (struct IONREC_TMP *) malloc(sizeof(struct IONREC_TMP));
  ptr = top;
  for (iRow=1;iRow<Nrows+1;iRow++) {
    read_fits_ionrec_row(fptr, iRow, Zel, &(ptr->p), &(ptr->l), colnum, colnames);
    if (iRow < Nrows) {
      ptr->next = (struct IONREC_TMP *) malloc(sizeof(struct IONREC_TMP));
      ptr = ptr->next;
    } else {
      ptr->next = NULL;
    }
  }
  
  /* Now, sort through the data.  Step 1: figure out how many transitions */
  
  Nrows = 0;
  ptr = top;
  while (ptr != NULL) {
    if ((ptr->p.element == Z) &&
	(ptr->p.ion_init != rmJ) &&
	(ptr->p.par_type != 0)) Nrows ++;
    ptr = ptr->next;
  }
  
  /* Make space for the data */
  *ion_dat_ptr = (struct IONREC_DATA *)
    malloc_safe(Nrows*sizeof(struct IONREC_DATA));
  ion_dat = *ion_dat_ptr;

  ptr = top;
  iRow = 0;
  while (ptr != NULL) {
    if ((ptr->p.element == Z) && 
	(ptr->p.ion_init == rmJ) && 
	(ptr->p.par_type != 0)) {
      ion_dat[iRow].element     = ptr->p.element;
      ion_dat[iRow].ion_init    = ptr->p.ion_init;
      ion_dat[iRow].ion_final   = ptr->p.ion_final;
      ion_dat[iRow].level_final = ptr->p.level_final;
      ion_dat[iRow].level_init  = ptr->p.level_init ;   
      ion_dat[iRow].tr_type     = ptr->p.tr_type;
      ion_dat[iRow].tr_index    = ptr->p.tr_index;     
      ion_dat[iRow].par_type    = ptr->p.par_type;     
      ion_dat[iRow].min_temp    = ptr->p.min_temp;  
      ion_dat[iRow].max_temp    = ptr->p.max_temp;  
      for (i=0;i<MAX_IONREC;i++) {
	ion_dat[iRow].Temperature[i] = ptr->p.Temperature[i];  
	ion_dat[iRow].IonRec_Par[i]  = ptr->p.IonRec_Par[i];  
      }
      if (ptr->l.lambda > 0) {
	ion_dat[iRow].lines = (struct IONREC_LINE_DATA *)
	  malloc_safe(sizeof(struct IONREC_LINE_DATA));
	ion_dat[iRow].lines->lambda = ptr->l.lambda;
	ion_dat[iRow].lines->lambda_err = ptr->l.lambda_err;
	ion_dat[iRow].lines->br_ratio = ptr->l.br_ratio;
	ion_dat[iRow].lines->next = NULL;
      }
      iRow++;
    }
    ptr = ptr->next;
  }
  
  /* Now we've got all the transition data in place.  Now for the lines. */
  ptr = top;
  while (ptr != NULL) {
    if ((ptr->p.element == Z) && 
	(ptr->p.ion_init == rmJ) && 
	(ptr->p.par_type == 0)) { /* Has line, no data */
      for (iRow=0;iRow<Nrows;iRow++) {
	if (ion_dat[iRow].tr_index == ptr->p.tr_index) {
	  lptr = ion_dat[iRow].lines;
	  while (lptr->next != NULL) lptr = lptr->next;
	  lptr->next = (struct IONREC_LINE_DATA *)
	    malloc_safe(sizeof(struct IONREC_LINE_DATA));
	  lptr = lptr->next;
	  lptr->lambda = ptr->l.lambda;
	  lptr->lambda_err = ptr->l.lambda_err;
	  lptr->br_ratio = ptr->l.br_ratio;
	  lptr->next = NULL;
	}
      }
    }
    top = ptr;
    ptr = ptr->next;
    free(top); /* Clear out the memory */ 
  }
  
  *Nvalues = Nrows;
  return datasum;
}

void read_fits_ionrec_row(fitsfile *fptr, int row, int Z, struct IONREC_DATA *ptr,
			  struct IONREC_LINE_DATA *line, 
			  int colnum[], char *colnames[]) {
    
  int status=0, anynul, i;
  char *ionrec_type;
  float var_float, arr_float[MAX_IONREC];
  float fnul_val;

  ionrec_type = (char *) malloc(MAXSTRLEN);

  ptr->element = Z;
  /*
  fits_read_col(fptr, TINT, colnum[0], row, 1, 1, 0, &(ptr->element), 
		&anynul, &status);
  if (status!=0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data", 
	     colnames[0]);
  }
  */

  /*
  fits_read_col(fptr, TINT, colnum[0], row, 1, 1, 0, &(ptr->ion_init), 
		&anynul, &status)
  if (status!=0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data", 
	     colnames[0]);
  }
  */

  fits_read_col(fptr, TINT, colnum[0], row, 1, 1, 0, &(ptr->ion_final), 
		&anynul, &status);
  if (status!=0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data", 
	     colnames[0]);
  }

  fits_read_col(fptr, TINT, colnum[1], row, 1, 1, 0, &(ptr->level_final), 
		&anynul, &status);
  if (status!=0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[1]);
  }

  fits_read_col(fptr, TINT, colnum[2], row, 1, 1, 0, &(ptr->level_init), 
		&anynul, &status);
  if (status!=0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[2]);
  }

  fits_read_col(fptr, TSTRING, colnum[3], row, 1, 1, 0, &ionrec_type, 
		&anynul, &status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[3]);
  } else {
    ptr->tr_type = 0;
    if (strncmp(ionrec_type,IONREC_NAME_RR,2)==0) ptr->tr_type=IONREC_TYPE_RR;
    if (strncmp(ionrec_type,IONREC_NAME_DR,2)==0) ptr->tr_type=IONREC_TYPE_DR;
    if (strncmp(ionrec_type,IONREC_NAME_CI,2)==0) ptr->tr_type=IONREC_TYPE_CI;
    if (strncmp(ionrec_type,IONREC_NAME_EA,2)==0) ptr->tr_type=IONREC_TYPE_EA;
    if (ptr->tr_type == 0) {
      fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	       colnames[3]);
    }
  }
  
  fits_read_col(fptr, TINT, colnum[4], row, 1, 1, 0, &(ptr->tr_index), 
		&anynul, &status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[4]);
  }

  fits_read_col(fptr, TINT, colnum[5], row, 1, 1, 0, &(ptr->par_type),
		&anynul, &status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[5]);
  }

  fits_read_col(fptr,TFLOAT,colnum[6], row, 1, 1, 0, &var_float,
		&anynul,&status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[6]);
  } else {
    ptr->min_temp = (double) var_float;
  }

  fits_read_col(fptr,TFLOAT,colnum[7], row, 1, 1, 0, &var_float,
		&anynul,&status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[7]);
  } else {
    ptr->max_temp = (double) var_float;
  }
    
  fits_read_col(fptr, TFLOAT, colnum[8], row, 1, MAX_IONREC, 0,arr_float,
		&anynul, &status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[8]);
  } else {
    for (i=0;i<MAX_IONREC;i++) 
      ptr->Temperature[i] = (double) arr_float[i];
  } 

  fits_read_col(fptr, TFLOAT, colnum[9], row, 1, MAX_IONREC, 0,arr_float,
		&anynul, &status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[9]);
  } else {
    for (i=0;i<MAX_IONREC;i++) 
      ptr->IonRec_Par[i] = (double) arr_float[i];
  } 
  
  fnul_val = -1.0;
  fits_read_col(fptr,TFLOAT,colnum[10], row, 1, 1, &fnul_val, &var_float,
		&anynul,&status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[10]);
  } else { 
    if (anynul == FALSE) {
      line->lambda = (double) var_float;
    } else {
      line->lambda = (double) fnul_val;
    }
  }

  fnul_val = -1.0;
  fits_read_col(fptr,TFLOAT,colnum[11], row, 1, 1, &fnul_val, &var_float,
		&anynul,&status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[11]);
  } else {
    if (anynul == FALSE) line->lambda = (double) var_float;
  }
  
  fnul_val = -1.0;
  fits_read_col(fptr,TFLOAT,colnum[12], row, 1, 1, &fnul_val, &var_float,
		&anynul,&status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[12]);
  } else {
    if (anynul == FALSE) {
      line->lambda_err = (double) var_float;
    } else {
      line->lambda_err = (double) fnul_val;
    }
  }

  fits_read_col(fptr,TFLOAT,colnum[13], row, 1, 1, 0, &var_float,
		&anynul,&status);
  if (status != 0) {
    fitsmess("read_fits_ionrec_data",status,"Error reading %s data",
	     colnames[13]);
  } else {
    line->br_ratio = (double) var_float;
  }
  free(ionrec_type);
}
		
