#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "messages.h"

#define RYDBERG 0.013605804
#define PI 3.1415926535
#define MAX_BHYD 10000

double G_hyd(int, int, double, double);
double B_hyd(int s, int l, int m, double eta);

double sigma_hydrogenic(int Z, int n, int l, double E) {

  /* From Boardman, 1964, ApJS.  Note Typo in l+1 case! */
  /* Compare to Karzas & Latter, 1961, ApJS.  */
  
  /* n, l are the quantum states.  E is the energy of the photon, in keV */

  double rho, eta;
  double chi;  /* Ionization energy, in keV */
  double Eelec; /* Initial electron energy, in Rydberg */
  double coeff=1.09768e-19; /* in cm^2 keV */
  double expterm, prodterm; 
  double commfact1, commfact2, rhoterm;
  double Gterm_m1, Gterm_p1;
  double sigma, sigma_m1, sigma_p1;
  double lp1;

  int iFact, iProd;

  /* Basic terms */

  chi = Z*Z*RYDBERG/(n*n);
  Eelec = (E - chi)/RYDBERG;
  if (Eelec <= 0) return 0.;  /* Set sigma = 0 at and below threshold to 
				 avoid nasty problems with infinities. */
  eta = Z/sqrt(Eelec);
  rho = eta / n ;
  lp1 = l+1.0;

  /* Exponential term */
  expterm = exp(-4*eta*atan(1./rho)) / (1-exp(-2*PI*eta));

  /* Common Factorial */
  commfact1 = 1.;
  commfact2 = 1.;
  if (n+l >= 2*l+2) {
    for (iFact=2*l+3;iFact<=n+l;iFact++) commfact1 *= iFact;
  } else {
    for (iFact=n+l+1; iFact<=2*l+2;iFact++) commfact2 *= iFact;
  }
  for (iFact=2;iFact<2*l+2;iFact++) commfact2 *= iFact;
  for (iFact=2;iFact<n-l;iFact++) commfact2 *= iFact;

  /* Common Product term */
  prodterm = 1.;
  for (iProd=1;iProd < l+1; iProd++) prodterm *= (iProd*iProd + eta*eta);

  /* Rho term */
  rhoterm = pow(rho,2*l+4.)/pow(1+rho*rho,2.*n);

  /* G terms */
  Gterm_p1 = (l+1-n)*G_hyd(l+1, n-l-1, eta, rho) +
    ((l+1+n)/(1+rho*rho))*G_hyd(l+1, n-l, eta, rho);

  if (l!=0) {
    Gterm_m1 = G_hyd(l, n-l-1, eta, rho) - 
      G_hyd(l, n-l+1, eta, rho)/pow(1.+rho*rho,2.);
  } else {
    Gterm_m1 = 0.;
  }

  /* Final answer */
  /* printf("%e %e %e %e %e\n",eta, commfact1/commfact2, prodterm, expterm, 
     rhoterm); */

  sigma_p1 = (coeff/E)*(pow(2.,4*l+6.)/3.)*(lp1*lp1/(2*l+1.))*
    (commfact1/commfact2)*(prodterm/(lp1*lp1+eta*eta))*
    expterm*rhoterm*eta*eta*Gterm_p1*Gterm_p1;
  /* printf("%e %e %e %e\n",coeff, eta, Gterm_p1, sigma_p1); */

  sigma_m1 = 0.0;

  if (l != 0) {
    sigma_m1 = (l*l/(64.*lp1*lp1))*(2*l+1)*(2*l+2)*(2*l+1)*(2*l)*
      (((1+rho*rho)*(1+rho*rho))/(eta*eta*rho*rho))*
      ((lp1*lp1+eta*eta)/(l*l+eta*eta))*
      (Gterm_m1*Gterm_m1/(Gterm_p1*Gterm_p1))*sigma_p1;
  }
  /* printf("%e %e %e %e %e\n",E, Gterm_m1, Gterm_p1, sigma_m1, sigma_p1); */
  sigma = sigma_m1 + sigma_p1;
  return sigma;
}

double G_hyd(int l, int m, double eta, double rho) {

  int s;
  double result=0, B_s;
  
  B_s = B_hyd(2*m, l, m, eta);
  /* printf("%d %d: %e ",l, m, B_s); */
  result = B_s;

  for (s=2*m-1;s>=0;s--) {
    B_s = B_hyd(s, l, m, eta);
    /* printf("%e ",B_s);  */
    result = result*rho + B_s;
  }
  /* printf("\n"); */

  return result;
}

double B_hyd(int s, int l, int m, double eta) {

  int t, iS;
  double d[MAX_BHYD];
  double term1, term2;
  double result;

  if ((s+1)*(s+1) > MAX_BHYD) {
    errmess("B_hyd","Increase MAX_BHYD to at least %d",(s+1)*(s+1));
  }

  d[0*(s+1) + 0] = 1.0;
  for(iS=1;iS<=s;iS++) {
    for(t=0;t<=iS;t++) {
      term1 = 0.;
      if ((iS > 0)&&(t>0)) term1 = d[(iS-1)*(s+1)+t-1];
      term2 = 0.;
      if ((iS > 1)&&((iS-2) >= t)) term2 = d[(iS-2)*(s+1)+t];
      d[iS*(s+1) + t] = - (1./(iS*(iS+2.*l-1.)))*
	((4*(iS-1-m)*term1) + (2*m+2-iS)*(2*m+2*l+1-iS)*term2);
    }
  }

  result = d[s*(s+1)+s];
  for(t=s-1;t>=0;t--) result = result*eta + d[s*(s+1) + t];

  return result;
}

