#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "messages.h"

#define TRUE 1
#define FALSE 0
#define RYDBERG 0.013605804
#define PI 3.1415926535
double G_hyd(int, int, double, double);
double B_hyd(int s, int l, int m, double eta);
char *roman_element(int Z, int J);

/* Static variables */
struct SIGMA_HYD_STORE {
  int Z;
  int n;
  int l;
  int Nbin;
  double E1; /* E value at midpoint; basically, just a quick checkpoint */
  double *sigma;
  struct SIGMA_HYD_STORE *next;
};
struct SIGMA_HYD_STORE *top;

void sigma_hydrogenic_arr(int Z, int n, int l, int Nbin, double bin_edge[], 
			  double sigma[]) { 

  /* From Boardman, 1964, ApJS.  Note Typo in l+1 case! */
  /* Compare to Karzas & Latter, 1961, ApJS.  */
  
  /* n, l are the quantum states.  E is the energy of the photon, in keV */

  double E, rho, eta;
  double chi;  /* Ionization energy, in keV */
  double Eelec; /* Initial electron energy, in Rydberg */
  double coeff=1.09768e-19; /* in cm^2 keV */
  double expterm, prodterm; 
  double commfact1, commfact2, rhoterm;
  double Gterm_m1, Gterm_p1;
  double sigma_m1, sigma_p1;
  double lp1;
  int iFact, iProd, iBin, match=FALSE, done=FALSE;
  struct SIGMA_HYD_STORE *ptr;

  /* First, check to see if we've done this calculation already */
  ptr = top;
  if (ptr != NULL) {
    while ((!done) && (!match)) {
      if ((ptr->Z == Z) && (ptr->n == n) && (ptr->l == l) &&
	  (ptr->Nbin == Nbin) && (ptr->E1 == bin_edge[Nbin/2])) {
	match = TRUE;
	for (iBin=0;iBin<Nbin;iBin++) sigma[iBin] = (ptr->sigma)[iBin];
	return;  /* And we're done! */
      } else {
	ptr=ptr->next;
	if (ptr == NULL) done=TRUE;
      }
    }
  }

  /*
  message("sigma_hydrogenic_arr","Doing calculation for %s n=%d, l=%d",
	  roman_element(Z,Z-1),n,l);
  */

  for (iBin=0;iBin<Nbin;iBin++) {
    E = bin_edge[iBin];
    
    chi = Z*Z*RYDBERG/(n*n);     /* Basic terms */
    Eelec = (E - chi)/RYDBERG;

    if (Eelec <= 0) {    /* Set sigma = 0 at and below threshold to avoid */
      sigma[iBin] = 0.;    /* nasty problems with infinities. */    
    } else {
      eta = Z/sqrt(Eelec);
      rho = eta / n ;
      lp1 = l+1.0;
      
      /* Exponential term */
      expterm = exp(-4*eta*atan(1./rho)) / (1-exp(-2*PI*eta));
      
      /* Common Factorial */
      commfact1 = 1.;
      commfact2 = 1.;
      if (n+l >= 2*l+2) {
	for (iFact=2*l+3;iFact<=n+l;iFact++) commfact1 *= iFact;
      } else {
	for (iFact=n+l+1; iFact<=2*l+2;iFact++) commfact2 *= iFact;
      }
      for (iFact=2;iFact<2*l+2;iFact++) commfact2 *= iFact;
      for (iFact=2;iFact<n-l;iFact++) commfact2 *= iFact;
      
      /* Common Product term */
      prodterm = 1.;
      for (iProd=1;iProd < l+1; iProd++) prodterm *= (iProd*iProd + eta*eta);
      
      /* Rho term */
      rhoterm = pow(rho,2*l+4.)/pow(1+rho*rho,2.*n);
      
      /* G terms */
      Gterm_p1 = (l+1-n)*G_hyd(l+1, n-l-1, eta, rho) +
	((l+1+n)/(1+rho*rho))*G_hyd(l+1, n-l, eta, rho);
      
      if (l!=0) {
	Gterm_m1 = G_hyd(l, n-l-1, eta, rho) - 
	  G_hyd(l, n-l+1, eta, rho)/pow(1.+rho*rho,2.);
      } else {
	Gterm_m1 = 0.;
      }
      
      /* Final answer */

      sigma_p1 = (coeff/E)*(pow(2.,4*l+6.)/3.)*(lp1*lp1/(2*l+1.))*
	(commfact1/commfact2)*(prodterm/(lp1*lp1+eta*eta))*
	expterm*rhoterm*eta*eta*Gterm_p1*Gterm_p1;
      
      sigma_m1 = 0.0;
      
      if (l != 0) {
	sigma_m1 = (l*l/(64.*lp1*lp1))*(2*l+1)*(2*l+2)*(2*l+1)*(2*l)*
	  (((1+rho*rho)*(1+rho*rho))/(eta*eta*rho*rho))*
	  ((lp1*lp1+eta*eta)/(l*l+eta*eta))*
	  (Gterm_m1*Gterm_m1/(Gterm_p1*Gterm_p1))*sigma_p1;
      }
      sigma[iBin] = sigma_m1 + sigma_p1;
    }
  }
  /* Now, store results if they are in fact new. */
  ptr = top;
  if (ptr == NULL) {
    top =(struct SIGMA_HYD_STORE *) malloc(sizeof(struct SIGMA_HYD_STORE));
    ptr = top;
    ptr->next = NULL;
  } else {
    while (ptr->next != NULL) ptr = ptr->next;
    ptr->next=(struct SIGMA_HYD_STORE *)malloc(sizeof(struct SIGMA_HYD_STORE));
    ptr = ptr->next;
    ptr->next = NULL;
  }
  ptr->Z = Z;  
  ptr->n = n;
  ptr->l = l;
  ptr->Nbin = Nbin;
  ptr->E1 = bin_edge[Nbin/2];
  ptr->sigma = (double *) malloc(Nbin*sizeof(double));
  for (iBin=0;iBin<Nbin;iBin++) (ptr->sigma)[iBin] = sigma[iBin];
  
}

void init_sigma_hydrogenic_arr() {

  top = NULL;

}
