#include <stdlib.h>
#include <stdio.h>

#include "messages.h"
#include "apec.h"
#define ERRLIMIT 1.e-3  /* Relative error */

/*! \file solve_rate_eqns.c */

char *roman_element(int Z, int J);
extern int dgesv_(long int *ndim, long int *nrhs, double *AA, long int *lda,
		  long int *INDX, double *BB, long int *ldb, long int *info);
void dgemv_(char *trans, long int *ndimX, long int *ndimY, 
	    double *alpha, double *AAch, 
	    long int *lda, double *B, long int *incx, 
	    double *beta, double *Bch, long int *incy);
void init_rate_matrices(int maxsize);

static double *AA=NULL, *B=NULL, *Bdebug=NULL;
static double *AAch=NULL, *Bch=NULL, *Rch=NULL;
static long int *ipiv=NULL;
static int *SkipLevel=NULL;
static int CurrSize=0;

void init_rate_matrices(maxsize) {
  /* Malloc arrays to their maximum size */

  if (maxsize > CurrSize) {
    if (CurrSize != 0) {
      free(AA); free(B); free(Bdebug); free(AAch); free(Bch); free(Rch);
      free(ipiv); free(SkipLevel);
    }
    CurrSize = maxsize;
    AA     = (double *) malloc_safe(maxsize*maxsize*sizeof(double));
    B      = (double *) malloc_safe(maxsize*sizeof(double));
    Bdebug = (double *) malloc_safe(maxsize*sizeof(double));

    AAch = (double *) malloc_safe(maxsize*maxsize*sizeof(double));
    Bch  = (double *) malloc_safe(maxsize*sizeof(double));
    Rch  = (double *) malloc_safe(maxsize*sizeof(double));

    ipiv = (long int *) malloc_safe(maxsize*sizeof(long int));
    SkipLevel = (int *) malloc_safe(maxsize*sizeof(int));
  }

}  

double *GetAA(void)   { return AA; }
double *GetB(void)    { return B; }
long int *Getipiv(void) { return ipiv; }
double *GetAAch(void) { return AAch; }
double *GetRch(void)  { return Rch; }

/* \brief Using precalculated rates, calculate level population of ion 
 *  \param debug Debugging output files [optional; set to NULL to ignore]
 *  \param trans_rates Structure holding all (de)excitation rates (cm^3/s)
 *  \param level_rates Ionization/Recombination level population rates (cm^3/s)
 *  \param Ne Electron density (in cm^-3)
 *  \param Lev_pop Output level population array (unitless)
 * This routine takes the excitation and de-excitation rates that have 
 * already been calculated, along with the recombination rates, and 
 * calculates the level population.  Does not (as yet) handle excitations
 * due to excitation-autoionization or inner-shell ionization. 
 */


void solve_rate_eqns(struct DEBUG_FILES *debug, 
		     struct TRANS_RATES *trans_rates,
		     struct LEVEL_RATES *level_rates, 
		     double Ne, double *Lev_pop) {

  int nlev;
  long int ndim;
  long int N, nrhs, lda, ldb, info;  /* Variables needed by dgesv */
  int iCol, iRow, oCol, oRow, iDiag, oDiag;
  double ColumnSum, LevSum;

  /* Check variables */
  char trans;
  double alpha, beta, Err;
  long int incx, incy;
  int lu, ll;
  
  nlev = trans_rates->nlev;
  ndim = nlev;

  if (debug != NULL) {
    if (debug->do_levpop && debug->ascii) {
      fprintf(debug->levpop.ascii,
	      "%4d %4d: log10(T), log10(n) (%s): %8g%8g \n",
	      trans_rates->Z,trans_rates->rmJ,
	      roman_element(trans_rates->Z,trans_rates->rmJ-1),
	      log10(trans_rates->Te),log10(Ne));
    }
  }

  /* First step is to discover if any levels are not going to be populated */
  /* because their rates are entirely zero. */

  for (iCol=0;iCol<nlev;iCol++) {
    SkipLevel[iCol] = FALSE;
    ColumnSum = 0.0;
    for (iRow=0;iRow<nlev;iRow++) {
      if (iCol!=iRow) ColumnSum += trans_rates->qexc[iCol*nlev+iRow];
    }
    if (ColumnSum == 0.0) {
      /* This happens if there is a level that has no transitions into */
      /* the level. In this case, the level can be ignored.            */  

      /* message("solve_rate_eqns", 
	 "Level %d of Z=%d, rmJ=%d has no way of being populated.",
	 iCol+1,trans_rates->Z,trans_rates->rmJ);      */

      if (iCol==0) {
	/* This occurs if there is no path to the ground state.     */
	/* This can happen in two ways:                             */ 
	/* Method 1: There are no transitions AT ALL in the matrix, */
	/*           because there is no data for this ion at this  */
	/*           temperature, so all the rates are zero.        */
	/* Method 2: Something is very badly broken.                */
	/*                                                          */
	/* So, the trick is, is this a type 1 or type 2 problem?    */
	/* Note that the following solution isn't the fastest, but  */
	/* since we're about to avoid a matrix inversion in either  */
	/* case, it's not a major problem                           */

	for (ll=1; ll<nlev;ll++) {
	  for (lu=ll; lu<nlev+1;lu++) {
	    if (trans_rates->qexc[(ll-1)*nlev+lu-1] != 0.0) { /* Type 2 */
	      message("solve_rate_eqns",
		   "There is no path to the ground state, but there are");
	      message("solve_rate_eqns",
		   "transitions between excited states %d and %d.",ll,lu);
	      
	      message("solve_rate_eqns",
		   "This indicates something is very wrong with the input");
	      message("solve_rate_eqns", "database for %s.",
		      roman_element(trans_rates->Z,trans_rates->rmJ-1));
	      /* Don't dump core, because this can be due to proton */
	      /* rate data, although it's a "bad thing" in general  */
	    }
	  }
	}
	/* If we made it here, it's a type 1.  Set the level population to*/
	/* be unexcited and move on to the next ion.                      */
	message("solve_rate_eqns","No collisional data for %s at T=%e",
		roman_element(trans_rates->Z,trans_rates->rmJ-1),
		trans_rates->Te);
	Lev_pop[0] = 1.0;
	for (iCol=1;iCol<nlev;iCol++) Lev_pop[iCol] = 0.0;
	return;
      }
      SkipLevel[iCol] = TRUE;
      ndim--;
    }
  }

  /* Now we know about all the levels that will be skipped.  Now fill */
  /* the matrix we'll invert. */
  
  /* Set the population conservation equation, sum(p_i) = 1.0 */

  for (iCol=0;iCol<ndim;iCol++) AA[iCol*ndim + 0] = 1.0;
  B[0] = 1.0;

  /* Now the transition matrix.  We ignore the ground state balancing */
  /* equation, to remove the degeneracy otherwise inherent in having  */
  /* individual balance equations for n levels plus the population    */
  /* conservation equation.  The offsets, oCol and oRow, are there to */
  /* handle missing levels.  This does everything except the diagonal */
  /* elements, which are handled next */

  oCol = 0;
  for (iCol=0;iCol<nlev;iCol++) {
    if (SkipLevel[iCol]) {
      oCol++;
    } else {
      oRow = 0;
      for (iRow=1;iRow<nlev;iRow++) {
	if (SkipLevel[iRow]) {
	  oRow++;
	} else {
	  if (iCol-oCol != iRow-oRow) {
	    AA[(iCol-oCol)*ndim + iRow-oRow]=trans_rates->qexc[iCol*nlev+iRow];
	  }
	}
      }
    }
  }

  /* Now to calculate the diagonal elements */
  oDiag = 0;
  for (iDiag=1;iDiag<nlev;iDiag++) {
    if (SkipLevel[iDiag]) {
      oDiag++;
    } else {
      AA[(iDiag-oDiag)*ndim + iDiag-oDiag] = 0.;
      for (iRow=0;iRow<nlev;iRow++) {
	if (!SkipLevel[iRow]) {
	  AA[(iDiag-oDiag)*ndim + iDiag-oDiag] += 
	    -trans_rates->qexc[iDiag*nlev + iRow];
	}
      }
    }
  }

  /* Now, we need the recombination coefficients for the RHS. */
  oRow = 0;
  for (iRow=1;iRow<nlev;iRow++) {
    if (SkipLevel[iRow]) {
      oRow++;
    } else {
      B[iRow-oRow] = level_rates->qrecomb[iRow]+level_rates->qsatel[iRow];
      B[iRow-oRow] *= -Ne*(level_rates->ionc);
      if (debug != NULL) {
	if (debug->do_levpop && debug->ascii) {
	  fprintf(debug->levpop.ascii,
		  "%4d %4d: RR = %e, DR = %e, B[%d] = %e\n", 
		  trans_rates->Z, trans_rates->rmJ, 
		  level_rates->qrecomb[iRow], level_rates->qsatel[iRow], 
		  iRow, B[iRow]);	
	}
      }
    }
  }

  /* Now set some variables needed by the matrix inversion routine. */

  nrhs = 1;
  lda = ndim;
  ldb = ndim;
  N   = ndim;
  
  /* Set up the backup variables for checking. */
  for (iCol=0;iCol<ndim;iCol++) {
    for (iRow=0;iRow<ndim;iRow++) {
      AAch[iCol*ndim + iRow] = AA[iCol*ndim + iRow];
    }
    Bch[iCol] = B[iCol];
    Rch[iCol] = 0.0;
  }

  for (iRow = 0; iRow < ndim; iRow++) Bdebug[iRow] = B[iRow];

  /* We're all set.  Now just do the inversion. */

  (void) dgesv_(&N, &nrhs, GetAA(), &lda, Getipiv(), GetB(), &ldb, &info);
  
  if (info != 0) {
    if (info < 0) {
      message("solve_rate_eqns",
	      "Error solving matrix: %d-th level bad for Z=%d, rmJ=%d.",
	      -info,trans_rates->Z,trans_rates->rmJ);
    } else {
      message("solve_rate_eqns",
	      "Error solving matrix: U(%d,%d) = 0 for Z=%d, rmJ=%d",
	      info, info, trans_rates->Z,trans_rates->rmJ);
    }
    /*
      for (iCol=0;iCol<ndim;iCol++) {
      printf("%d %e \n",iCol, Bch[iCol]);
      for (iRow=0;iRow<ndim;iRow++) {
      printf("%d %d %e \n",iCol, iRow, AAch[iCol*ndim + iRow]);
      }
      }
    */
  }
  /* Now check the result */
  trans = 'N';
  lda = ndim;
  alpha = 1.0;
  beta =  0.0;
  incx = 1;
  incy = 1;

  dgemv_(&trans,&ndim, &ndim, &alpha, GetAAch(), &lda, GetB(), 
	 &incx, &beta, GetRch(), &incy);

  /* Check the matrix results */
  LevSum = 0.0;

  for (iRow=0;iRow<ndim;iRow++) {
    LevSum += B[(ndim-1)-iRow];
    if (Rch[iRow] == Bch[iRow]) {
      Err = 0.0;
    } else {
      Err = fabs((Rch[iRow] - Bch[iRow])/(Rch[iRow] + Bch[iRow]));
      if (Bch[iRow] == 0.0) Err = fabs(Rch[iRow]);
    }
    if (Err > ERRLIMIT) {
      message("solve_rate_eqns","Pop[%3d] = %e, Rch=%e, Bch=%e, Error=%e",
	      iRow, B[iRow], Rch[iRow], Bch[iRow], Err);
    }
  }
  if (fabs(LevSum - 1.) > ERRLIMIT) {
    message("solve_rate_eqns","Error: Total level population is %e.",LevSum);
  }

  if (debug != NULL) {
    if (debug->do_levpop && debug->ascii) {
      fprintf(debug->levpop.ascii,"%4d %4d: %8g%8g ",trans_rates->Z,
	      trans_rates->rmJ, log10(trans_rates->Te),log10(Ne));
    }
  }

  /* Set the Lev_pop array equal to solution vector B.  */
  oRow = 0;
  for (iRow=0; iRow < nlev; iRow++) {
    if (SkipLevel[iRow]) {
      Lev_pop[iRow] = 0.0;
      oRow++;
    } else {
      if (debug != NULL) {
	if (debug->do_levpop && debug->ascii) {
	  fprintf(debug->levpop.ascii,"%8g ",B[iRow-oRow]);
	}
	if (debug->do_levpop && debug->fits) {
	  if (iRow == 0) {
	    fd_write_row(&debug->levpop.fits,
			 trans_rates->Z,
			 trans_rates->rmJ,
			 roman_element(trans_rates->Z,trans_rates->rmJ-1),
			 log10(trans_rates->Te),
			 log10(Ne),
			 iRow, 
			 0.0, 
			 0.0, 
			 0.0, 
			 B[iRow-oRow]);
	  } else {
	    fd_write_row(&debug->levpop.fits,
			 trans_rates->Z,
			 trans_rates->rmJ,
			 roman_element(trans_rates->Z,trans_rates->rmJ-1),
			 log10(trans_rates->Te),
			 log10(Ne),
			 iRow, 
			 level_rates->qrecomb[iRow], 
			 level_rates->qsatel[iRow], 
			 Bdebug[iRow], 
			 B[iRow-oRow]);
	  }
	}
      }
      
      if (B[iRow] < 0.0) {
	message("solve_rate_eqns","Err:Level pop for Z=%d, rmJ=%d, n=%d is %e",
		trans_rates->Z, trans_rates->rmJ, iRow, B[iRow]);
	message("solve_rate_eqns","Set to zero.");
	B[iRow] = 0.0;
      }

      Lev_pop[iRow] = B[iRow - oRow];
    }
  }

  if (debug != NULL) {
    if (debug->do_levpop && debug->ascii) {
      fprintf(debug->levpop.ascii,"\n");
      fflush(debug->levpop.ascii);
    }
  }

}
