PRO write_elev, filename, Z, rmJ, data, COMMENTS=comments, ENERGY_REF=eref,$
                PHOT_REF=pref
;
;  Writes a Elev FITS file using the data in the structure data, for
;  element Z and ion stage rmJ.  The code first checks to see if
;  filename exists.  If it doesn't, it creates filename and then
;  writes the primary header.  If it does, the code checks to make
;  sure this particular ion stage hasn't already been written to the
;  file.  Assuming all is OK, we are positioned at the end of the file.
;  Then the header is created, and the data is written to the file.
;

;
;  Check to see if file exists
;
result = findfile(filename,count=count)
if (count eq 0) then begin      ; must create file and primary HDU
    fxhmake, primary_header, /extend, /initialize, /date
    fxaddpar, primary_header, 'E_LEVEL','Atomic Energy levels'
    fxaddpar, primary_header, 'FILENAME','IDL_routine','Source of data'
    fxaddpar, primary_header, 'ORIGIN','CXC','Chandra X-ray Center'
    fxaddpar, primary_header, 'HDUCLASS','ATOMIC','Atomic data'
    fxaddpar, primary_header, 'HDUCLAS1','E_LEVEL','Atomic Energy levels'
    version = '1.1.0'
    fxaddpar, primary_header, 'HDUVERS1', version, 'Version of datafile'
    fxwrite, filename, primary_header
endif

if (count gt 1) then begin
    print,'More than 1 file exists with this name: ',result
    stop
endif

if (count eq 1) then begin
    Ext = match_header(filename, Z, rmJ, 'E_LEVEL')
    if (Ext gt 0) then begin
        print,'Extension ',Ext,' already has data for Z, rmJ = ',Z,rmJ,$
          ' in file ',filename
        return
    endif
endif
;
;  Now we're all set, ready to create the new header.
;
integer = 1L
realvar = 1.0
array = fltarr(20)
strings = '01234567890123456789
bigstrings = '0123456789012345678901234567890123456789'
nrows = n_elements(data)
extname = ionname(Z,rmJ)
fxbhmake, header, nrows, extname, 'Ion Name'
fxaddpar, header, 'HDUCLASS', 'ATOMIC', 'Atomic Data'
fxaddpar, header, 'HDUCLAS1','E_LEVEL', 'Emission line tables'
fxaddpar, header, 'ELEMENT', Z, 'Number of protons in element'
fxaddpar, header, 'ION_STAT',rmJ-1,'Ion state (0 = neutral)'
fxaddpar, header, 'ION_NAME',extname, 'Ion Name'
fxaddpar, header, 'N_LEVELS',nrows,'Number of energy levels'
version = '1.1.0'
fxaddpar, header, 'HDUVERS1', version, 'Version of datafile'
fxbaddcol, 1, header, bigstrings, 'Elec_Config'
fxbaddcol, 2, header, realvar, 'Energy',TUNIT='eV'
fxbaddcol, 3, header, realvar, 'E_Error',TUNIT='eV'
fxbaddcol, 4, header, integer, 'N_quan'
fxbaddcol, 5, header, integer, 'L_quan'
fxbaddcol, 6, header, realvar, 'S_quan'
fxbaddcol, 7, header, integer, 'Lev_Deg'
fxbaddcol, 8, header, integer, 'Phot_Type'
fxbaddcol, 9, header, array  , 'Phot_Par'
fxbaddcol,10, header, strings, 'Energy_Ref'
fxbaddcol,11, header, strings, 'Phot_Ref'
;
; Now add any additional comments
;
    for i=0,n_elements(comments)-1 do begin
        if (strpos(comments(i),'COMMENT') eq 0) then begin
            commstr = strmid(comments(i),8)
            add_reference, header, commstr, /header
        endif else begin
            add_reference, header, comments(i), /header
        endelse
    endfor
;
;  Now open the file and write the new header
;
fxbcreate, unit, filename, header
;
;  Write the data
;
for i=0,nrows-1 do begin
    fxbwrite, unit, data(i).Elec_Config, 1, i+1
    fxbwrite, unit, data(i).Energy,      2, i+1
    fxbwrite, unit, data(i).E_Error,     3, i+1
    fxbwrite, unit, data(i).N_quan,      4, i+1
    fxbwrite, unit, data(i).L_quan,      5, i+1
    fxbwrite, unit, data(i).S_quan,      6, i+1
    fxbwrite, unit, data(i).Lev_Deg,     7, i+1
    fxbwrite, unit, data(i).Phot_Type,   8, i+1
    array = data(i).Phot_Par
    fxbwrite, unit, array,               9, i+1
    if (n_elements(eref) eq 0) then begin
        fxbwrite, unit, data(i).Energy_Ref,  10, i+1
    endif else begin
        fxbwrite, unit, eref,                10, i+1
    endelse
    if (n_elements(pref) eq 0) then begin
        fxbwrite, unit, data(i).Phot_Ref,    11, i+1
    endif else begin
        fxbwrite, unit, pref,                11, i+1
    endelse
endfor
;
; Close the file
;
fxbfinish, unit

return
end
