%
% This file reads in the model definition from either the 
% standard model (provided by APEC in AM->standard_model_file)
% or a user specific model (stored in a file names _apec_model_).
%
% Changes in the model require exiting and starting sherpa and
% the apec model software. 
%
% Ronnie Hoogerwerf, SAO, 06/24/2002
%
uses("SL-str","SL-ll");

define read_apec_model_definitions();
private define element_format();

define read_apec_model_definitions() {

   variable fp, line, words, nelem = 0, elems = NULL, i;
   variable tmp;

   % Determine which file to read.
   i = 0;
   while (andelse {i < 3} {IO->file_exists(AM->model_files[i]) == 0}) i++;

   vmessage("==> Reading apec model definition from file \"%s\"",
	    AM->model_files[i]);
   fp = fopen(IO->replace_envs(AM->model_files[i]),"r");

   if (fp == NULL) {
      verror("Error opening \"%s\", \"%s\" or \"%s\" file",
	     AM->model_file,AM->home_model_file,
	     AM->standard_model_file);
   }


   % Loop through file.
   while (-1 != fgets(&line,fp)) {
      % Skip comments
      if (line[0] == '%') continue;

      % Split model definition line and remove spaces
      words = Str->split_and_trim(line,'=');

      % Analyze model definition.
      switch (words[0]) 
      { case "ntemp" : 
	AM->N_Temp_comp_max = integer(words[1]);
	AM->N_Temp_comp = AM->N_Temp_comp_max;}
      { case "element" :
	% First convert the element name to our convention: "Xx"
        words[1] = element_format(words[1]);
	elems = LL->append(words[1],elems);
	nelem++;}
   }

   % close model definition file
   () = fclose(fp);

   AM->nelem = nelem;
   AM->npar = nelem + (2*AM->N_Temp_comp_max);

   elems = LL->ll_to_array(elems);

   AM->ElementsZ = Integer_Type[nelem];
   AM->Abundances = Double_Type[nelem];
   AM->abunds = String_Type[nelem];
   for (i = 0; i < nelem; i++) {
      
      AM->ElementsZ[i] = AM->Elem_to_Z[strlow(elems[i])];
      AM->Abundances[i] = 0.0;
      AM->abunds[i] = "abund" + elems[i]; 
   }

   tmp = "apec model: call as\n";
   tmp = tmp + "apec(";
   for (i = 1; i <= AM->N_Temp_comp_max; i++) tmp = tmp + "T" + string(i) + ", ";
   tmp = tmp + "\n     ";
   for (i = 1; i <= AM->N_Temp_comp_max; i++) tmp = tmp + "norm" + string(i) + ", ";
   tmp = tmp + "\n     ";
   for (i = 0; i < nelem; i++) {
      tmp = tmp + AM->abunds[i] + ", ";
      if ((0 == (i+1) mod 5)) tmp = tmp + "\n     ";
   }
   if (1 == (i+1) mod 5) tmp = tmp + "\n     ";
   tmp = tmp + "binlo, binhi) where\n";
   tmp = tmp + " T<i>      = Temperature of component <i> (in keV)\n";
   tmp = tmp + " norm<i>   = emission measure / 4 pi D^2\n";
   tmp = tmp + " abund<El> = abundance of element <El> (in solar units)\n";
   tmp = tmp + " binlo and binhi are in Angstrom\n";

   AM->apec_usage_message = tmp;

}



private define element_format (word) {

   variable char1, char2;

   if (strlen(word) == 1) {
      if ( word[0] >= 'a') {
	 word = char(word[0] - 32);
      } 
   } else if (strlen(word) == 2) {
      char1 = word[0]; char2 = word[1];
      if (char1 >= 'a') {
	 char1 = char(char1 - 32);
      } else {
	 char1 = char(char1);
      }
      if ( char2 < 'a') {
	 char2 = char(char2 + 32);
      } else {
	 char2 = char(char2);
      }
      word = char1 + char2;
   }

   return word;

}

define copy_apec_model() {

   variable dest, file;
   variable usage = "Usage: copy_apec_model([destination])";
   
   switch (_NARGS)
   { case 0 : dest = "."; }
   { case 1 : dest = (); }
   { message(usage); _pop_n(_NARGS); return; }

   if (dest[strlen(dest)-1] == '/') dest = dest[[:strlen(dest)-2]];

   file = dest + "/" + ".apec_model";
   file = IO->replace_envs(file);

   () = IO->file_copy(IO->replace_envs(AM->standard_model_file),
		      file);

}
