%
% This file sets up a namespace for the Apec Models (AM) and sets 
% a couple of variables which will be used in the other routines
% present in the AM package.
%
% Ronnie Hoogerwerf, SAO, 04/09/2002.
%

%
% Some usefull (and necessary) type definitions
%
variable am_types;

if (__is_initialized(&am_types) == 0) {
   typedef struct { continuum, datarange } Continuum_Type;
   typedef struct { broaden, num, bin_min, bin_max, dbin, nbin, fwhm, type, unit, func, kernel } Broaden_Type;
   typedef struct { dataset, temp, dens, norm, abund, Z, ion} FixLine_Type;
   typedef struct { dataset, Z, ion, ul, ll} RemoveLine_Type;
   am_types = 1;
}

implements("AM");	% Sets up name space for Apec Model variables
%
variable am_runonce;
variable am_aped_loaded;
%
% Check if am_runonce is initialized. If not set up a set of variables
% that will be used in the am code.
%
if (__is_initialized(&am_runonce) == 0) {
   
   % will contain version number of this apec model code
   variable AM_version;
   % are we in verbose mode (verbose = 1) or in quiet mode (verbose = 0)
   variable verbose = 0;
   % Temporary variable until get_mdl_models() is fixed
   variable getmdlmodels = NULL;
   % Convert Kelvins into keVs and vise versa
   variable K_to_keV = 8.617342e-8;
   % maximum number of datasets
   variable max_datasets = 100; 

   % 
   % Parameters used to determine the apec model specifics,
   % i.e., the number of temp. components and which elements
   % are to be included in the model. 
   % Default is 1 T component and 13 elements.
   %
   variable npar = 17;                       % # of model parameters = 2*Ntemp + nelem + ContinuumFile + FitMode
   variable nelem = 13;                      % # of elements
   variable apec_usage_message;              % error message for incorrectly calling apec function
   variable abunds;                          % array containing the parameter names of the abundances
   variable model_file = ".apec_model";      % user input file for model specification in cwd
   variable home_model_file = "$HOME/.apec_model";     
                                             % user input file for model specification in home directory
   variable standard_model_file = "$APEC_SLANG_DIR/.apec_model"; 
                                             % global input file for model specification
   variable model_files = [model_file,home_model_file,standard_model_file];

   % Default settings for number of T components. Note that
   % any default model only have 1 active T component, even if
   % the maximum is larger than 1, when the model is first created
   % using source # = apec[name].
   variable N_Temp_comp = 1;           % Number of temperature components.
   variable N_Temp_comp_max = 1;       % Maximum number of temperatue comp.

   %
   % Some variables to store and manage the calculation of the
   % continuum. Storing the (fixed) continuum speeds up the 
   % calculation by a factor of 50-100 and is especially noticable
   % for large wavelength ranges/
   %
   variable continuum = NULL;
   variable ncont = 0;
   variable continua = Continuum_Type[max_datasets];

   %
   % Variable to store broadening information such as fwhm,
   % kernel type, calculated kernel etc., etc.
   %
   variable broaden = Broaden_Type[max_datasets];

   %
   % Variable containing information of spectral lines that are
   % fixed at a certain temp, dens, norm, and abundance. The
   % fixing can be done for individual elements or ions.
   %
   variable fix = NULL;
   variable remove = NULL;

   %
   % Different modes for the apec model
   variable Fit_Continuum_only = 1;        % Fit Continuum only.
   variable Fit_Lines_only = 2;            % Fit Lines only.
   variable Fit_Continuum_and_lines = 3;   % Fit Continuum and Lines.
   variable Continuum_only = 4;            % Fit nothing, use this to fix continuum.
   variable Fit = Fit_Continuum_and_lines; % Default fitting of lines + cont.

   %
   % Associative array linking element name (as a string) to the
   % atomic number (Z as an integer).
   variable Elem_to_Z = Assoc_Type[Integer_Type];
   Elem_to_Z["h"]  =  1;               % Assoc array linking Element name to
   Elem_to_Z["he"] =  2;               % atomic number
   Elem_to_Z["li"] =  3;   Elem_to_Z["be"] =  4;
   Elem_to_Z["b"]  =  5;   Elem_to_Z["c"]  =  6;
   Elem_to_Z["n"]  =  7;   Elem_to_Z["o"]  =  8;
   Elem_to_Z["f"]  =  9;   Elem_to_Z["ne"] = 10;
   Elem_to_Z["na"] = 11;   Elem_to_Z["mg"] = 12;
   Elem_to_Z["al"] = 13;   Elem_to_Z["si"] = 14;
   Elem_to_Z["p"]  = 15;   Elem_to_Z["s"]  = 16;
   Elem_to_Z["cl"] = 17;   Elem_to_Z["ar"] = 18;
   Elem_to_Z["k"]  = 19;   Elem_to_Z["ca"] = 20;
   Elem_to_Z["sc"] = 21;   Elem_to_Z["ti"] = 22;
   Elem_to_Z["v"]  = 23;   Elem_to_Z["cr"] = 24;
   Elem_to_Z["mn"] = 25;   Elem_to_Z["fe"] = 26;
   Elem_to_Z["co"] = 27;   Elem_to_Z["ni"] = 28;
   Elem_to_Z["cu"] = 29;   Elem_to_Z["zn"] = 30;

   % 
   % Default model elements.
   variable Nelem = 13;
   variable Elements  = ["C","N","O","F","Ne","Mg","Al","Si","S","Ar","Ca","Fe","Ni"];
   variable ElementsZ = [  6,  7,  8,  9, 10, 12, 13, 14, 16, 18, 20, 26, 28];
            % Array used as input to ISIS to define elements in spectrum which
            % have abundances different from metal_abund.
   variable Abundances =[0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0];
            % Array specifying the abundance values different from metal_abund
            % for the correspondig element in Elements.

   %
   % Associative array that links the string representation of
   % the ionozation state to the numerical one.
   variable Roman_to_I = Assoc_Type[Integer_Type];
   Roman_to_I["i"]      =  1;   Roman_to_I["ii"]     =  2;
   Roman_to_I["iii"]    =  3;   Roman_to_I["iv"]     =  4;
   Roman_to_I["v"]      =  5;   Roman_to_I["vi"]     =  6;
   Roman_to_I["vii"]    =  7;   Roman_to_I["viii"]   =  8;
   Roman_to_I["ix"]     =  9;   Roman_to_I["x"]      = 10;
   Roman_to_I["xi"]     = 11;   Roman_to_I["xii"]    = 12;
   Roman_to_I["xiii"]   = 13;   Roman_to_I["xiv"]    = 14;
   Roman_to_I["xv"]     = 15;   Roman_to_I["xvi"]    = 16;
   Roman_to_I["xvii"]   = 17;   Roman_to_I["xviii"]  = 18;
   Roman_to_I["xix"]    = 19;   Roman_to_I["xx"]     = 20;
   Roman_to_I["xxi"]    = 21;   Roman_to_I["xxii"]   = 22;
   Roman_to_I["xxiii"]  = 23;   Roman_to_I["xxiv"]   = 24;
   Roman_to_I["xxv"]    = 25;   Roman_to_I["xxvi"]   = 26;
   Roman_to_I["xxvii"]  = 27;   Roman_to_I["xxviii"] = 28;
   Roman_to_I["xxix"]   = 29;   Roman_to_I["xxx"]    = 30;

   %
   % Set of variables to prevent typing strings on the command
   % line
   public variable _h_  =  1; public variable _H_  =  1;
   public variable _he_ =  2; public variable _He_ =  2;
   public variable _li_ =  3; public variable _Li_ =  3;
   public variable _be_ =  4; public variable _Be_ =  4;
   public variable _b_  =  5; public variable _B_  =  5;
   public variable _c_  =  6; public variable _C_  =  6;
   public variable _n_  =  7; public variable _N_  =  7;
   public variable _o_  =  8; public variable _O_  =  8;
   public variable _f_  =  9; public variable _F_  =  9;
   public variable _ne_ = 10; public variable _Ne_ = 10;
   public variable _na_ = 11; public variable _Na_ = 11;
   public variable _mg_ = 12; public variable _Mg_ = 12;
   public variable _al_ = 13; public variable _Al_ = 13;
   public variable _si_ = 14; public variable _Si_ = 14;
   public variable _p_  = 15; public variable _P_  = 15;
   public variable _s_  = 16; public variable _S_  = 16;
   public variable _cl_ = 17; public variable _Cl_ = 17;
   public variable _ar_ = 18; public variable _Ar_ = 18;
   public variable _k_  = 19; public variable _K_  = 19;
   public variable _ca_ = 20; public variable _Ca_ = 20;
   public variable _sc_ = 21; public variable _Sc_ = 21;
   public variable _ti_ = 22; public variable _Ti_ = 22;
   public variable _v_  = 23; public variable _V_  = 23;
   public variable _cr_ = 24; public variable _Cr_ = 24;
   public variable _mn_ = 25; public variable _Mn_ = 25;
   public variable _fe_ = 26; public variable _Fe_ = 26;
   public variable _co_ = 27; public variable _Co_ = 27;
   public variable _ni_ = 28; public variable _Ni_ = 28;
   public variable _cu_ = 29; public variable _Cu_ = 29;
   public variable _zn_ = 30; public variable _Zn_ = 30;
   public variable _i_      =  1; public variable _I_      =  1;
   public variable _ii_     =  2; public variable _II_     =  2;
   public variable _iii_    =  3; public variable _III_    =  3;
   public variable _iv_     =  4; public variable _IV_     =  4;
   public variable _v_      =  5; public variable _V_      =  5;
   public variable _vi_     =  6; public variable _VI_     =  6;
   public variable _vii_    =  7; public variable _VII_    =  7;
   public variable _viii_   =  8; public variable _VIII_   =  8;
   public variable _ix_     =  9; public variable _IX_     =  9;
   public variable _x_      = 10; public variable _X_      = 10;
   public variable _xi_     = 11; public variable _XI_     = 11;
   public variable _xii_    = 12; public variable _XII_    = 12;
   public variable _xiii_   = 13; public variable _XIII_   = 13;
   public variable _xiv_    = 14; public variable _XIV_    = 14;
   public variable _xv_     = 15; public variable _XV_     = 15;
   public variable _xvi_    = 16; public variable _XVI_    = 16;
   public variable _xvii_   = 17; public variable _XVII_   = 17;
   public variable _xviii_  = 18; public variable _XVIII_  = 18;
   public variable _xix_    = 19; public variable _XIX_    = 19;
   public variable _xx_     = 20; public variable _XX_     = 20;
   public variable _xxi_    = 21; public variable _XXI_    = 21;
   public variable _xxii_   = 22; public variable _XXII_   = 22;
   public variable _xxiii_  = 23; public variable _XXIII_  = 23;
   public variable _xxiv_   = 24; public variable _XXIV_   = 24;
   public variable _xxv_    = 25; public variable _XXV_    = 25;
   public variable _xxvi_   = 26; public variable _XXVI_   = 26;
   public variable _xxvii_  = 27; public variable _XXVII_  = 27;
   public variable _xxviii_ = 28; public variable _XXVIII_ = 28;
   public variable _xxix_   = 29; public variable _XXIX_   = 29;
   public variable _xxx_    = 30; public variable _XXX_    = 30;

   %
   % Model default, minimum, and maximum values
   %
   variable Tdef = 1.0, Tmin = 0.01, Tmax =  68.0, Ttf = 1;
   variable Ndef = 1.0, Nmin = 0.0,  Nmax =  1e10, Ntf = 1;
   variable Adef = 1.0, Amin = 0.0,  Amax = 100.0, Atf = 1;

   
   %
   % Load ISIS 
   %
   message("==> Importing ISIS");
   import("isis","isis");  % put all of isis in its own namespace!

   %
   % Load Guide 
   %
   message("==> Importing Guide");
   import("guide");

   %
   % One would like to import the APED database and APEC models now, 
   % However, this is inpossible to do at this point in this file. We
   % have moved the following two commands to the end of this file. This
   % quick fix seems to work.
   %
   % message("==> Importing APED models");
   % plasma(aped);

   variable am_runonce = 1;           % Variable to see if initialization has
                                      % occured.
}

%
% load aped/apec database of spectral lines and continuum emission
%
if (__is_initialized(&am_aped_loaded) == 0) {
   message("==> Importing APED");
   isis->plasma(aped);
   variable am_aped_loaded = 1;
}

%
% Now compile and load all slang files belonging to the AM package
%
if (NULL == getenv("APEC_SLANG_DIR")) {
   message("The Apec Model package requires the environmental variable");
   message("$APEC_SLANG_DIR to be defined (check with 'echo $APEC_SLANG_DIR')");
   message("Issue the following command \'setenv APEC_SLANG_DIR \"/data/aped/ronnieh/slang/apec\"\'");
   message("or include it in your .cshrc file\n");
} else {
   () = evalfile("SL-uses.sl");      % header facility
   () = evalfile("SL-include.sl");   % usefull routines
   () = evalfile("MATH-include.sl");

   () = evalfile("AM-utilities.sl"); % usefull routines for AM

   () = evalfile("expand_env.sl");   % Borrowed from Randall Smith: allows 
                                     % use of environmental variables in file
                                     % names

   () = evalfile("AM-definitions.sl");
   () = evalfile("AM-fix.sl");
   () = evalfile("AM-remove-lines.sl");

   () = evalfile("AM-set-get-show.sl"); % Change, retreive and view 
                                        % AM-variables

   () = evalfile("AM-elements.sl"); % Handles manipulation of element 
                                    % abundances and in/exclusion of elements
                                    % in the model.

   () = evalfile("AM-model.sl"); % Defines calculation method for apec model
                                 % and provides the definitions for model used
                                 % by Sherpa so that the apec model becomes an
                                 % integral part of Sherpa.

   () = evalfile("AM-broaden.sl");
   () = evalfile("AM-save-load-model.sl");

}


if (NULL != getenv("APEC_SLANG_DIR")) {
   AM->AM_version = IO->get_version_number();
   % Report success
   vmessage("\n\n Apec Models version %s initialized.\n\n",AM->AM_version);
       
   % define apec model.
   read_apec_model_definitions(); 
   AM_make_model();

   %message("\n\n*** IMPORTANT ***\n");
   %message("    This APEC model contains a temporary fix for a bug in");
   %message("    CIAO 3.0.1. If you are running CIAO 3.0.2 AND you see");
   %message("    this message please notify ronnieh@cfa.harvard.edu.\n");
   %message("*** IMPORTANT ***\n\n");

}
