%
%
%  Routines to make reading of simple ascii data tables easy. 
%  The routine "read_block" takes an ascii file and reads it,
%  skipping comments and empty lines, and returns a 2D array
%  containing the information in a type defined by the largest
%  common denominator. The file also provides routines to extract
%  individual columns or rows from the data.
%
%  R. Hoogerwerf, SAO, 04/23/2002
%
%

uses("SL-ll","SL-io");
use_namespace("IO");


%----

define file_copy(file1,file2) {

   variable fp1, fp2, buf, stat;
   variable usage = "Usage: IO->file_copy(filename1,filename2);\n       where filename1 and filename2 are of type string";

   % check if both file1 and file2 are of type string
   !if (typeof(file1) == String_Type and 
	typeof(file2) == String_Type) {
      message(usage);
      return NULL;
   }

   % expand any possible environmental variables
   file1 = IO->replace_envs(file1);
   file2 = IO->replace_envs(file2);

   % open original file and check if file exists
   fp1 = open(file1,O_RDONLY);
   if (fp1 == NULL) {
      verror("IO->file_copy: Unable to open file %s",file1);
   }

   % open destination file and check if file opening is successful
   fp2 = open(file2,O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
   if (fp2 == NULL) {
      verror("IO->file_copy: Unable to open file %s",file2);
   }

   stat = read(fp1,&buf,8);
   while (stat != -1 and stat != 0) {
      () = write(fp2,buf);
      stat = read(fp1,&buf,8);
   }


   () = close(fp1);
   () = close(fp2);

   % must be a success
   return 1;

}


%----

define get_version_number() {

   variable filename, version;
   
   switch (_NARGS)
   { case 0 : filename = "$APEC_SLANG_DIR/VERSION"; }
   { case 1 : filename = (); }
   { message("Usage: Utils->get_version_number([filename]);");
     _pop_n(_NARGS);
     return; }
   
   filename = IO->replace_envs(filename);

   !if (IO->file_exists(filename)) {
     vmessage("Utils->get_version_number: File \"%s\" does not exist.",filename);
     return;
   }

   variable fp = fopen(filename,"r"), line;

   if (NULL == fp) {
      vmessage("Utils->get_version_number: Error opening file \"%s\".",filename);
      return;
   }

   while (-1 != fgets(&line, fp)) {
      % ignore comments
      line = Str->remove_excess_whitespace(line);
      % ignore comments and empty lines
      if ((line[0] == '%') or (strlen(line) == 0)) {
	 continue;
      }

      () = sscanf(line,"%s",&version);
      break;
   }
   
   () = fclose(fp);

   version = Str->remove_excess_whitespace(version);
   version = typecast(version,String_Type);
   return version;

}

%----

define file_exists() {

   variable usage = "Usage: IO->file_exists(filename [,dirname])";
   variable filename, dirname = NULL;
   variable slash, ls;

   switch(_NARGS)
   { case 1 : filename = (); dirname = "."; }
   { case 2 : (filename, dirname) = (); }
   { message(usage); return NULL; }
    
   %
   % Get rid of any environmental variables
   %
   filename = replace_envs(filename);
   dirname = replace_envs(dirname);

   slash = Str->rposition('/',filename);

   if (slash != NULL) {
      if (dirname == ".") {
	 % extract dirname
	 dirname = filename[[0:slash-1]];
      } else {
	 dirname = dirname + filename[[0:slash-1]];
      }
   } else {
      slash = -1;
   }
   filename = filename[[slash+1:]];
   
   ls = listdir(dirname);
   if (Utils->member(filename,ls) != NULL) {
      return 1;
   } else {
      return 0;
   }

}

%----

define replace_envs() {

   variable str, penv, env, slash;

   switch(_NARGS)
   { case 1 : str = (); penv = 0;}
   { error("incorrect number of arguments to IO->replace_envs"); }

   if (is_defined("expand_env") == 0) () = evalfile("expand_env.sl");

   if (penv == 0) {
      penv = Str->position('\$',str);
   } else {
      penv = Str->position('\$',str,penv+1);
   }

   if (penv != NULL) {

      slash = Str->position('/',str,penv);
      if (NULL == slash) slash = Str->position('\$',str,penv+1);
      if (NULL == slash) {
	 env = str[[penv:]];
	 if (penv == 0) {
	    str = expand_env(env);
	 } else {
	    str = str[[0:penv-1]] + expand_env(env);
	 }
      } else {
	 env = str[[penv:slash-1]];
	 if (penv == 0) {
	    str = expand_env(env) + str[[slash:]];
	 } else {
	    str = str[[0:penv-1]] + expand_env(env) + str[[slash:]];
	 }
      }
   }

   if (Str->position('\$',str)) return IO->replace_envs(str);

   return str;
   
}


%----

define get_user_reply() {

   variable usage = "IO->get_user_reply(answer, message [, return_type])";
   variable ans, msg, type = NULL;
   
   if (Utils->check_nargs(_NARGS,2,3) == 0) {
      message(usage);
      return;
   }
   
   if (_NARGS == 2) {
      (ans, msg) = ();
   } else {
      (ans, msg, type) = ();
   }
   
   () = printf("%s", msg);
   if (-1 == fgets(ans, stdin)) {
      % we are dealing with a file that is not open or does not exist
      error("Error: unable to open standard input.");
   }

   % strip trialing carriage returns...
   @ans = strtrim_end(@ans); 

   if (@ans == "") {
      @ans = NULL;
   } else if (type == Integer_Type) {
      @ans = integer(@ans);
   } else if (type == Float_Type or type == Double_Type) {
      @ans = atof(@ans);
   }

}

%----

define get_ncol (file) {
   
   % Determine the number of columns in an ascii file containing 
   % rows and columns of data.

   variable nc = 0;
   variable fp = fopen(file,"r");
   variable line, words;

   if (fp == NULL) { 
      verror("Could not open file %s", file);
   }
   
   while (-1 != fgets(&line, fp)) {
      line = Str->remove_excess_whitespace(line);
      % ignore comments
      if ((line[0] == '%') or (strlen(line) == 0)) {
	 continue;
      }
      words = strchop(line,' ',0);
      while (andelse {nc < length(words)} {words[nc][0] != '%'}) {
	 nc++;
      }
      break;
   }
       
   () = fclose(fp);

   return nc;

}

%----

define get_nrow (file) {

   % Determine the number of rows in an ascii file containing 
   % rows and columns of data. Comments ("%") and empty lines
   % are ignored.

   variable nr = 0;
   variable fp = fopen(file,"r");
   variable line;

   if (fp == NULL) { 
      verror("Could not open file %s", file);
   }
   
   while (-1 != fgets(&line, fp)) {

      line = Str->remove_excess_whitespace(line);

      % ignore comments and empty lines
      if ((line[0] == '%') or (strlen(line) == 0)) {
	 continue;
      } else {
	 % increment nrow
	 nr++; 
      }
   }
       
   () = fclose(fp);

   return nr;

}

%----

% This function may be used recursively.
private define max_type(type1, type2) {

   % Determines the largest common denominator of the types found
   % in the file. 

   if (Utils->member(type1,[Integer_Type, Float_Type, Double_Type]) == NULL) 
       return NULL;
   if (Utils->member(type2,[Integer_Type, Float_Type, Double_Type]) == NULL) 
       return NULL;
   if (type1 == Integer_Type and type2 == Float_Type) return Float_Type;
   if (type1 == Integer_Type and type2 == Double_Type) return Double_Type;
   if (type1 == Float_Type and type2 == Double_Type) return Double_Type;

   max_type(type2, type1);

}

%----

define get_type(file) {

   % Determine the largest common denominator of the data tape present
   % in the data block.

   variable type = NULL, nc = get_ncol(file), i;
   variable fp = fopen(file,"r");
   variable line, words;

   if (fp == NULL) { 
      verror("Could not open file %s", file);
   }
   
   while (-1 != fgets(&line, fp)) {
      % ignore comments

      line = Str->remove_excess_whitespace(line);
      % ignore comments and empty lines
      if ((line[0] == '%') or (strlen(line) == 0)) {
	 continue;
      }

      words = strchop(line,' ',0);
      for (i = 0; i < nc; i++) {
	 if (type == NULL) {
	    % intialize 'type'
	    type = _slang_guess_type(words[i]);
	 } else if (type != _slang_guess_type(words[i])) {
	    % found a type different from 'type'; set 'type' to 
	    % most general type.
	    type = max_type(type,_slang_guess_type(words[i]));
	 }
      }
   }
       
   () = fclose(fp);

   return type;

}

%----

define read_block (file) {

   % Read rows and columns from a file and return a 2D array containing
   % the information.

   variable nr = get_nrow(file);        % get rows
   variable nc = get_ncol(file);        % get columns

   if (orelse {nr == 0} {nc == 0}) return NULL;

   variable type = get_type(file);      % get the type
   variable block = type[nr,nc];        % define the block
   
   variable fp = fopen(file,"r");
   variable line, words, format, value;
   variable _r, _c;
   
   if (type == Integer_Type) {
      format = "%d";
   } else if (type == Float_Type) {
      format = "%e";
   } else if (type == Double_Type) {
      format = "%e";
   }

   if (fp == NULL) { 
      verror("Could not open file %s", file);
   }
   
   _r = 0;
   while (-1 != fgets(&line, fp)) {
      % ignore comments

      line = Str->remove_excess_whitespace(line);
      % ignore comments and empty lines
      if ((line[0] == '%') or (strlen(line) == 0)) {
	 continue;
      }

      words = strchop(line,' ',0);
      for (_c = 0; _c < nc; _c++) {
	 () = sscanf(words[_c],format,&value);
         block[_r,_c] = typecast(value,type);
      }
      _r++;
   }
       
   () = fclose(fp);

   return block;
   
}

%----

define extract_column(block,col) {

   % Extract a column from the data block

   variable usage = "IO->extract_column(Array_Type block, Integer_Type col)";

   if (Array_Type != typeof(block)) {
       message(usage);
   }

   variable dims, ndims, type;
   (dims,ndims,type) = array_info(block);
   
   if (andelse {ndims == 2} {dims[1] > col}) {
       return block[*,col];
   } else {
       message("Number of dimensions or dimensions itself are out of range");
       return NULL;
   }

}

%----

define extract_row(block,row) {

   % Extract a row from the data block

   variable usage = "IO->extract_row(Array_Type block, Integer_Type row)";

   if (Array_Type != typeof(block)) {
       message(usage);
   }

   variable dims, ndims, type;
   (dims,ndims,type) = array_info(block);
   
   if (andelse {ndims == 2} {dims[0] > row}) {
       return block[row,*];
   } else {
       message("Number of dimensions or dimensions itself are out of range");
       return NULL;
   }

}

%----

define write_block() {

   %
   % Write a 2-D array or a list of 1-D arrays to file.
   %

   variable usage = "IO->write_block(filename, array [, array ...])";
   variable filename, fp, block = NULL, ll = NULL, n, m, i, j, arrays;
   variable dims, ndims, type;
   variable str, tmpstr, fmt;

   if (_NARGS < 2) {
       vmessage(usage);
       return;
   } else if (_NARGS == 2) {
       (filename, block) = ();
       % get dimensions
       (dims, ndims, type) = array_info(block);
       if (ndims == 1) {
          n = dims[0];
          m = 1;
          reshape(block, [n,m]);
          fmt = Utils->get_format(block[0,0]) + " ";
       } else if (ndims == 2) {
          n = dims[0];
          m = dims[1];
          fmt = Utils->get_format(block[0,0]) + " ";
       } else {
          error("IO->write_block deals only with 1-D or 2-D arrays");
       }
       %
       % Write to file
       %
       fp = fopen(filename,"w");
       if (fp == NULL) {
          verror("IO->write_block: error opening/creating file %s",
                 filename);
       }
       for (i = 0; i < n; i++) {
          str = "";
          for (j = 0; j < m; j++) {
             str = str + sprintf(fmt,block[i,j]);
          }
          () = fprintf(fp,"%s\n",str);
       }
       () = fclose(fp);
   } else {
       arrays = __pop_args(_NARGS-1);
       (filename) = ();
       % put arrays into a linked list. 
       ll = LL->ll_from_args(__push_args(arrays));
       % get dimensions
       n = length(LL->nth(0,ll));
       m = LL->length(ll);
       %
       % check if all arrays have the same dimension
       for (i = 1; i < m; i++) {
          if (length(LL->nth(i,ll)) != n) {
             error("IO->write_block: arrays have different dimensions");
          }
       }
       %
       % Write to file
       %
       fp = fopen(filename,"w");
       if (fp == NULL) {
          verror("IO->write_block: error opening/creating file %s",
                 filename);
       }
       for (i = 0; i < n; i++) {
          str = "";
          for (j = 0; j < m; j++) {
             block = LL->nth(j,ll);
             fmt = Utils->get_format(block[0]) + " ";
             str = str + sprintf(fmt,block[i]);
          }
          () = fprintf(fp,"%s\n",str);
       }
        () = fclose(fp);
   }

}

%----



