#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>

#include "slang.h"
#include "apec.h"
#include "_apec_slang.h"

static unsigned char _initialized;

int init_plasma_module(void);
void plasma(void);
void ncomp_plasma(void);
void load_apec_model(void);
void cleanup_apec_model(void);
void elem_ion(void);
void wvl(void);
void emiss(void);
void ll_ul(void);
void em_power_law(void);
void em_broken_power_law(void);
void em_double_power_law(void);
void em_chebyshev(void);
void em_polynomial(void);
static SLang_Intrin_Fun_Type Plasma_Intrinsics [] = {
  MAKE_INTRINSIC_0("plasma", plasma, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("ncomp_plasma", ncomp_plasma, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("load_apec_model", load_apec_model, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("cleanup_apec_model", cleanup_apec_model, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("elem_ion", elem_ion, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("wvl", wvl, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("emiss", emiss, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("ll_ul", ll_ul, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("em_power_law", em_power_law, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("em_broken_power_law", em_broken_power_law, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("em_double_power_law", em_double_power_law, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("em_chebyshev", em_chebyshev, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("em_polynomial", em_polynomial, SLANG_VOID_TYPE),
  SLANG_END_INTRIN_FUN_TABLE
};

int init_plasma_module(void) {
  if (_initialized) return 0;
  _initialized = 1;
  if (-1 == SLadd_intrin_fun_table (Plasma_Intrinsics, NULL)) {
    return -1;
  }
  SLang_vmessage("Plasma module installed.");
  return 0;
}


void cleanup_apec_model(void) {

  free_all_apec_models();

}

/* 
   From SLang call this function as


*/

void load_apec_model(void) {

  /* Inputs are the user defined name of the model and 
     line and continuum filenames of the apec model outputs */

  char *model_name;
  char *line_filename;
  char *coco_filename;
  char *filestem;
  int free_model_name, free_line_filename, free_coco_filename, free_filestem;

  if (SLang_Num_Function_Args != 4) {
    SLang_verror (SL_USAGE_ERROR, "Usage: load_apec_model()");
    return;
  }

  /* 
     pop filestem
   */
  if (-1 == SLang_pop_string(&filestem,&free_filestem)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping filestem");
    return;
  }
  /* 
     pop coco_filename
   */
  if (-1 == SLang_pop_string(&coco_filename,&free_coco_filename)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping coco_filename");
    return;
  }
  /* 
     pop line_filename
   */
  if (-1 == SLang_pop_string(&line_filename,&free_line_filename)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping line_filename");
    return;
  }
  /* 
     pop model_name
   */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }

  /* load apec model in non-verbose mode */
  if (0 != load_plasma(model_name,line_filename,coco_filename,filestem,0)) {
    SLang_verror(SL_USER_ERROR,"Failed to read apec model.");
   return;
  }

  if (free_filestem == 1) SLfree(filestem);
  if (free_coco_filename == 1) SLfree(coco_filename);
  if (free_line_filename == 1) SLfree(line_filename);
  if (free_model_name == 1) SLfree(model_name);

}

/* 
   From SLang call this function as

   plasma(modelname,
	  spectrum_type,         lines/continuum/both
	  linelistname,
          kT,
	  density,
	  norm,
	  abundances,
	  redshift,
	  kev,
	  binlo,
	  binhi)

   We always expect binlo AND binhi for this routine. We will provide
   SLang scripts to deal with integrate on/off.

*/

void plasma(void) {

  /* Inputs */
  char *model_name;
  int spectrum_type = 0;
  SLang_Array_Type *sl_linelist;
  long nlinelist;
  long *linelist;
  double dbl_kT, dbl_density, dbl_norm;
  float kT, density, norm;
  SLang_Array_Type *sl_abundances;
  int nabundances;
  float *abundances;
  float redshift;
  int kev;
  SLang_Array_Type *sl_binlo, *sl_binhi, *sl_spectrum;
  int i, nbin;
  float *binlo, *binhi, *spectrum;
  double value;
  int free_model_name=0, linelist_supplied=0;

  if (SLang_Num_Function_Args != 11) {
    SLang_verror (SL_USAGE_ERROR, "Usage: plasma()");
    return;
  }

  /* Pop binhi */
  if (-1 == SLang_pop_array_of_type(&sl_binhi, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* Pop binlo */
  if (-1 == SLang_pop_array_of_type(&sl_binlo, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* 
     Pop keV from the stack
   */
  if (-1 == SLang_pop_integer(&kev)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping kev flag");
    return;
  }
  /* 
     Pop redshift from the stack
   */
  if (-1 == SLang_pop_double(&value, NULL, NULL)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping redshift");
    return;
  }
  redshift = (float)value;
  /* 
     Pop abundances from the stack
   */
  if (-1 == SLang_pop_array_of_type(&sl_abundances,SLANG_DOUBLE_TYPE)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping abundances pointer");
    return;
  }
  /* 
     Pop kT, density, and norm from the stack
   */
  if ((-1 == SLang_pop_double(&dbl_norm, NULL, NULL)) || 
      (-1 == SLang_pop_double(&dbl_density, NULL, NULL)) || 
      (-1 == SLang_pop_double(&dbl_kT, NULL, NULL))) { 
    SLang_verror(SL_STACK_OVERFLOW, "Error popping norm, denisty, or kT");
    return;
  }
  /* convert to float */
  kT      = (float)dbl_kT;
  density = (float)dbl_density;
  norm    = (float)dbl_norm;
  /* 
     Pop linelist from the stack. If linelist is NULL than we mean to
     do all lines
   */
  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    linelist_supplied = 0;
  } else {
    if (-1 == SLang_pop_array_of_type(&sl_linelist,SLANG_LONG_TYPE)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping linelist pointer");
      return;
    }
    linelist_supplied = 1;
  }
  /* 
     Pop flag from the stack
   */
  if (-1 == SLang_pop_integer(&spectrum_type)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping spectrum_type");
    return;
  }
  /* 
     First argument, last to pop from stack: Pop model name from the stack
   */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }


  /* 
     Transfer SLang_arrays to c-arrays
  */
  nbin = sl_binlo->num_elements;
  if (nbin != sl_binhi->num_elements) {
    SLang_verror(SL_USER_ERROR,"Arrays binlo and binhi are not of the same length");
    return;
  }
  binlo    = (float *)malloc(sizeof(float)*nbin);
  binhi    = (float *)malloc(sizeof(float)*nbin);
  spectrum = (float *)malloc(sizeof(float)*nbin);
  for(i=0;i<nbin;i++) {
    SLang_get_array_element(sl_binlo, &i, &value);
    binlo[i] = (float)value;
    SLang_get_array_element(sl_binhi, &i, &value);
    binhi[i] = (float)value;
    spectrum[i] = 0.0;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nabundances = sl_abundances->num_elements;
  abundances = (float *)malloc(sizeof(float)*nabundances);
  for(i=0;i<nabundances;i++) {
    SLang_get_array_element(sl_abundances, &i, &value);
    abundances[i] = (float)value;
  }
  if (nabundances < 29) {
    SLang_verror(SL_USER_ERROR,"The abundances array must have at least 29 elements to accomodate all elements in the APED database.");
    return;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  if (linelist_supplied == 1) {
    nlinelist = sl_linelist->num_elements;
    linelist = (long *)malloc(sizeof(long)*nlinelist);
    for(i=0;i<nlinelist;i++) {
      SLang_get_array_element(sl_linelist, &i, &(linelist[i]));
    }
  } else {
    nlinelist = 0;
    linelist = NULL;
  }

  if (-1 == extract_spectrum_for_modelname(model_name,
					   kT,density,norm,abundances,
					   redshift,spectrum_type,kev,
					   nlinelist,linelist,nbin, 
					   binlo,binhi,spectrum)) {
    return;
  }

  sl_spectrum = SLang_create_array(SLANG_DOUBLE_TYPE, 0, NULL, &nbin, 1);
  if (sl_spectrum == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nbin;i++) {
    value = (double)spectrum[i];
    if (-1 == SLang_set_array_element(sl_spectrum,&i,&value)) {
      SLang_free_array(sl_spectrum);
      return;
    }
  }

  if (free_model_name == 1) SLfree(model_name);
  SLang_free_array(sl_binlo);
  SLang_free_array(sl_binhi);
  free(binlo);
  free(binhi);
  free(spectrum);
  if (linelist_supplied == 1) {
    SLang_free_array(sl_linelist);
    free(linelist);
  }

  SLang_push_array(sl_spectrum,1);

}

/* 
   From SLang call this function as

   plasma(modelname,
	  spectrum_type,         lines/continuum/both
	  linelistname,
          kT[nfit],
	  density[nfit],
	  norm[nfit],
	  abundances,
	  redshift,
	  kev,
	  binlo,
	  binhi)

   We always expect binlo AND binhi for this routine. We will provide
   SLang scripts to deal with integrate on/off.

*/

void ncomp_plasma(void) {

  /* Inputs */
  char *model_name;
  int spectrum_type = 0;
  SLang_Array_Type *sl_linelist;
  long nlinelist;
  long *linelist;
  SLang_Array_Type *sl_kT, *sl_density, *sl_norm;
  int nfit;
  float *kT, *density, *norm;
  SLang_Array_Type *sl_abundances;
  int nabundances;
  float *abundances;
  float redshift;
  int kev;
  SLang_Array_Type *sl_binlo, *sl_binhi, *sl_spectrum;
  int i, nbin;
  float *binlo, *binhi, *spectrum;
  double value;
  int free_model_name=0, linelist_supplied=0;

  if (SLang_Num_Function_Args != 11) {
    SLang_verror (SL_USAGE_ERROR, "Usage: plasma()");
    return;
  }

  /* Pop binhi */
  if (-1 == SLang_pop_array_of_type(&sl_binhi, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* Pop binlo */
  if (-1 == SLang_pop_array_of_type(&sl_binlo, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* 
     Pop keV from the stack
   */
  if (-1 == SLang_pop_integer(&kev)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping kev flag");
    return;
  }
  /* 
     Pop redshift from the stack
   */
  if (-1 == SLang_pop_double(&value, NULL, NULL)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping redshift");
    return;
  }
  redshift = (float)value;
  /* 
     Pop abundances from the stack
   */
  if (-1 == SLang_pop_array_of_type(&sl_abundances,SLANG_DOUBLE_TYPE)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping abundances pointer");
    return;
  }
  /* 
     Pop norm from the stack
   */
  if (-1 == SLang_pop_array_of_type(&sl_norm,SLANG_DOUBLE_TYPE)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping norm pointer");
    return;
  }
  /* 
     Pop density from the stack
   */
  if (-1 == SLang_pop_array_of_type(&sl_density,SLANG_DOUBLE_TYPE)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping density pointer");
    return;
  }
  /* 
     Pop kT from the stack
   */
  if (-1 == SLang_pop_array_of_type(&sl_kT,SLANG_DOUBLE_TYPE)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping kT pointer");
    return;
  }
  /* 
     Pop linelist from the stack. If linelist is NULL than we mean to
     do all lines
   */
  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    linelist_supplied = 0;
  } else {
    if (-1 == SLang_pop_array_of_type(&sl_linelist,SLANG_LONG_TYPE)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping linelist pointer");
      return;
    }
    linelist_supplied = 1;
  }
  /* 
     Pop flag from the stack
   */
  if (-1 == SLang_pop_integer(&spectrum_type)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping spectrum_type");
    return;
  }
  /* 
     First argument, last to pop from stack: Pop model name from the stack
   */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }

  /* 
     Transfer SLang_arrays to c-arrays
  */
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nbin = sl_binlo->num_elements;
  if (nbin != sl_binhi->num_elements) {
    SLang_verror(SL_USER_ERROR,"Arrays binlo and binhi are not of the same length");
    return;
  }
  binlo    = (float *)malloc(sizeof(float)*nbin);
  binhi    = (float *)malloc(sizeof(float)*nbin);
  spectrum = (float *)malloc(sizeof(float)*nbin);
  for(i=0;i<nbin;i++) {
    SLang_get_array_element(sl_binlo, &i, &value);
    binlo[i] = (float)value;
    SLang_get_array_element(sl_binhi, &i, &value);
    binhi[i] = (float)value;
    spectrum[i] = 0.0;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nfit = sl_kT->num_elements;
  if (nfit != sl_density->num_elements ||
      nfit != sl_norm->num_elements) {
    SLang_verror(SL_USER_ERROR,"Arrays kT, density, and norm are not of the same length");
    return;
  }
  kT      = (float *)malloc(sizeof(float)*nfit);
  density = (float *)malloc(sizeof(float)*nfit);
  norm    = (float *)malloc(sizeof(float)*nfit);
  for(i=0;i<nfit;i++) {
    SLang_get_array_element(sl_kT, &i, &value);
    kT[i] = (float)value;
    SLang_get_array_element(sl_density, &i, &value);
    density[i] = (float)value;
    SLang_get_array_element(sl_norm, &i, &value);
    norm[i] = (float)value;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nabundances = sl_abundances->num_elements;
  abundances = (float *)malloc(sizeof(float)*nabundances);
  for(i=0;i<nabundances;i++) {
    SLang_get_array_element(sl_abundances, &i, &value);
    abundances[i] = (float)value;
  }
  
  if (nabundances < 29) {
    SLang_verror(SL_USER_ERROR,"The abundances array must have at least 29 elements to accomodate all elements in the APED database.");
    return;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  if (linelist_supplied == 1) {
    nlinelist = sl_linelist->num_elements;
    linelist = (long *)malloc(sizeof(long)*nlinelist);
    for(i=0;i<nlinelist;i++) {
      SLang_get_array_element(sl_linelist, &i, &(linelist[i]));
    }
  } else {
    nlinelist = 0;
    linelist = NULL;
  }

  printf("%d %f %f\n",nfit,kT[0],norm[0]);
  if (-1 == extract_ncomp_spectrum_for_modelname(model_name,nfit,
						 kT,density,norm,abundances,
						 redshift,spectrum_type,kev,
						 nlinelist,linelist,nbin, 
						 binlo,binhi,spectrum)) {
    return;
  }

  sl_spectrum = SLang_create_array(SLANG_DOUBLE_TYPE, 0, NULL, &nbin, 1);
  if (sl_spectrum == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nbin;i++) {
    value = (double)spectrum[i];
    if (-1 == SLang_set_array_element(sl_spectrum,&i,&value)) {
      SLang_free_array(sl_spectrum);
      return;
    }
  }


  if (free_model_name == 1) SLfree(model_name);
  SLang_free_array(sl_binlo);
  SLang_free_array(sl_binhi);
  free(binlo);
  free(binhi);
  free(spectrum);
  if (linelist_supplied == 1) {
    SLang_free_array(sl_linelist);
    free(linelist);
  }

  SLang_push_array(sl_spectrum,1);

}

/*



 */
void elem_ion() {
  /*
    In: modelname (String_Type)
        elements  (Array_Type)
        ions      (Array_Type)
    Out: index list
  */

  SLang_Array_Type *sl_elem, *sl_ion, *sl_linelist;
  char *model_name;
  int i, free_model_name, stack, val=0;
  long *elem, *ion, *linelist, nelem, nion, nlinelist;
  
  if ((SLang_Num_Function_Args != 3) && (SLang_Num_Function_Args != 2)) {
    SLang_verror (SL_USAGE_ERROR, "Usage: elem_ion()");
    return;
  }

  nion = 0;
  if (SLang_Num_Function_Args == 3) {
    /* 
       pop ions
    */
    stack = SLang_peek_at_stack();
    if (stack == SLANG_NULL_TYPE) {
      if (SLang_pop_null()==-1) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
      }
      nion = 0;
    } else if (stack == SLANG_ARRAY_TYPE) {
      if (-1 == SLang_pop_array_of_type(&sl_ion,SLANG_LONG_TYPE)) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping ion pointer");
	return;
      }
      nion= sl_ion->num_elements;
      ion = (long *)malloc(sizeof(long)*nion);
      for(i=0;i<nion;i++) 
        SLang_get_array_element(sl_ion, &i, &(ion[i]));
    } else {
      /* must be dealing with a number */
      if (-1 == SLang_pop_integer(&val)) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping ion");
	return;
      }
      nion = 1;
      ion = (long *)malloc(sizeof(long));
      ion[0] = (long)val;
    }
  }
  /* 
     pop elements
   */
  stack = SLang_peek_at_stack();
  if (stack == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    nelem = 0;
  } else if (stack == SLANG_ARRAY_TYPE) {
    if (-1 == SLang_pop_array_of_type(&sl_elem,SLANG_LONG_TYPE)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping element pointer");
      return;
    }
    nelem= sl_elem->num_elements;
    elem = (long *)malloc(sizeof(long)*nelem);
    for(i=0;i<nelem;i++) 
      SLang_get_array_element(sl_elem, &i, &(elem[i]));
  } else {
    /* must be dealing with a number */
     if (-1 == SLang_pop_integer(&val)) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping elem");
	return;
     }
    nelem = 1;
    elem = (long *)malloc(sizeof(long));
    elem[0] = val;
  }
  /* 
     pop modelname
  */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }

  nlinelist = 0;
  linelist = NULL;
  linelist = _elem_ion_(model_name,nelem,elem,nion,ion,&nlinelist);

  i = (int)nlinelist;
  sl_linelist = SLang_create_array(SLANG_LONG_TYPE, 0, NULL, &i, 1);
  if (sl_linelist == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nlinelist;i++) {
    if (-1 == SLang_set_array_element(sl_linelist,&i,&(linelist[i]))) {
      SLang_free_array(sl_linelist);
      return;
    }
  }

  if (nlinelist > 0) free(linelist);
  if (nion > 0) free(ion);
  if (nelem > 0) free(elem);
  if (free_model_name == 1) SLfree(model_name);

  SLang_push_array(sl_linelist,1);

}

void wvl() {
  /*
    In: modelname
        wvl_min
        wvl_max
	kev (optional)
    Out: index list
  */
  
  SLang_Array_Type *sl_linelist;
  char *model_name;
  double wvl_min,wvl_max;
  int i, kev, free_model_name;
  long *linelist, nlinelist;
  
  if ((SLang_Num_Function_Args != 4) && (SLang_Num_Function_Args != 3) && (SLang_Num_Function_Args != 2))  {
    SLang_verror (SL_USAGE_ERROR, "Usage: wvl()");
    return;
  }

  kev = 0;
  if (SLang_Num_Function_Args == 4) {
    /* 
       pop kev
    */
    if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
      if (SLang_pop_null()==-1) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
      }
      kev = 0;
    } else {
      if (-1 == SLang_pop_integer(&kev)) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping kev");
	return;
      }
    }
  } 

  wvl_max = -1.0;
  if (SLang_Num_Function_Args == 3 || SLang_Num_Function_Args == 4) {
    /* 
     pop wvl_max
    */
    if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
      if (SLang_pop_null()==-1) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
      }
      wvl_max = -1.0;
    } else {
      if (-1 == SLang_pop_double(&wvl_max,NULL,NULL)) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping wvl_max");
	return;
      }
    }
  }
  /* 
     pop wvl_min
   */
  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    wvl_min = -1.0;
  } else {
    if (-1 == SLang_pop_double(&wvl_min,NULL,NULL)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping wvl_min");
      return;
    }
  }
  /* 
     pop modelname
  */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }
  
  nlinelist = 0;
  linelist = NULL;
  linelist = _wvl_(model_name,wvl_min,wvl_max,kev,&nlinelist);

  i = (int)nlinelist;
  sl_linelist = SLang_create_array(SLANG_LONG_TYPE, 0, NULL, &i, 1);
  if (sl_linelist == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nlinelist;i++) {
    if (-1 == SLang_set_array_element(sl_linelist,&i,&(linelist[i]))) {
      SLang_free_array(sl_linelist);
      return;
    }
  }

  if (nlinelist > 0) free(linelist);
  if (free_model_name == 1) SLfree(model_name);

  SLang_push_array(sl_linelist,1);
}

void emiss() {
  /*
    In: modelname
        emiss_min
        emiss_max
    Out: index list
  */

  SLang_Array_Type *sl_linelist;
  char *model_name;
  double emiss_min,emiss_max;
  int i, free_model_name;
  long *linelist, nlinelist;
  
  if ((SLang_Num_Function_Args != 3) && (SLang_Num_Function_Args != 2))  {
    SLang_verror (SL_USAGE_ERROR, "Usage: emiss()");
    return;
  }

  emiss_max = -1.0;
  if (SLang_Num_Function_Args == 3) {
    /* 
       pop emiss_max
    */
    if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
      if (SLang_pop_null()==-1) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
      }
      emiss_max = -1.0;
    } else {
      if (-1 == SLang_pop_double(&emiss_max,NULL,NULL)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping emiss_max");
      return;
      }
    }
  }
  /* 
     pop emiss_min
   */
  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    emiss_min = -1.0;
  } else {
    if (-1 == SLang_pop_double(&emiss_min,NULL,NULL)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping emiss_min");
      return;
    }
  }
  /* 
     pop modelname
  */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }
  
  nlinelist = 0;
  linelist = NULL;
  linelist = _emiss_(model_name,emiss_min,emiss_max,&nlinelist);

  i = (int)nlinelist;
  sl_linelist = SLang_create_array(SLANG_LONG_TYPE, 0, NULL, &i, 1);
  if (sl_linelist == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nlinelist;i++) {
    if (-1 == SLang_set_array_element(sl_linelist,&i,&(linelist[i]))) {
      SLang_free_array(sl_linelist);
      return;
    }
  }

  if (nlinelist > 0) free(linelist);
  if (free_model_name == 1) SLfree(model_name);

  SLang_push_array(sl_linelist,1);

}

void ll_ul() {
  /*
    In: modelname
        ll
        ul
    Out: index list
  */

  SLang_Array_Type *sl_ll, *sl_ul, *sl_linelist;
  char *model_name;
  int i, free_model_name,stack,val;
  long *ll, *ul, *linelist, nll, nul, nlinelist;
  
  if ((SLang_Num_Function_Args != 3) && (SLang_Num_Function_Args != 2)){
    SLang_verror (SL_USAGE_ERROR, "Usage: ll_ul()");
    return;
  }

  nul = 0;
  if (SLang_Num_Function_Args == 3) {
    /* 
       pop uls
    */
    stack = SLang_peek_at_stack();
    if (stack == SLANG_NULL_TYPE) {
      if (SLang_pop_null()==-1) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
      }
      nul = 0;
    } else if (stack == SLANG_ARRAY_TYPE) {
      if (-1 == SLang_pop_array_of_type(&sl_ul,SLANG_LONG_TYPE)) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping ul pointer");
	return;
      }
      nul= sl_ul->num_elements;
      ul = (long *)malloc(sizeof(long)*nul);
      for(i=0;i<nul;i++) 
        SLang_get_array_element(sl_ul, &i, &(ul[i]));
    } else {
      /* must be dealing with a number */
      if (-1 == SLang_pop_integer(&val)) {
	SLang_verror(SL_STACK_OVERFLOW, "Error popping ul");
	return;
      }
      nul = 1;
      ul = (long *)malloc(sizeof(long));
      ul[0] = (long)val;
    }
  }
  /* 
     pop elements
   */
  stack = SLang_peek_at_stack();
  if (stack == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    nll = 0;
  } else if (stack == SLANG_ARRAY_TYPE) {
    if (-1 == SLang_pop_array_of_type(&sl_ll,SLANG_LONG_TYPE)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping element pointer");
      return;
    }
    nll= sl_ll->num_elements;
    ll = (long *)malloc(sizeof(long)*nll);
    for(i=0;i<nll;i++) 
      SLang_get_array_element(sl_ll, &i, &(ll[i]));
  } else {
    /* must be dealing with a number */
    if (-1 == SLang_pop_integer(&val)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping ll");
      return;
    }
    nll = 1;
    ll = (long *)malloc(sizeof(long));
    ll[0] = (long)val;
  }
  /* 
     pop modelname
  */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }
  
  nlinelist = 0;
  linelist = NULL;
  linelist = _ll_ul_(model_name,nll,ll,nul,ul,&nlinelist);

  i = (int)nlinelist;
  sl_linelist = SLang_create_array(SLANG_LONG_TYPE, 0, NULL, &i, 1);
  if (sl_linelist == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nlinelist;i++) {
    if (-1 == SLang_set_array_element(sl_linelist,&i,&(linelist[i]))) {
      SLang_free_array(sl_linelist);
      return;
    }
  }

  if (nlinelist > 0) free(linelist);
  if (nul > 0) free(ul);
  if (nll > 0) free(ll);
  if (free_model_name == 1) SLfree(model_name);

  SLang_push_array(sl_linelist,1);

}

/*
  em_power_law(modelname,
	  spectrum_type,         lines/continuum/both
	  linelistname,
	  nT,
	  Tmin,
	  Tmax,
	  log,
	  gamma,
	  ampl,
	  density,
	  abundances,
	  redshift,
	  kev,
	  binlo,
	  binhi)

*/

void em_power_law(void) {

  /* Inputs */
  char *model_name;
  float gamma, ampl;
  double dbl_gamma, dbl_ampl;
  int nT, log;
  float Tmin,Tmax;
  double dbl_Tmin,dbl_Tmax;
  int spectrum_type = 0;
  SLang_Array_Type *sl_linelist;
  long nlinelist;
  long *linelist;
  double dbl_density;
  float density;
  SLang_Array_Type *sl_abundances;
  int nabundances;
  float *abundances;
  float redshift;
  int kev;
  SLang_Array_Type *sl_binlo, *sl_binhi, *sl_spectrum;
  int i, nbin;
  float *binlo, *binhi, *spectrum;
  double value;
  int free_model_name=0, linelist_supplied=0;

  if (SLang_Num_Function_Args != 15) {
    SLang_verror (SL_USAGE_ERROR, "Usage: em_power_law()");
    return;
  }

  /* Pop binhi */
  if (-1 == SLang_pop_array_of_type(&sl_binhi, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* Pop binlo */
  if (-1 == SLang_pop_array_of_type(&sl_binlo, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* 
     Pop keV from the stack
   */
  if (-1 == SLang_pop_integer(&kev)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping kev flag");
    return;
  }
  /* 
     Pop redshift from the stack
   */
  if (-1 == SLang_pop_double(&value, NULL, NULL)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping redshift");
    return;
  }
  redshift = (float)value;
  /* 
     Pop abundances from the stack
   */
  if (-1 == SLang_pop_array_of_type(&sl_abundances,SLANG_DOUBLE_TYPE)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping abundances pointer");
    return;
  }
  /* 
     Pop gamma, ampl, density, log, Tmax, Tmin, nT from the stack
   */
  if ((-1 == SLang_pop_double(&dbl_density, NULL, NULL)) || 
      (-1 == SLang_pop_double(&dbl_ampl, NULL, NULL)) || 
      (-1 == SLang_pop_double(&dbl_gamma, NULL, NULL)) ||
      (-1 == SLang_pop_integer(&log)) ||
      (-1 == SLang_pop_double(&dbl_Tmax, NULL, NULL)) ||
      (-1 == SLang_pop_double(&dbl_Tmin, NULL, NULL)) ||
      (-1 == SLang_pop_integer(&nT))) { 
    SLang_verror(SL_STACK_OVERFLOW, "Error popping density, ampl, gamma, log, Tmax, Tmin, or nT");
    return;
  }
  /* convert to float */
  density = (float)dbl_density;
  ampl    = (float)dbl_ampl;
  gamma   = (float)dbl_gamma;
  Tmin    = (float)dbl_Tmin;
  Tmax    = (float)dbl_Tmax;
  /* 
     Pop linelist from the stack. If linelist is NULL than we mean to
     do all lines
   */
  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    linelist_supplied = 0;
  } else {
    if (-1 == SLang_pop_array_of_type(&sl_linelist,SLANG_LONG_TYPE)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping linelist pointer");
      return;
    }
    linelist_supplied = 1;
  }
  /* 
     Pop flag from the stack
   */
  if (-1 == SLang_pop_integer(&spectrum_type)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping spectrum_type");
    return;
  }
  /* 
     First argument, last to pop from stack: Pop model name from the stack
   */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }


  /* 
     Transfer SLang_arrays to c-arrays
  */
  nbin = sl_binlo->num_elements;
  if (nbin != sl_binhi->num_elements) {
    SLang_verror(SL_USER_ERROR,"Arrays binlo and binhi are not of the same length");
    return;
  }
  binlo    = (float *)malloc(sizeof(float)*nbin);
  binhi    = (float *)malloc(sizeof(float)*nbin);
  spectrum = (float *)malloc(sizeof(float)*nbin);
  for(i=0;i<nbin;i++) {
    SLang_get_array_element(sl_binlo, &i, &value);
    binlo[i] = (float)value;
    SLang_get_array_element(sl_binhi, &i, &value);
    binhi[i] = (float)value;
    spectrum[i] = 0.0;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nabundances = sl_abundances->num_elements;
  abundances = (float *)malloc(sizeof(float)*nabundances);
  for(i=0;i<nabundances;i++) {
    SLang_get_array_element(sl_abundances, &i, &value);
    abundances[i] = (float)value;
  }
  if (nabundances < 29) {
    SLang_verror(SL_USER_ERROR,"The abundances array must have at least 29 elements to accomodate all elements in the APED database.");
    return;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  if (linelist_supplied == 1) {
    nlinelist = sl_linelist->num_elements;
    linelist = (long *)malloc(sizeof(long)*nlinelist);
    for(i=0;i<nlinelist;i++) {
      SLang_get_array_element(sl_linelist, &i, &(linelist[i]));
    }
  } else {
    nlinelist = 0;
    linelist = NULL;
  }

  if (-1 == _em_power_law_(model_name,
			   gamma,ampl,
			   nT, Tmin, Tmax, log,
			   density,abundances,
			   redshift,spectrum_type,kev,
			   nlinelist,linelist,nbin, 
			   binlo,binhi,spectrum)) {
    return;
  }

  sl_spectrum = SLang_create_array(SLANG_DOUBLE_TYPE, 0, NULL, &nbin, 1);
  if (sl_spectrum == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nbin;i++) {
    value = (double)spectrum[i];
    if (-1 == SLang_set_array_element(sl_spectrum,&i,&value)) {
      SLang_free_array(sl_spectrum);
      return;
    }
  }

  if (free_model_name == 1) SLfree(model_name);
  SLang_free_array(sl_binlo);
  SLang_free_array(sl_binhi);
  free(binlo);
  free(binhi);
  free(spectrum);
  if (linelist_supplied == 1) {
    SLang_free_array(sl_linelist);
    free(linelist);
  }

  SLang_push_array(sl_spectrum,1);

}

/*
  em_power_law(modelname,
	  spectrum_type,         lines/continuum/both
	  linelistname,
	  nT,
	  Tmin,
	  Tmax,
	  log,
	  gamma1,
	  gamma2,
	  Tbreak,
	  ampl,
	  density,
	  abundances,
	  redshift,
	  kev,
	  binlo,
	  binhi)

*/

void em_broken_power_law(void) {

  /* Inputs */
  char *model_name;
  float gamma1, gamma2, ampl, Tbreak;
  double dbl_gamma1, dbl_gamma2, dbl_ampl, dbl_Tbreak;
  int nT, log;
  float Tmin,Tmax;
  double dbl_Tmin,dbl_Tmax;
  int spectrum_type = 0;
  SLang_Array_Type *sl_linelist;
  long nlinelist;
  long *linelist;
  double dbl_density;
  float density;
  SLang_Array_Type *sl_abundances;
  int nabundances;
  float *abundances;
  float redshift;
  int kev;
  SLang_Array_Type *sl_binlo, *sl_binhi, *sl_spectrum;
  int i, nbin;
  float *binlo, *binhi, *spectrum;
  double value;
  int free_model_name=0, linelist_supplied=0;

  if (SLang_Num_Function_Args != 17) {
    SLang_verror (SL_USAGE_ERROR, "Usage: em_power_law()");
    return;
  }

  /* Pop binhi */
  if (-1 == SLang_pop_array_of_type(&sl_binhi, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* Pop binlo */
  if (-1 == SLang_pop_array_of_type(&sl_binlo, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* 
     Pop keV from the stack
   */
  if (-1 == SLang_pop_integer(&kev)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping kev flag");
    return;
  }
  /* 
     Pop redshift from the stack
   */
  if (-1 == SLang_pop_double(&value, NULL, NULL)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping redshift");
    return;
  }
  redshift = (float)value;
  /* 
     Pop abundances from the stack
   */
  if (-1 == SLang_pop_array_of_type(&sl_abundances,SLANG_DOUBLE_TYPE)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping abundances pointer");
    return;
  }
  /* 
     Pop gamma, ampl, density, log, Tmax, Tmin, nT from the stack
   */
  if ((-1 == SLang_pop_double(&dbl_density, NULL, NULL)) || 
      (-1 == SLang_pop_double(&dbl_ampl, NULL, NULL)) || 
      (-1 == SLang_pop_double(&dbl_Tbreak, NULL, NULL)) ||
      (-1 == SLang_pop_double(&dbl_gamma2, NULL, NULL)) ||
      (-1 == SLang_pop_double(&dbl_gamma1, NULL, NULL)) ||
      (-1 == SLang_pop_integer(&log)) ||
      (-1 == SLang_pop_double(&dbl_Tmax, NULL, NULL)) ||
      (-1 == SLang_pop_double(&dbl_Tmin, NULL, NULL)) ||
      (-1 == SLang_pop_integer(&nT))) { 
    SLang_verror(SL_STACK_OVERFLOW, "Error popping density, ampl, gamma, log, Tmax, Tmin, or nT");
    return;
  }
  /* convert to float */
  density = (float)dbl_density;
  ampl    = (float)dbl_ampl;
  Tbreak  = (float)dbl_Tbreak;
  gamma1   = (float)dbl_gamma1;
  gamma2   = (float)dbl_gamma2;
  Tmin    = (float)dbl_Tmin;
  Tmax    = (float)dbl_Tmax;
  /* 
     Pop linelist from the stack. If linelist is NULL than we mean to
     do all lines
   */
  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    linelist_supplied = 0;
  } else {
    if (-1 == SLang_pop_array_of_type(&sl_linelist,SLANG_LONG_TYPE)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping linelist pointer");
      return;
    }
    linelist_supplied = 1;
  }
  /* 
     Pop flag from the stack
   */
  if (-1 == SLang_pop_integer(&spectrum_type)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping spectrum_type");
    return;
  }
  /* 
     First argument, last to pop from stack: Pop model name from the stack
   */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }


  /* 
     Transfer SLang_arrays to c-arrays
  */
  nbin = sl_binlo->num_elements;
  if (nbin != sl_binhi->num_elements) {
    SLang_verror(SL_USER_ERROR,"Arrays binlo and binhi are not of the same length");
    return;
  }
  binlo    = (float *)malloc(sizeof(float)*nbin);
  binhi    = (float *)malloc(sizeof(float)*nbin);
  spectrum = (float *)malloc(sizeof(float)*nbin);
  for(i=0;i<nbin;i++) {
    SLang_get_array_element(sl_binlo, &i, &value);
    binlo[i] = (float)value;
    SLang_get_array_element(sl_binhi, &i, &value);
    binhi[i] = (float)value;
    spectrum[i] = 0.0;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nabundances = sl_abundances->num_elements;
  abundances = (float *)malloc(sizeof(float)*nabundances);
  for(i=0;i<nabundances;i++) {
    SLang_get_array_element(sl_abundances, &i, &value);
    abundances[i] = (float)value;
  }
  if (nabundances < 29) {
    SLang_verror(SL_USER_ERROR,"The abundances array must have at least 29 elements to accomodate all elements in the APED database.");
    return;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  if (linelist_supplied == 1) {
    nlinelist = sl_linelist->num_elements;
    linelist = (long *)malloc(sizeof(long)*nlinelist);
    for(i=0;i<nlinelist;i++) {
      SLang_get_array_element(sl_linelist, &i, &(linelist[i]));
    }
  } else {
    nlinelist = 0;
    linelist = NULL;
  }

  if (-1 == _em_broken_power_law_(model_name,
				  gamma1,gamma2,Tbreak,ampl,
				  nT, Tmin, Tmax, log,
				  density,abundances,
				  redshift,spectrum_type,kev,
				  nlinelist,linelist,nbin, 
				  binlo,binhi,spectrum)) {
    return;
  }

  sl_spectrum = SLang_create_array(SLANG_DOUBLE_TYPE, 0, NULL, &nbin, 1);
  if (sl_spectrum == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nbin;i++) {
    value = (double)spectrum[i];
    if (-1 == SLang_set_array_element(sl_spectrum,&i,&value)) {
      SLang_free_array(sl_spectrum);
      return;
    }
  }

  if (free_model_name == 1) SLfree(model_name);
  SLang_free_array(sl_binlo);
  SLang_free_array(sl_binhi);
  free(binlo);
  free(binhi);
  free(spectrum);
  if (linelist_supplied == 1) {
    SLang_free_array(sl_linelist);
    free(linelist);
  }

  SLang_push_array(sl_spectrum,1);

}

/*
  em_power_law(modelname,
	  spectrum_type,         lines/continuum/both
	  linelistname,
	  nT,
	  Tmin,
	  Tmax,
	  log,
	  gamma,
	  beta,
	  alpha,
	  Tbreak,
	  ampl,
	  density,
	  abundances,
	  redshift,
	  kev,
	  binlo,
	  binhi)

*/

void em_double_power_law(void) {

  /* Inputs */
  char *model_name;
  float gamma, beta, alpha, Tbreak, ampl;
  double dbl_gamma, dbl_beta, dbl_alpha, dbl_Tbreak, dbl_ampl;
  int nT, log;
  float Tmin,Tmax;
  double dbl_Tmin,dbl_Tmax;
  int spectrum_type = 0;
  SLang_Array_Type *sl_linelist;
  long nlinelist;
  long *linelist;
  double dbl_density;
  float density;
  SLang_Array_Type *sl_abundances;
  int nabundances;
  float *abundances;
  float redshift;
  int kev;
  SLang_Array_Type *sl_binlo, *sl_binhi, *sl_spectrum;
  int i, nbin;
  float *binlo, *binhi, *spectrum;
  double value;
  int free_model_name=0, linelist_supplied=0;

  if (SLang_Num_Function_Args != 18) {
    SLang_verror (SL_USAGE_ERROR, "Usage: em_power_law()");
    return;
  }

  /* Pop binhi */
  if (-1 == SLang_pop_array_of_type(&sl_binhi, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* Pop binlo */
  if (-1 == SLang_pop_array_of_type(&sl_binlo, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* 
     Pop keV from the stack
   */
  if (-1 == SLang_pop_integer(&kev)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping kev flag");
    return;
  }
  /* 
     Pop redshift from the stack
   */
  if (-1 == SLang_pop_double(&value, NULL, NULL)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping redshift");
    return;
  }
  redshift = (float)value;
  /* 
     Pop abundances from the stack
   */
  if (-1 == SLang_pop_array_of_type(&sl_abundances,SLANG_DOUBLE_TYPE)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping abundances pointer");
    return;
  }
  /* 
     Pop gamma, ampl, density, log, Tmax, Tmin, nT from the stack
   */
  if ((-1 == SLang_pop_double(&dbl_density, NULL, NULL)) || 
      (-1 == SLang_pop_double(&dbl_ampl, NULL, NULL)) || 
      (-1 == SLang_pop_double(&dbl_Tbreak, NULL, NULL)) ||
      (-1 == SLang_pop_double(&dbl_alpha, NULL, NULL)) ||
      (-1 == SLang_pop_double(&dbl_beta, NULL, NULL)) ||
      (-1 == SLang_pop_double(&dbl_gamma, NULL, NULL)) ||
      (-1 == SLang_pop_integer(&log)) ||
      (-1 == SLang_pop_double(&dbl_Tmax, NULL, NULL)) ||
      (-1 == SLang_pop_double(&dbl_Tmin, NULL, NULL)) ||
      (-1 == SLang_pop_integer(&nT))) { 
    SLang_verror(SL_STACK_OVERFLOW, "Error popping density, ampl, gamma, log, Tmax, Tmin, or nT");
    return;
  }
  /* convert to float */
  density = (float)dbl_density;
  ampl    = (float)dbl_ampl;
  Tbreak  = (float)dbl_Tbreak;
  alpha   = (float)dbl_alpha;
  beta    = (float)dbl_beta;
  gamma   = (float)dbl_gamma;
  Tmin    = (float)dbl_Tmin;
  Tmax    = (float)dbl_Tmax;
  /* 
     Pop linelist from the stack. If linelist is NULL than we mean to
     do all lines
   */
  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    linelist_supplied = 0;
  } else {
    if (-1 == SLang_pop_array_of_type(&sl_linelist,SLANG_LONG_TYPE)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping linelist pointer");
      return;
    }
    linelist_supplied = 1;
  }
  /* 
     Pop flag from the stack
   */
  if (-1 == SLang_pop_integer(&spectrum_type)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping spectrum_type");
    return;
  }
  /* 
     First argument, last to pop from stack: Pop model name from the stack
   */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }


  /* 
     Transfer SLang_arrays to c-arrays
  */
  nbin = sl_binlo->num_elements;
  if (nbin != sl_binhi->num_elements) {
    SLang_verror(SL_USER_ERROR,"Arrays binlo and binhi are not of the same length");
    return;
  }
  binlo    = (float *)malloc(sizeof(float)*nbin);
  binhi    = (float *)malloc(sizeof(float)*nbin);
  spectrum = (float *)malloc(sizeof(float)*nbin);
  for(i=0;i<nbin;i++) {
    SLang_get_array_element(sl_binlo, &i, &value);
    binlo[i] = (float)value;
    SLang_get_array_element(sl_binhi, &i, &value);
    binhi[i] = (float)value;
    spectrum[i] = 0.0;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nabundances = sl_abundances->num_elements;
  abundances = (float *)malloc(sizeof(float)*nabundances);
  for(i=0;i<nabundances;i++) {
    SLang_get_array_element(sl_abundances, &i, &value);
    abundances[i] = (float)value;
  }
  if (nabundances < 29) {
    SLang_verror(SL_USER_ERROR,"The abundances array must have at least 29 elements to accomodate all elements in the APED database.");
    return;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  if (linelist_supplied == 1) {
    nlinelist = sl_linelist->num_elements;
    linelist = (long *)malloc(sizeof(long)*nlinelist);
    for(i=0;i<nlinelist;i++) {
      SLang_get_array_element(sl_linelist, &i, &(linelist[i]));
    }
  } else {
    nlinelist = 0;
    linelist = NULL;
  }

  if (-1 == _em_double_power_law_(model_name,
			   gamma,beta,alpha,Tbreak,ampl,
			   nT, Tmin, Tmax, log,
			   density,abundances,
			   redshift,spectrum_type,kev,
			   nlinelist,linelist,nbin, 
			   binlo,binhi,spectrum)) {
    return;
  }

  sl_spectrum = SLang_create_array(SLANG_DOUBLE_TYPE, 0, NULL, &nbin, 1);
  if (sl_spectrum == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nbin;i++) {
    value = (double)spectrum[i];
    if (-1 == SLang_set_array_element(sl_spectrum,&i,&value)) {
      SLang_free_array(sl_spectrum);
      return;
    }
  }

  if (free_model_name == 1) SLfree(model_name);
  SLang_free_array(sl_binlo);
  SLang_free_array(sl_binhi);
  free(binlo);
  free(binhi);
  free(spectrum);
  if (linelist_supplied == 1) {
    SLang_free_array(sl_linelist);
    free(linelist);
  }

  SLang_push_array(sl_spectrum,1);

}

/*
  em_chebyshev(modelname,
	  spectrum_type,         lines/continuum/both
	  linelistname,
	  nT,
	  Tmin,
	  Tmax,
	  log,
	  ampl,
	  density,
	  abundances,
	  redshift,
	  kev,
	  binlo,
	  binhi)

*/

void em_chebyshev(void) {

  /* Inputs */
  char *model_name;
  SLang_Array_Type *sl_ampl;
  float *ampl;
  int nampl;
  int nT, log;
  float Tmin,Tmax;
  double dbl_Tmin,dbl_Tmax;
  int spectrum_type = 0;
  SLang_Array_Type *sl_linelist;
  long nlinelist;
  long *linelist;
  double dbl_density;
  float density;
  SLang_Array_Type *sl_abundances;
  int nabundances;
  float *abundances;
  float redshift;
  int kev;
  SLang_Array_Type *sl_binlo, *sl_binhi, *sl_spectrum;
  int i, nbin;
  float *binlo, *binhi, *spectrum;
  double value;
  int free_model_name=0, linelist_supplied=0;

  if (SLang_Num_Function_Args != 14) {
    SLang_verror (SL_USAGE_ERROR, "Usage: em_chebyshev()");
    return;
  }

  /* Pop binhi */
  if (-1 == SLang_pop_array_of_type(&sl_binhi, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* Pop binlo */
  if (-1 == SLang_pop_array_of_type(&sl_binlo, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* 
     Pop keV from the stack
   */
  if (-1 == SLang_pop_integer(&kev)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping kev flag");
    return;
  }
  /* 
     Pop redshift from the stack
   */
  if (-1 == SLang_pop_double(&value, NULL, NULL)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping redshift");
    return;
  }
  redshift = (float)value;
  /* 
     Pop abundances from the stack
   */
  if (-1 == SLang_pop_array_of_type(&sl_abundances,SLANG_DOUBLE_TYPE)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping abundances pointer");
    return;
  }
  /* 
     Pop density from the stack
   */
  if (-1 == SLang_pop_double(&dbl_density, NULL, NULL)) { 
    SLang_verror(SL_STACK_OVERFLOW, "Error popping density");
    return;
  }
  density = (float)dbl_density;
  /* 
     pop amplitude array 
  */
  if (-1 == SLang_pop_array_of_type(&sl_ampl, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping ampl pointer");
     return;
  }
  /* 
     Pop log, Tmax, Tmin, nT from the stack
   */
  if ((-1 == SLang_pop_integer(&log)) ||
      (-1 == SLang_pop_double(&dbl_Tmax, NULL, NULL)) ||
      (-1 == SLang_pop_double(&dbl_Tmin, NULL, NULL)) ||
      (-1 == SLang_pop_integer(&nT))) { 
    SLang_verror(SL_STACK_OVERFLOW, "Error popping density, ampl, gamma, log, Tmax, Tmin, or nT");
    return;
  }
  /* convert to float */
  Tmin    = (float)dbl_Tmin;
  Tmax    = (float)dbl_Tmax;
  /* 
     Pop linelist from the stack. If linelist is NULL than we mean to
     do all lines
   */
  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    linelist_supplied = 0;
  } else {
    if (-1 == SLang_pop_array_of_type(&sl_linelist,SLANG_LONG_TYPE)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping linelist pointer");
      return;
    }
    linelist_supplied = 1;
  }
  /* 
     Pop flag from the stack
   */
  if (-1 == SLang_pop_integer(&spectrum_type)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping spectrum_type");
    return;
  }
  /* 
     First argument, last to pop from stack: Pop model name from the stack
   */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }


  /* 
     Transfer SLang_arrays to c-arrays
  */
  nbin = sl_binlo->num_elements;
  if (nbin != sl_binhi->num_elements) {
    SLang_verror(SL_USER_ERROR,"Arrays binlo and binhi are not of the same length");
    return;
  }
  binlo    = (float *)malloc(sizeof(float)*nbin);
  binhi    = (float *)malloc(sizeof(float)*nbin);
  spectrum = (float *)malloc(sizeof(float)*nbin);
  for(i=0;i<nbin;i++) {
    SLang_get_array_element(sl_binlo, &i, &value);
    binlo[i] = (float)value;
    SLang_get_array_element(sl_binhi, &i, &value);
    binhi[i] = (float)value;
    spectrum[i] = 0.0;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nampl = sl_ampl->num_elements;
  ampl    = (float *)malloc(sizeof(float)*nampl);
  for(i=0;i<nampl;i++) {
    SLang_get_array_element(sl_ampl, &i, &value);
    ampl[i] = (float)value;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nabundances = sl_abundances->num_elements;
  abundances = (float *)malloc(sizeof(float)*nabundances);
  for(i=0;i<nabundances;i++) {
    SLang_get_array_element(sl_abundances, &i, &value);
    abundances[i] = (float)value;
  }
  if (nabundances < 29) {
    SLang_verror(SL_USER_ERROR,"The abundances array must have at least 29 elements to accomodate all elements in the APED database.");
    return;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  if (linelist_supplied == 1) {
    nlinelist = sl_linelist->num_elements;
    linelist = (long *)malloc(sizeof(long)*nlinelist);
    for(i=0;i<nlinelist;i++) {
      SLang_get_array_element(sl_linelist, &i, &(linelist[i]));
    }
  } else {
    nlinelist = 0;
    linelist = NULL;
  }

  if (-1 == _em_chebyshev_(model_name,
			   nampl,ampl,
			   nT, Tmin, Tmax, log,
			   density,abundances,
			   redshift,spectrum_type,kev,
			   nlinelist,linelist,nbin, 
			   binlo,binhi,spectrum)) {
    return;
  }

  sl_spectrum = SLang_create_array(SLANG_DOUBLE_TYPE, 0, NULL, &nbin, 1);
  if (sl_spectrum == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nbin;i++) {
    value = (double)spectrum[i];
    if (-1 == SLang_set_array_element(sl_spectrum,&i,&value)) {
      SLang_free_array(sl_spectrum);
      return;
    }
  }

  if (free_model_name == 1) SLfree(model_name);
  SLang_free_array(sl_binlo);
  SLang_free_array(sl_binhi);
  free(binlo);
  free(binhi);
  free(spectrum);
  if (linelist_supplied == 1) {
    SLang_free_array(sl_linelist);
    free(linelist);
  }

  SLang_push_array(sl_spectrum,1);

}


/*
  em_polynomial(modelname,
          spectrum_type,         lines/continuum/both
          linelistname,
          nT,
	  Tmin,
	  Tmax,
	  log,
	  ampl,
	  density,
	  abundances,
	  redshift,
	  kev,
	  binlo,
	  binhi)

*/

void em_polynomial(void) {

  /* Inputs */
  char *model_name;
  SLang_Array_Type *sl_ampl;
  float *ampl;
  int nampl;
  int nT;
  int log;
  float Tmin,Tmax;
  double dbl_Tmin,dbl_Tmax;
  int spectrum_type = 0;
  SLang_Array_Type *sl_linelist;
  long nlinelist;
  long *linelist;
  double dbl_density;
  float density;
  SLang_Array_Type *sl_abundances;
  int nabundances;
  float *abundances;
  float redshift;
  int kev;
  SLang_Array_Type *sl_binlo, *sl_binhi, *sl_spectrum;
  int i, nbin;
  float *binlo, *binhi, *spectrum;
  double value;
  int free_model_name=0, linelist_supplied=0;

  if (SLang_Num_Function_Args != 14) {
    SLang_verror (SL_USAGE_ERROR, "Usage: em_polynomial()");
    return;
  }

  /* Pop binhi */
  if (-1 == SLang_pop_array_of_type(&sl_binhi, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* Pop binlo */
  if (-1 == SLang_pop_array_of_type(&sl_binlo, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
     return;
  }
  /* 
     Pop keV from the stack
   */
  if (-1 == SLang_pop_integer(&kev)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping kev flag");
    return;
  }
  /* 
     Pop redshift from the stack
   */
  if (-1 == SLang_pop_double(&value, NULL, NULL)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping redshift");
    return;
  }
  redshift = (float)value;
  /* 
     Pop abundances from the stack
   */
  if (-1 == SLang_pop_array_of_type(&sl_abundances,SLANG_DOUBLE_TYPE)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping abundances pointer");
    return;
  }
  /* 
     Pop density from the stack
   */
  if (-1 == SLang_pop_double(&dbl_density, NULL, NULL)) { 
    SLang_verror(SL_STACK_OVERFLOW, "Error popping density");
    return;
  }
  density = (float)dbl_density;
  /* 
     pop amplitude array 
  */
  if (-1 == SLang_pop_array_of_type(&sl_ampl, SLANG_DOUBLE_TYPE)) {
     SLang_verror(SL_STACK_OVERFLOW, "Error popping ampl pointer");
     return;
  }
  /* 
     Pop log, Tmax, Tmin, nT from the stack
   */
  if ((-1 == SLang_pop_integer(&log)) ||
      (-1 == SLang_pop_double(&dbl_Tmax, NULL, NULL)) ||
      (-1 == SLang_pop_double(&dbl_Tmin, NULL, NULL)) ||
      (-1 == SLang_pop_integer(&nT))) { 
    SLang_verror(SL_STACK_OVERFLOW, "Error popping log, Tmax, Tmin, or nT");
    return;
  }
  /* convert to float */
  Tmin    = (float)dbl_Tmin;
  Tmax    = (float)dbl_Tmax;
  /* 
     Pop linelist from the stack. If linelist is NULL than we mean to
     do all lines
   */
  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    linelist_supplied = 0;
  } else {
    if (-1 == SLang_pop_array_of_type(&sl_linelist,SLANG_LONG_TYPE)) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping linelist pointer");
      return;
    }
    linelist_supplied = 1;
  }
  /* 
     Pop flag from the stack
   */
  if (-1 == SLang_pop_integer(&spectrum_type)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping spectrum_type");
    return;
  }
  /* 
     First argument, last to pop from stack: Pop model name from the stack
   */
  if (-1 == SLang_pop_string(&model_name,&free_model_name)) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping model_name");
    return;
  }


  /* 
     Transfer SLang_arrays to c-arrays
  */
  nbin = sl_binlo->num_elements;
  if (nbin != sl_binhi->num_elements) {
    SLang_verror(SL_USER_ERROR,"Arrays binlo and binhi are not of the same length");
    return;
  }
  binlo    = (float *)malloc(sizeof(float)*nbin);
  binhi    = (float *)malloc(sizeof(float)*nbin);
  spectrum = (float *)malloc(sizeof(float)*nbin);
  for(i=0;i<nbin;i++) {
    SLang_get_array_element(sl_binlo, &i, &value);
    binlo[i] = (float)value;
    SLang_get_array_element(sl_binhi, &i, &value);
    binhi[i] = (float)value;
    spectrum[i] = 0.0;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nampl = sl_ampl->num_elements;
  ampl    = (float *)malloc(sizeof(float)*nampl);
  for(i=0;i<nampl;i++) {
    SLang_get_array_element(sl_ampl, &i, &value);
    ampl[i] = (float)value;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  nabundances = sl_abundances->num_elements;
  abundances = (float *)malloc(sizeof(float)*nabundances);
  for(i=0;i<nabundances;i++) {
    SLang_get_array_element(sl_abundances, &i, &value);
    abundances[i] = (float)value;
  }
  if (nabundances < 29) {
    SLang_verror(SL_USER_ERROR,"The abundances array must have at least 29 elements to accomodate all elements in the APED database.");
    return;
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  if (linelist_supplied == 1) {
    nlinelist = sl_linelist->num_elements;
    linelist = (long *)malloc(sizeof(long)*nlinelist);
    for(i=0;i<nlinelist;i++) {
      SLang_get_array_element(sl_linelist, &i, &(linelist[i]));
    }
  } else {
    nlinelist = 0;
    linelist = NULL;
  }

  if (-1 == _em_polynomial_(model_name,
			   nampl,ampl,
			   nT, Tmin, Tmax, log,
			   density,abundances,
			   redshift,spectrum_type,kev,
			   nlinelist,linelist,nbin, 
			   binlo,binhi,spectrum)) {
    return;
  }

  sl_spectrum = SLang_create_array(SLANG_DOUBLE_TYPE, 0, NULL, &nbin, 1);
  if (sl_spectrum == NULL) {
    SLang_verror (SL_MALLOC_ERROR, "Bad position");
    return;
  }
  for (i=0;i<nbin;i++) {
    value = (double)spectrum[i];
    if (-1 == SLang_set_array_element(sl_spectrum,&i,&value)) {
      SLang_free_array(sl_spectrum);
      return;
    }
  }

  if (free_model_name == 1) SLfree(model_name);
  SLang_free_array(sl_binlo);
  SLang_free_array(sl_binhi);
  free(binlo);
  free(binhi);
  free(spectrum);
  if (linelist_supplied == 1) {
    SLang_free_array(sl_linelist);
    free(linelist);
  }

  SLang_push_array(sl_spectrum,1);

}
