#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>

#include "slang.h"

#define TRUE 1
#define FALSE 0
#define HC_KEV 12.398
#define MAXBIN 100000

static unsigned char _initialized;

int init_plasma_module(void);
int bisearch( int inc, int Nbin, double lambda, double *bin_lo,double *bin_hi);
void plasma_line(void);
void plasma_coco(void);

static SLang_Intrin_Fun_Type Plasma_Intrinsics [] = {
  MAKE_INTRINSIC_0("plasma_line", plasma_line, SLANG_VOID_TYPE),
  MAKE_INTRINSIC_0("plasma_coco", plasma_coco, SLANG_VOID_TYPE),
  SLANG_END_INTRIN_FUN_TABLE
};

int init_plasma_module(void) {
  if (_initialized) return 0;
  _initialized = 1;
  if (-1 == SLadd_intrin_fun_table (Plasma_Intrinsics, NULL)) {
    return -1;
  }
  SLang_vmessage("Plasma module installed.");
  return 0;
}

int bisearch( int inc,int Nbin,double lambda,double *bin_lo,double *bin_hi) {

  /* Function to do a bisection search for which element of the energy array*/
  /* the input energy lies in. */

  int low, high, result;

  low = 0;
  high = Nbin-1;

  while ((high-low) > 1 ) {

    result = (low + high) / 2;
    if (inc) {
      if (lambda > bin_lo[result]) {
	low = result;
      } else {
	high = result;
      }
    } else {
      if (lambda < bin_lo[result]) {
	low = result;
      } else {
	high = result;
      }
    }
  }
  result--;
  if (result < 0) result = 0;
  while ((lambda - bin_lo[result])*(lambda - bin_hi[result]) > 0) {
    result++;
    if (result > Nbin-1) SLang_verror(SL_USER_ERROR, "Can't find lambda bin");
  }
  return result;
}

void plasma_line(void) {

  /* Inputs */
  double norm;
  double lambda, emis;
  double bin_lo[MAXBIN], bin_hi[MAXBIN], spec[MAXBIN];
  int ArrSize;
  int keV, Nlines;
  SLang_Array_Type *sl_binlo, *sl_binhi;
  SLang_Array_Type *sl_emis, *sl_Lambda;
  int integrated;

  /* Outputs */
  SLang_Array_Type *sl_spec  = NULL;

  int i, Nbin, inc, pos;
  double minLam, maxLam;

  if (SLang_Num_Function_Args != 6) {
    SLang_verror (SL_USAGE_ERROR, "Usage: plasma_line()");
    return;
  }

  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    integrated = 0;
    sl_binhi = NULL;
  } else {
    integrated = 1;
    if (SLang_pop_array_of_type(&sl_binhi, SLANG_DOUBLE_TYPE) == -1 ) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
    }
  }
  /* Pull off in reverse order */
  if( (SLang_pop_array_of_type(&sl_binlo, SLANG_DOUBLE_TYPE) == -1 ) ||
      (SLang_pop_integer(&keV) == -1 ) || 
      (SLang_pop_array_of_type(&sl_emis, SLANG_DOUBLE_TYPE) == -1 ) ||
      (SLang_pop_array_of_type(&sl_Lambda, SLANG_DOUBLE_TYPE) == -1 ) ||
      (SLang_pop_double(&norm, NULL, NULL) == -1 )) {
    SLang_verror(SL_STACK_OVERFLOW, "Error popping input data");
  }

  /* Now we have the necessary data, create the spectrum array */
  Nbin = sl_binlo->num_elements;
  if (Nbin > MAXBIN) SLang_verror(SL_USER_ERROR,"Too many bins");
  for (i=0;i<Nbin;i++) spec[i] = 0.0;
  for (i=0;i<Nbin;i++) SLang_get_array_element(sl_binlo, &i, &(bin_lo[i]));
  inc = (bin_lo[0] < bin_lo[1]);
  if (integrated==1) {
    for (i=0;i<Nbin;i++) SLang_get_array_element(sl_binhi, &i, &(bin_hi[i]));
  } else {
    if (inc) {
      for (i=0;i<Nbin-1;i++) bin_hi[i] = bin_lo[i+1];
      bin_hi[Nbin-1] = bin_lo[Nbin-1] + (bin_hi[Nbin-2] - bin_lo[Nbin-2]);
    } else {
      for (i=1;i<Nbin;i++) bin_hi[i] = bin_lo[i-1];
      bin_hi[0] = bin_lo[0] + (bin_hi[1] - bin_lo[1]);
    }      
  }
  if (inc) {
    if (keV) {
      minLam = HC_KEV/bin_hi[Nbin-1];
      maxLam = HC_KEV/bin_lo[0];
    } else {
      minLam = bin_lo[0];
      maxLam = bin_hi[Nbin-1];
    }
  } else {
    if (keV) {
      minLam = HC_KEV/bin_lo[0];
      maxLam = HC_KEV/bin_hi[Nbin-1];
    } else {
      minLam = bin_hi[Nbin-1];
      maxLam = bin_lo[0];
    }
  }
  /* And calculate the results */
  Nlines = sl_Lambda->num_elements;

   for (i=0;i<Nlines;i++) {
    SLang_get_array_element(sl_Lambda, &i, &lambda);
    if ((lambda > minLam) && (lambda < maxLam)) {
      SLang_get_array_element(sl_emis, &i, &emis);
      if (keV) lambda = HC_KEV/lambda;  /* Now in correct units */
      pos = bisearch(inc, Nbin, lambda, bin_lo, bin_hi);
      if ((bin_lo[pos] > lambda)||(bin_hi[pos] < lambda)) {
	SLang_verror (SL_USER_ERROR, "Bad position"); 
      }
      spec[pos] += norm*emis;
    }
  }
  if (!integrated) for (i=0;i<Nbin;i++) spec[i]=spec[i]/(bin_hi[i]-bin_lo[i]);

  sl_spec = SLang_create_array(SLANG_DOUBLE_TYPE, 0, NULL, &Nbin, 1);
  if (sl_spec == NULL) SLang_verror (SL_MALLOC_ERROR, "Bad position");
  for (i=0;i<Nbin;i++) {
    if (SLang_set_array_element (sl_spec, &i, &(spec[i])) == -1) {
      SLang_free_array (sl_spec);
      return;
    }
  }
  SLang_push_array(sl_spec, 1);
}

void plasma_coco(void) {

  /* Inputs */
  double norm;
  SLang_Array_Type *sl_binlo, *sl_binhi;
  SLang_Array_Type *sl_pseudo, *sl_E_pseudo;
  SLang_Array_Type *sl_continuum, *sl_E_cont;
  SLang_Array_Type *sl_Flux, *sl_E;
  int N_pseudo, N_cont, N, keV, integrated;
  double bin_lo[MAXBIN], bin_hi[MAXBIN], spec[MAXBIN];

  /* Outputs */
  SLang_Array_Type *sl_spec  = NULL;

  int i, j, k, Nbin, inc;
  int pos;
  double emis, En_arr[MAXBIN], Emiss_arr[MAXBIN];
  double slope;

  if (SLang_Num_Function_Args != 10) {
    SLang_verror (SL_USAGE_ERROR, "Usage: plasma_coco()");
  }

  if (SLang_peek_at_stack() == SLANG_NULL_TYPE) {
    if (SLang_pop_null()==-1) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping NULL pointer");
    }
    integrated = 0;
    sl_binhi = NULL;
  } else {
    integrated = 1;
    if (SLang_pop_array_of_type(&sl_binhi, SLANG_DOUBLE_TYPE) == -1 ) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping binhi pointer");
    }
  }
  /* Pull off in reverse order */
  if( (SLang_pop_array_of_type(&sl_binlo, SLANG_DOUBLE_TYPE) == -1 ) ||
      (SLang_pop_integer(&keV) == -1 ) || 
      (SLang_pop_array_of_type(&sl_pseudo, SLANG_DOUBLE_TYPE) == -1 ) ||
      (SLang_pop_array_of_type(&sl_E_pseudo, SLANG_DOUBLE_TYPE) == -1 ) ||
      (SLang_pop_integer(&N_pseudo) == -1 ) ||
      (SLang_pop_array_of_type(&sl_continuum, SLANG_DOUBLE_TYPE) == -1 ) ||
      (SLang_pop_array_of_type(&sl_E_cont, SLANG_DOUBLE_TYPE) == -1 ) ||
      (SLang_pop_integer(&N_cont) == -1 ) ||
      (SLang_pop_double(&norm, NULL, NULL) == -1 )) {
      SLang_verror(SL_STACK_OVERFLOW, "Error popping input data");
  }
  /* Now we have the necessary data, create the spectrum array */
  Nbin = sl_binlo->num_elements;

  for (i=0;i<Nbin;i++) spec[i] = 0.0;
  for (i=0;i<Nbin;i++) SLang_get_array_element(sl_binlo, &i, &(bin_lo[i]));
  inc = (bin_lo[0] < bin_lo[1]);
  if (integrated==1) {
    for (i=0;i<Nbin;i++) SLang_get_array_element(sl_binhi, &i, &(bin_hi[i]));
  } else {
    if (inc) {
      for (i=0;i<Nbin-1;i++) bin_hi[i] = bin_lo[i+1];
      bin_hi[Nbin-1] = bin_lo[Nbin-1] + (bin_hi[Nbin-2] - bin_lo[Nbin-2]);
    } else {
      for (i=1;i<Nbin;i++) bin_hi[i] = bin_lo[i-1];
      bin_hi[0] = bin_lo[0] + (bin_hi[1] - bin_lo[1]);
    }      
  }

  if ((N_cont > MAXBIN)||(N_pseudo > MAXBIN)) {
    SLang_vmessage("Number of bins in continuum (%d) or pseudo (%d) too large (> %d)",
		    N_cont, N_pseudo, MAXBIN);
  }

  /* And calculate the spectrum, beginning with the continuum */
  for (j=0;j<2;j++) {
    if (j==0) {
      sl_E = sl_E_cont;
      sl_Flux = sl_continuum;
      N = N_cont;
    } else {
      sl_E = sl_E_pseudo;
      sl_Flux = sl_pseudo;
      N = N_pseudo;
    }

    for (i=0;i<N;i++) { /* Note:APEC data ALWAYS in increasing energy order */
      if (keV) {   /* Units of ph cm^3 /s /keV */
	SLang_get_array_element(sl_E, &i, &(En_arr[i]));
	SLang_get_array_element(sl_Flux, &i, &(Emiss_arr[i]));
      } else { /* increasing, but in wavelength units */
	SLang_get_array_element(sl_E, &i, &(En_arr[(N-1)-i]));
	SLang_get_array_element(sl_Flux, &i, &(Emiss_arr[(N-1)-i]));
	En_arr[(N-1)-i] = HC_KEV/En_arr[(N-1)-i];
	Emiss_arr[(N-1)-i] *= HC_KEV/(En_arr[(N-1)-i]*En_arr[(N-1)-i]);
      }
    }
    
    pos = 0;
    for (k=0;k<Nbin;k++) {
      if (inc) {i = k;} else {i = Nbin-1-k;}
      emis = 0.0;
      while ((bin_lo[i] > En_arr[pos+1])&&(pos<N-1)) pos++;
      slope=(Emiss_arr[pos+1]-Emiss_arr[pos])/(En_arr[pos+1]-En_arr[pos]);
      if (pos < N-1) {
	if (bin_hi[i] <= En_arr[pos+1]) {
	  emis = 2*Emiss_arr[pos] + slope*
	    ((bin_lo[i]-En_arr[pos]) + (bin_hi[i] - En_arr[pos]));
	  emis *= (bin_hi[i]-bin_lo[i])/2; /* integrate over average */
	} else {
	  emis = Emiss_arr[pos+1] + Emiss_arr[pos] + 
	    slope*(bin_lo[i]-En_arr[pos]);
	  emis *= (En_arr[pos+1]-bin_lo[i])/2;
	  pos++; /* Note: Slope value no longer accurate! */
	  while ((bin_hi[i] > En_arr[pos+1])&&(pos < N-1)) {
	    emis += (Emiss_arr[pos+1] + Emiss_arr[pos])*
	      (En_arr[pos+1]-En_arr[pos])/2;
	    pos++;
	  }
	  if (pos < N-1) {
	    slope=(Emiss_arr[pos+1]-Emiss_arr[pos])/
	      (En_arr[pos+1]-En_arr[pos]);
	    emis += (2*Emiss_arr[pos] + (bin_hi[i]-En_arr[pos])*slope)*
	      (bin_hi[i]-En_arr[pos])/2;
	  }
	}
      }       
      spec[i] += norm*emis;
    }
  }
  
  
  if (!integrated) for (i=0;i<Nbin;i++) spec[i]=spec[i]/(bin_hi[i]-bin_lo[i]);

  sl_spec = SLang_create_array(SLANG_DOUBLE_TYPE, 0, NULL, &Nbin, 1);
  if (sl_spec == NULL) SLang_verror (SL_MALLOC_ERROR, "Malloc error");
  for (i=0;i<Nbin;i++) {
    if (SLang_set_array_element (sl_spec, &i, &(spec[i])) == -1) {
      SLang_free_array (sl_spec);
      return;
    }
  }
  SLang_push_array(sl_spec, 1);
}

