%
%
%
% Ronnie Hoogerwerf, SAO, 01/07/2004
%
use_namespace("LF");

public define keV_to_K();
public define K_to_keV();
public define A_to_keV();

public define linefree() {

   %
   % Return the linefree regions of a spectrum for a certain
   % temperature, wavelength, emissivity region. The linefree
   % regions are output on the screen will be returned as a
   % linked list for further manipulation.
   %

   variable usage = "lf = linefree(T-min, T-max , Lambda-min , Lambda-max, Emissivity-min, Emissivity-max, Minimum Gap size, FWHM of lines);\nwhere  T-min            : Minimum temperature in Kelvin or keV (required parameter)\n                          Default : " + string(LF->Tmin) + " Kelvin\n       T-max            : Minimum temperature in Kelvin or keV (optional)\n                          Default : " + string(LF->Tmax) + " Kelvin\n       Lambda-min       : Minimum wavelength in Angstrom (optional)\n                          Default : " + string(LF->Lmin) + " Angstrom\n       Lambda-max       : Maximum wavelength in Angstrom (optional)\n                          Default : " + string(LF->Lmax) + " Angstrom\n       Emissivity-min   : Minimum emissivity in ph cm^3 s^-1 (optional)\n                          Default : " + string(LF->epsmin) + " ph cm^3 s^-1\n       Emissivity-max   : Maximum emissivity in ph cm^3 s^-1 (optional)\n                          Default : " + string(LF->epsmax) + " ph cm^3 s^-1\n       Minimum gap size : Minimum size in Angstrom to be included in linefree regions\n                          Default : " + string(LF->mingap) + " Angstrom (optional)\n       FWHM             : Assumed FWHM of spectral lines (optional)\n                          Default : 0.0\n";
   variable Tmin = LF->Tmin, 
            Tmax = LF->Tmax,
            Lmin = LF->Lmin,
            Lmax = LF->Lmax,
            epsmin = LF->epsmin, 
            epsmax = LF->epsmax,
            kevmin, kevmax;
   variable subset;
   variable epsmax_inrange;
   variable mingap = NULL;  % Angstrom
   variable fwhm = NULL;  % Angstrom
   
   %
   % Check number of arguments and read in values
   %
   switch (_NARGS) 
   { case 1 : Tmin = (); Tmax = Tmin; } 
   { case 2 : (Tmin, Tmax) = (); }
   { case 4 : (Tmin, Tmax, Lmin, Lmax) = (); }
   { case 5 : (Tmin, Tmax, Lmin, Lmax, epsmin) = (); }
   { case 6 : (Tmin, Tmax, Lmin, Lmax, epsmin, epsmax) = (); }
   { case 7 : (Tmin, Tmax, Lmin, Lmax, epsmin, epsmax, mingap) = (); }
   { case 8 : (Tmin, Tmax, Lmin, Lmax, epsmin, epsmax, mingap, fwhm) = (); }
   { _pop_n(_NARGS); message(usage); return;}
   

   % handle undefined limits; set them to their extremes
   if (Tmin == NULL) Tmin = LF->Tmin;
   if (Tmax == NULL) Tmax = LF->Tmax;
   if (Lmin == NULL) Lmin = LF->Lmin;
   if (Lmax == NULL) Lmax = LF->Lmax;
   if (epsmin == NULL) epsmin = LF->epsmin;
   if (epsmax == NULL) epsmax = LF->epsmax;
   if (mingap == NULL) mingap = LF->mingap;
   if (fwhm   == NULL) fwhm = 0.0;


   % Handle all stuff possibly wrong with the temperature input.
   if (Tmin > Tmax) {
      % switch temperatures
      variable dum = Tmin;
      Tmin = Tmax;
      Tmax = dum;
   }
   if (Tmin < LF->Tmin and Tmax < LF->Tmin) {
      % assume temperature is in keVs.
      Tmin = keV_to_K(Tmin);
      Tmax = keV_to_K(Tmax);
   } 
   if (Tmin < LF->Tmin) {
      % readjust Tmin to smallest possible temperature in apec models
      Tmin = LF->Tmin;
   } else if (Tmin > LF->Tmax) {
      Tmin = LF->Tmax;
   } 
   if (Tmax < LF->Tmin) {
      % readjust Tmin to smallest possible temperature in apec models
      Tmax = LF->Tmin;
   } else if (Tmax > LF->Tmax) {
      Tmax = LF->Tmax;
   }
   kevmin = K_to_keV(Tmin);
   kevmax = K_to_keV(Tmax);
       
   % Handle all stuff possibly wrong with the wavelength input.
   
   if (Lmin > Lmax) {
      % switch wavelengths
      variable ldum;
      ldum = Lmin;
      Lmin = Lmax;
      Lmax = ldum;
   }
   if (Lmin < LF->Lmin) {
      % readjust Lmin to smallest possible wavelength
      Lmin = LF->Lmin;
   }
   if (Lmax > LF->Lmax) {
      % readjust Lmin to largest possible wavelength
      Lmax = LF->Lmax;
   }

   %
   % Perform selections using the where command
   %

   % 0: Make copy of data_short
   subset = LF->data_short;

   % 1: Filter on temperature
   subset = subset[where(Tmax >= subset[*,LF->iTmin] and 
			 Tmin <= subset[*,LF->iTmax]),*];

   % 2: Filter on wavelength
   subset = subset[where(subset[*,LF->iLambda] >= Lmin and 
			 subset[*,LF->iLambda] <= Lmax),*];
   epsmax_inrange = LF_Math->max(subset[*,LF->iEpeak]);
   
   % 3: Filter on emissivity
   subset = subset[where(subset[*,LF->iEpeak] >= epsmin and 
			 subset[*,LF->iEpeak] <= epsmax),*];


   %
   % 4: now search the list for areas without spectral lines
   %
   variable wl = subset[*,LF->iLambda];
   variable nwl = length(wl);
   variable gaptype = struct{start, end, next};
   variable gap = @gaptype, next = gap, tmp;
   variable i, ngap = 0;

   % a: check from minimum edge to first line
   if (wl[0] - Lmin > mingap+fwhm) {
      % Found first gap
      next.start = Lmin;
      next.end   = wl[0]-fwhm/2.0;
      next.next  = NULL;
      ngap++;
   }
   % b: search all lines
   for (i = 1; i < nwl; i++) {
      if (wl[i] - wl[i-1] > mingap+fwhm) {
	 % Found another gap
	 if (ngap == 0) {
	    next.start = wl[i-1]+fwhm/2.0;
	    next.end   = wl[i]-fwhm/2.0;
	    next.next  = NULL;
	 } else {
	    tmp = @gaptype;
	    tmp.start = wl[i-1]+fwhm/2.0;
	    tmp.end   = wl[i]-fwhm/2.0;
	    tmp.next  = NULL;
	    next.next = tmp;
	    next = tmp;
	 }
	 ngap++;
      }
   }
   % c: check from last line to maximum edge
   if (Lmax - wl[nwl-1] > mingap+fwhm) {
      % Found last gap.
      if (ngap == 0) {
	 next.start = wl[i-1]+fwhm/2.0;
	 next.end   = wl[i]-fwhm/2.0;
	 next.next  = NULL;
      } else {
	 tmp = @gaptype;
	 tmp.start = wl[nwl-1]+fwhm/2.0;
	 tmp.end   = Lmax;
	 tmp.next  = NULL;
	 next.next = tmp;
	 next = tmp;
      }
      ngap++;
   }
   next = NULL;
   vmessage("Line free regions for %15.3f < Temperature < %15.3f K", Tmin, Tmax);
   vmessage("                          (%5.2f kev)                     (%5.2f kev)", kevmin, kevmax);
   vmessage("                      %15.3f < Wavelength  < %15.3f A", Lmin, Lmax);
   vmessage("                      %15.5e < Emissivity  < %15.5e ph cm^3 s^-1", epsmin, epsmax);
   vmessage("  Detected %d regions without spectral lines.",ngap);
   message("       in Angstrom               in keV       ");
   message("     From   :    To         From   :    To   ");
   message("  ---------------------  ---------------------");
   foreach (gap) {
      tmp = ();
      vmessage("  %10.4f:%10.4f  %10.4f:%10.4f",
	       tmp.start,tmp.end,A_to_keV(tmp.end),A_to_keV(tmp.start));

   }

 
   return gap;

}

public define Angstrom_to_keV (a) {

   %
   % Convert Angstroms to keV
   %

   return 12.3984 / a;

}

public define A_to_keV (a) {
   return Global->Angstrom_to_keV(a);
}

public define keV_to_Angstrom (kev) {

   %
   % Convert keV to Angstrom
   %

   return 12.3984 / kev;

}

public define keV_to_A (kev) {
   return Global->keV_to_Angstrom(kev);
}

public define keV_to_K (kev) {
   
   return kev / 8.617342e-8;

}


public define K_to_keV (K) {

   return K * 8.617342e-8; 

}
