/** \file ApedDatabase.cxx
    \brief Implementation of ApedDatabase class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include "ApedDatabase.h"
#include "profit/Energy.h"

#include <limits>

namespace profit {

  ApedDatabase::ApedDatabase(const std::string & file_name): m_peak_emis_table(file_name, "PEAKEMIS"), m_matches() {
    // Perform an action which will throw an exception if this is not an APED database.
    // TODO: rationalize this so that one can reliably tell for ALL supported databases whether this one is APED.
    std::string ext_name = m_peak_emis_table.getKeyword("EXTNAME");
  }

  void ApedDatabase::filter(const IEnergy & min_energy, const IEnergy & max_energy) {
    // Max energy -> min wavelength and vice versa.
    m_peak_emis_table.filter("Lambda", max_energy.wavelength(), min_energy.wavelength());
  }

  void ApedDatabase::filter(double min_emissivity) {
    m_peak_emis_table.filter("Peakemissivity", min_emissivity, std::numeric_limits<double>::max());
  }

  const AtomDbImp::AtomicLineCont & ApedDatabase::getMatches() const {
    // Erase the current set of matches.
    m_matches.clear();

    InputTable::size_type num_rec = m_peak_emis_table.getNumRecords();

    // Make sure container has enough room for the matches.
    m_matches.resize(num_rec);

    // Read all data from input table.
    for (InputTable::size_type index = 0; index != num_rec; ++index) {
      Record rec = m_peak_emis_table.getRecord(index);
      m_matches[index] = AtomicLine("APED", rec.getField("lambda"), int(rec.getField("element")), int(rec.getField("ion")),
        int(rec.getField("upperlev")), int(rec.getField("lowerlev")), rec.getField("peakemissivity"));
    }

    return m_matches;
  }

}
