/** \file Application.cxx
    \brief Implementation of Application class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <iostream>
#include <limits>
#include <stdexcept>

#include "profit/Application.h"
#include "profit/ProfitApplication.h"
#include "profit/UnitTest.h"
#include "profit/UnitTestGui.h"

namespace profit {

  // Static variable definitions.
  std::string Application::s_application_name;
  std::list<std::string> Application::s_arg;
  bool Application::s_debug_mode(false);
  bool Application::s_gui_mode(true);
  bool Application::s_hit_mode(false);
  bool Application::s_test_mode(false);
  bool Application::s_failed(false);
  bool Application::s_version_mode(false);

  void Application::parseCommandLine(int argc, char ** argv) {
    std::string method("parseCommandLine(...)");
    if (0 >= argc && 0 == argv)
      error(method) << "Application::parseCommandLine: argument argc is non-positive and argv is 0" << std::endl;
    else if (0 >= argc)
      error(method) << "Application::parseCommandLine: argument argc is non-positive" << std::endl;
    else if (0 == argv)
      error(method) << "Application::parseCommandLine: argument argv is 0" << std::endl;
    else {
      // 0th argument is application name.
      s_application_name = argv[0];

      // Remove directory part of application name.
      std::string::size_type last_slash = s_application_name.find_last_of("/\\");
      if (std::string::npos != last_slash) s_application_name = s_application_name.substr(last_slash + 1);

      // Loop over remaining arguments, looking for known switches.
      for (int ii = 1; ii != argc; ++ii) {
        std::string this_arg(argv[ii]);
        if (0 == this_arg.find("-b") || 0 == this_arg.find("--batch")) s_gui_mode = false;
        else if (0 == this_arg.find("-d") || 0 == this_arg.find("--debug")) s_debug_mode = true;
        else if (0 == this_arg.find("-h") || 0 == this_arg.find("--help")) s_version_mode = true;
        else if (0 == this_arg.find("-s") || 0 == this_arg.find("--suicide")) s_hit_mode = true;
        else if (0 == this_arg.find("-t") || 0 == this_arg.find("--test")) s_test_mode = true;
        else if (0 == this_arg.find("-v") || 0 == this_arg.find("--version")) s_version_mode = true;
        else if (0 != this_arg.find("-")) s_arg.push_back(this_arg);
      }
    }
  }

  Application * Application::create() {
    std::string method("Application::create()");
    Application * application = 0;
    // Check debug mode.
    if (s_debug_mode) {
      // In debug mode, so create application outside of any try block so exceptions may be caught.
      application = createApp();
    } else {
      try {
        // Not in debug mode, so create application here and report problems.
        application = createApp();
      } catch (const std::exception & x) {
        error(method) << "Application::create() caught exception " << x.what() << std::endl;
      } catch (...) {
        error(method) << "Application::create() caught unknown exception." << std::endl;
      }
    }
    return application;
  }

  void Application::setTestMode(bool test_mode) { s_test_mode = test_mode; }

  Application::~Application() {}

  void Application::run() {
    std::string method("Application::run()");
    // Check debug mode.
    if (s_version_mode) {
      info("") << "version " << version() << std::endl;
    } else if (s_debug_mode) {
      // In debug mode, so run application outside of any try block so exceptions may be caught.
      runApp();
    } else {
      try {
        // Not in debug mode, so run application here and report problems.
        runApp();
      } catch (const std::exception & x) {
        error(method) << "Application::run() caught exception " << x.what() << std::endl;
      } catch (...) {
        error(method) << "Application::run() caught unknown exception." << std::endl;
      }
    }
  }

  bool Application::failed() const { return s_failed; }

  const std::string & Application::version() const { return m_version; }

  Application * Application::createApp() {
    Application * application = 0;
    // Check application run modes.
    if (s_test_mode && !s_gui_mode) {
      // Test mode without GUI.
      application = new UnitTest;
    } else if (s_test_mode && s_gui_mode) {
      // Test mode with GUI.
      application = new UnitTestGui;
    } else {
      // Profit application.
      application = new ProfitApplication;
    }
    return application;
  }

  std::ostream & Application::error(const std::string & method) {
    s_failed = true;
    std::cerr.precision(std::numeric_limits<double>::digits10 + 3);
    std::cerr << std::scientific;
    if (method.empty())
      return std::cerr << s_application_name << ": ERROR: ";
    return std::cerr << s_application_name << ": " << method << " ERROR: ";
  }

  std::ostream & Application::info(const std::string & method) {
    if (method.empty())
      return std::clog << s_application_name << ": INFO: ";
    return std::clog << s_application_name << ": " << method << " INFO: ";
  }

  std::ostream & Application::warn(const std::string & method) {
    if (method.empty())
      return std::clog << s_application_name << ": WARNING: ";
    return std::clog << s_application_name << ": " << method << " WARNING: ";
  }

  Application::Application(): m_version("UNKNOWN") {}
}
