/** \file AtomicLine.cxx
    \brief Implementation of AtomicLine class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <iomanip>
#include <sstream>

#include "profit/AtomicLine.h"
#include "profit/Energy.h"
#include "roman_element.h"

namespace profit {

  AtomicLine::AtomicLine(): m_database(), m_lambda(0.), m_peak_emissivity(0.), m_element(0), m_ion(0), m_upper_level(0),
    m_lower_level(0) {}

  AtomicLine::AtomicLine(const std::string & database, double lambda, int element, int ion, int upper_level, int lower_level,
    double peak_emissivity): m_database(database), m_lambda(lambda), m_peak_emissivity(peak_emissivity), m_element(element),
    m_ion(ion), m_upper_level(upper_level), m_lower_level(lower_level) {}

  double AtomicLine::getLambda() const { return m_lambda; }

  int AtomicLine::getElement() const { return m_element; }

  int AtomicLine::getIon() const { return m_ion; }

  int AtomicLine::getUpperLevel() const { return m_upper_level; }

  int AtomicLine::getLowerLevel() const { return m_lower_level; }

  double AtomicLine::getPeakEmissivity() const { return m_peak_emissivity; }

  std::string AtomicLine::getLabel(const AtomicLine & reference_line, bool wavelength_space) const {
    std::ostringstream fmt_matches;
    fmt_matches << std::left << std::showpoint;

    // Write either wavelength or energy.
    if (wavelength_space) {
      fmt_matches.precision(5);
      fmt_matches << std::setw(8) << Wavelength(m_lambda).wavelength();
    } else {
      fmt_matches.precision(4);
      fmt_matches << std::fixed << std::setw(8) << Wavelength(m_lambda).energy();
    }
    fmt_matches << std::setw(12) << getRomanName() <<
      std::right << std::setw(5) << m_upper_level << "->" << std::left << std::setw(4) << m_lower_level <<
      std::setprecision(3) << std::setw(10) << m_peak_emissivity / reference_line.getPeakEmissivity();

    return fmt_matches.str();
  }

  std::string AtomicLine::getRomanName() const {
    const char * roman_name = roman_element(m_element, m_ion);
    std::string std_string_roman_name(roman_name);
    free(const_cast<char *>(roman_name));
    return std_string_roman_name;
  }

  const std::string & AtomicLine::database() const { return m_database; }

  bool AtomicLine::operator ==(const AtomicLine & line) const {
    return m_database == line.m_database &&
      m_lambda == line.m_lambda &&
      m_peak_emissivity == line.m_peak_emissivity &&
      m_element == line.m_element &&
      m_ion == line.m_ion &&
      m_upper_level == line.m_upper_level &&
      m_lower_level == line.m_lower_level;
  }
}
