/** \file AtomicTransition.cxx
    \brief Implementation of AtomicTransition class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <iomanip>

extern "C" {
#include "apec.h"
}

#include "profit/AtomicLevel.h"
#include "profit/AtomicTransition.h"
#include "profit/Energy.h"
#include "profit/OStream.h"

namespace profit {

  AtomicTransition::AtomicTransition(): m_lower_level(), m_upper_level(), m_coll(), m_rad(), m_sat(), m_Z(0), m_rmJ(0) {}

  AtomicTransition::AtomicTransition(int Z, int rmJ, AtomicLevel lower_level, AtomicLevel upper_level):
    m_lower_level(lower_level), m_upper_level(upper_level), m_Z(Z), m_rmJ(rmJ) {}

  AtomicTransition::~AtomicTransition() {}

  void AtomicTransition::addProcess(const CollProcess & process) { m_coll.push_back(process); }

  void AtomicTransition::addProcess(const RadProcess & process) { m_rad.push_back(process); }

  void AtomicTransition::addProcess(const SatProcess & process) { m_sat.push_back(process); }

  const AtomicLevel & AtomicTransition::getLowerLevel() const { return m_lower_level; }

  const AtomicLevel & AtomicTransition::getUpperLevel() const { return m_upper_level; }

  const AtomicTransition::CollCont & AtomicTransition::getCollision() const { return m_coll; }

  const AtomicTransition::RadCont & AtomicTransition::getRadiative() const { return m_rad; }

  const AtomicTransition::SatCont & AtomicTransition::getSatellite() const { return m_sat; }

  OStream & AtomicTransition::write(OStream & os, bool wavelength_space) const {
    os << "Level #  Electronic Configuration   Energy  n  l   s  g" << std::endl;

    os << m_lower_level.description() << std::endl;
    os << m_upper_level.description() << std::endl;

    if (m_upper_level.level() < 10000) {
      // Write title describing fields.
      if (wavelength_space)
        os << "Wavelength (Error)   Trans. Rate    Osc. Strength" << std::endl;
      else
        os << "Energy     (Error)   Trans. Rate    Osc. Strength" << std::endl;

      if (!m_rad.empty()) {
        double lambda = m_rad.front()->lambda;
        double lambda_err = m_rad.front()->lambda_err;
        double A = 0.;
        for (AtomicTransition::RadCont::const_iterator itor = m_rad.begin(); itor != m_rad.end(); ++itor) {
          A += (*itor)->A;
        }

        double osc = getUpperLevel().degeneracy() * A * lambda * lambda / 6.67025168e15;
        os.precision(9);
        os << std::left << std::setw(11);
        if (wavelength_space)
          os << lambda;
        else
          os << Wavelength(lambda).energy();
        if (lambda_err > 0.) {
          os << "(" << std::setw(5) << std::scientific << std::setprecision(1);
          if (wavelength_space) {
            os << lambda_err;
          } else {
            os << Wavelength(lambda).energy() * lambda_err / lambda;
          }
          os << std::setw(2) << ")";
          os.unsetf(std::ios::scientific);
        } else {
          os << "          ";
        }
        os << std::setprecision(4) << std::setw(15) << A << osc;
      }
    } else {
      // Write title describing fields.
      if (wavelength_space)
        os << "Wavelength (Error)   DR Rate        Res. Energy" << std::endl;
      else
        os << "Energy     (Error)   DR Rate        Res. Energy" << std::endl;
      if (!m_sat.empty()) {
        for (AtomicTransition::SatCont::const_iterator itor = m_sat.begin(); itor != m_sat.end(); ++itor) {
          double lambda = (*itor)->lambda;
          double lambda_err = (*itor)->lambda_err;
          double q_exc = (*itor)->q_exc;
          double energy = (*itor)->E_excite;
          os << std::left << std::setw(11);
          if (wavelength_space)
            os << lambda;
          else
            os << Wavelength(lambda).energy();
          if (lambda_err > 0.) {
            os << "(" << std::setw(5) << std::scientific << std::setprecision(1);
            if (wavelength_space) {
              os << lambda_err << std::setw(2) << ")";
            } else {
              os << Wavelength(lambda).energy() * lambda_err / lambda;
            }
            os << std::setw(2) << ")";
            os.unsetf(std::ios::scientific);
          } else {
            os << "          ";
          }
          os << std::setprecision(4) << std::setw(15) << q_exc << energy;
        }
      }
    }
    return os;
  }

  OStream & operator <<(OStream & os, const AtomicTransition & transition) { return transition.write(os); }

}
