/** \file DataEntryPanel.cxx
    \brief Implementation of DataEntryPanel class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <stdexcept>

#include "TGClient.h"

#include "profit/DataEntryPanel.h"
#include "profit/DataEntryWidget.h"

ClassImp(profit::DataEntryPanel)

namespace profit {

  DataEntryPanel::DataEntryPanel(const std::string & title, TGFrame * parent, UInt_t w, UInt_t h): m_widget(),
    m_hints_expand(kLHintsExpandX, 5, 5), m_hints_centerX(kLHintsCenterX, 5, 5),
    m_title(title), m_parent(parent), m_transient_frame(0), m_widget_area(0), m_width(w), m_height(h), m_accept(false) {
  }

  DataEntryPanel::~DataEntryPanel() {
    handleCloseWindow();
    for (std::list<DataEntryWidget *>::reverse_iterator itor = m_widget.rbegin(); itor != m_widget.rend(); ++itor)
      delete (*itor);
  }

  DataEntryWidget * DataEntryPanel::addWidget(const std::string & label, const std::string & style) {
    DataEntryWidget * widget = new DataEntryWidget(style, label);
    if (0 == widget) throw std::runtime_error("DataEntryPanel::addWidget could not allocate DataEntryWidget");
    m_widget.push_back(widget);
    return widget;
  }

  void DataEntryPanel::run() {
    buildGui();
    m_transient_frame->MapSubwindows();
    m_transient_frame->Layout();
    m_transient_frame->MapWindow();

    gClient->WaitFor(m_transient_frame);
  }

  void DataEntryPanel::handleOK() {
    m_accept = true;
    m_transient_frame->SendCloseMessage();
  }

  void DataEntryPanel::handleCancel() {
    m_accept = false;
    m_transient_frame->SendCloseMessage();
  }

  void DataEntryPanel::handleCloseWindow() {
    for (std::list<DataEntryWidget *>::reverse_iterator itor = m_widget.rbegin(); itor != m_widget.rend(); ++itor)
      (*itor)->clearDisplay();
    delete m_cancel; m_cancel = 0;
    delete m_ok; m_ok = 0;
    delete m_horizontal_frame; m_horizontal_frame = 0;
    delete m_widget_area; m_widget_area = 0;
    // Do not delete m_transient_frame -- Root does that already.
    m_transient_frame = 0;
  }

  bool DataEntryPanel::accept() { return m_accept; }

  void DataEntryPanel::buildGui() {
    m_transient_frame = new TGTransientFrame(gClient->GetRoot(), m_parent, m_width, m_height);
    if (0 == m_transient_frame) {
      handleCloseWindow();
      throw std::runtime_error("DataEntryPanel constructor could not allocate TGTransientFrame");
    }
    m_transient_frame->SetWindowName(m_title.c_str());
    m_transient_frame->CenterOnParent();
    m_transient_frame->Connect("CloseWindow", "profit::DataEntryPanel", this, "handleCloseWindow()");

    m_widget_area = new TGVerticalFrame(m_transient_frame);
    if (0 == m_widget_area) {
      handleCloseWindow();
      throw std::runtime_error("DataEntryPanel constructor could not allocate TGVerticalFrame");
    }
    m_transient_frame->AddFrame(m_widget_area, &m_hints_expand);

    m_horizontal_frame = new TGHorizontalFrame(m_transient_frame);
    if (0 == m_horizontal_frame) {
      handleCloseWindow();
      throw std::runtime_error("DataEntryPanel constructor could not allocate TGHorizontalFrame");
    }
    m_transient_frame->AddFrame(m_horizontal_frame, &m_hints_centerX);

    m_ok = new TGTextButton(m_horizontal_frame, "OK");
    if (0 == m_ok) {
      handleCloseWindow();
      throw std::runtime_error("DataEntryPanel constructor could not allocate OK button");
    }
    m_horizontal_frame->AddFrame(m_ok, &m_hints_centerX);
    m_ok->Connect("Clicked()", "profit::DataEntryPanel", this, "handleOK()");
    m_ok->SetToolTipText("Accept current settings and return.");

    m_cancel = new TGTextButton(m_horizontal_frame, "Cancel");
    if (0 == m_cancel) {
      handleCloseWindow();
      throw std::runtime_error("DataEntryPanel constructor could not allocate Cancel button");
    }
    m_horizontal_frame->AddFrame(m_cancel, &m_hints_centerX);
    m_cancel->Connect("Clicked()", "profit::DataEntryPanel", this, "handleCancel()");
    m_cancel->SetToolTipText("Reject changes to settings and return.");

    // Create children widgets.
    for (std::list<DataEntryWidget *>::iterator itor = m_widget.begin(); itor != m_widget.end(); ++itor)
      (*itor)->buildGui(m_widget_area, 100, 100);
  }

}
