/** \file FileImport.cxx
    \brief Implementation of FileImport and related classes.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include "profit/FileImport.h"
#include "profit/Gui.h"

#include "TGButton.h"
#include "TGClient.h"
#include "TGComboBox.h"
#include "TGFrame.h"
#include "TGLabel.h"
#include "TGLayout.h"
#include "TGTextEntry.h"

#include <stdexcept>

#include <iostream>

ClassImp(profit::ExtensionInterpreter)
ClassImp(profit::FileImport)
ClassImp(profit::FileInterpreter)

namespace profit {

  QuantityInterpreter::QuantityInterpreter(TGCompositeFrame * parent): m_tracker() {
    TGCompositeFrame * data_loc_frame = new TGHorizontalFrame(parent);
    m_tracker.trackFrame("Data interpretation frame", data_loc_frame, parent, kLHintsExpandX, 3, 3, 3, 3);

    TGLabel * label = new TGLabel(data_loc_frame, "Quantity:");
    m_tracker.trackFrame("Data label", label, data_loc_frame, kLHintsCenterY);

    TGComboBox * data_combo_box = new TGComboBox(data_loc_frame, 0);
    m_tracker.trackFrame("Data popup", data_combo_box, data_loc_frame);
    data_combo_box->Resize(150, 22);

    int item_number = 0;
    data_combo_box->AddEntry("Counts", item_number++);
    data_combo_box->AddEntry("Errors", item_number++);
    data_combo_box->AddEntry("Count Rate", item_number++);
    data_combo_box->AddEntry("Flux", item_number++);
    data_combo_box->AddEntry("-------", item_number++);
    data_combo_box->AddEntry("Energy", item_number++);
    data_combo_box->AddEntry("Energy Bin Width", item_number++);
    data_combo_box->AddEntry("Energy Bin Low", item_number++);
    data_combo_box->AddEntry("Energy Bin High", item_number++);
    data_combo_box->AddEntry("-------", item_number++);
    data_combo_box->AddEntry("Wavelength", item_number++);
    data_combo_box->AddEntry("-------", item_number++);
    data_combo_box->AddEntry("Wavenumber", item_number++);

    label = new TGLabel(data_loc_frame, "Column:");
    m_tracker.trackFrame("Column label", label, data_loc_frame, kLHintsCenterY, 5);

    TGComboBox * col_combo_box = new TGComboBox(data_loc_frame, 0);
    m_tracker.trackFrame("Extension popup", col_combo_box, data_loc_frame, kLHintsNormal);
    col_combo_box->Resize(100, 22);

    item_number = 0;
    col_combo_box->AddEntry("COUNTS", item_number++);
    col_combo_box->AddEntry("E_MIN", item_number++);
    col_combo_box->AddEntry("E_MAX", item_number++);

    label = new TGLabel(data_loc_frame, "Units:");
    m_tracker.trackFrame("Units label", label, data_loc_frame, kLHintsCenterY, 5);

    TGComboBox * unit_combo_box = new TGComboBox(data_loc_frame, 0);
    m_tracker.trackFrame("Units popup", unit_combo_box, data_loc_frame, kLHintsNormal);
    unit_combo_box->Resize(100, 22);

    item_number = 0;
    unit_combo_box->AddEntry("Dimensionless", item_number++);
    unit_combo_box->AddEntry("-------", item_number++);
//    unit_combo_box->AddEntry(new TGLineLBEntry(data_loc_frame, item_number++), 0);
    unit_combo_box->AddEntry("keV", item_number++);
    unit_combo_box->AddEntry("MeV", item_number++);
    unit_combo_box->AddEntry("-------", item_number++);
    unit_combo_box->AddEntry("Counts/s", item_number++);
    unit_combo_box->AddEntry("Angstroms", item_number++);

  }

  QuantityInterpreter::~QuantityInterpreter() { m_tracker.clear(); }

  ExtensionInterpreter::ExtensionInterpreter(FileImport * import, TGCompositeFrame * parent): m_tracker(), m_interpreter(),
    m_import(import), m_interp_frame(0) {
    TGCompositeFrame * ext_frame = new TGGroupFrame(parent, "Extension interpreter");
    m_tracker.trackFrame("Extension main frame", ext_frame, parent, kLHintsExpandX, 3, 3, 3, 3);

    TGCompositeFrame * data_loc_frame = new TGHorizontalFrame(ext_frame);
    m_tracker.trackFrame("Data location frame", data_loc_frame, ext_frame, kLHintsExpandX, 3, 3, 3, 3);

    TGLabel * label = new TGLabel(data_loc_frame, "Extension:");
    m_tracker.trackFrame("Extension label", label, data_loc_frame, kLHintsCenterY);

    TGComboBox * ext_combo_box = new TGComboBox(data_loc_frame, 0);
    m_tracker.trackFrame("Extension popup", ext_combo_box, data_loc_frame);
    ext_combo_box->Resize(100, 22);

    int item_number = 0;
    ext_combo_box->AddEntry("SPECTRUM", item_number++);
    ext_combo_box->AddEntry("EBOUNDS", item_number++);

    label = new TGLabel(data_loc_frame, "Row #:");
    m_tracker.trackFrame("Spectrum number label", label, data_loc_frame, kLHintsCenterY, 5);

    TGComboBox * spec_combo_box = new TGComboBox(data_loc_frame, 0);
    m_tracker.trackFrame("Extension popup", spec_combo_box, data_loc_frame, kLHintsNormal);
    spec_combo_box->Resize(100, 22);

    item_number = 0;
    spec_combo_box->AddEntry("1", item_number++);
    spec_combo_box->AddEntry("2", item_number++);

    m_interp_frame = new TGVerticalFrame(ext_frame);
    m_tracker.trackFrame("Quantity interpreter frame", m_interp_frame, ext_frame, kLHintsExpandX);

    m_interpreter.push_back(QuantityInterpreter(m_interp_frame));

    TGButton * button = new TGTextButton(ext_frame, "Add Quantity");
    m_tracker.trackFrame("Add quantity button", button, ext_frame, kLHintsNormal, 5, 3);
    button->Connect("Clicked", "profit::ExtensionInterpreter", this, "handleAddInterpreter()");
  }

  ExtensionInterpreter::~ExtensionInterpreter() {
    while (!m_interpreter.empty()) {
      m_interpreter.pop_back();
    }
    m_tracker.clear();
  }

  void ExtensionInterpreter::handleAddInterpreter() {
    m_interpreter.push_back(QuantityInterpreter(m_interp_frame));

    m_import->updateDisplay();
  }

  FileInterpreter::FileInterpreter(FileImport * import, TGCompositeFrame * parent): m_tracker(), m_interpreter(), m_import(import),
    m_interp_frame(0) {
    TGCompositeFrame * file_frame = new TGGroupFrame(parent, "File interpreter");
    m_tracker.trackFrame("File interpreter main frame", file_frame, parent, kLHintsExpandX, 5, 5, 5, 5);

    TGCompositeFrame * name_frame = new TGHorizontalFrame(file_frame);
    m_tracker.trackFrame("File name frame", name_frame, file_frame, kLHintsExpandX, 3, 3, 3, 3);

    TGLabel * label = new TGLabel(name_frame, "File name:");
    m_tracker.trackFrame("File name label", label, name_frame);

    TGTextEntry * name_text = new TGTextEntry(name_frame, "");
    m_tracker.trackFrame("File name text entry", name_text, name_frame, kLHintsExpandX, 3, 3, 0, 0);

    TGButton * button = new TGTextButton(name_frame, "...");
    m_tracker.trackFrame("File open button", button, name_frame);

    // TODO: can drop expandY so that form doesn't expand too far?
    m_interp_frame = new TGVerticalFrame(file_frame, 10, 10);
    m_tracker.trackFrame("Extension interpreter container frame", m_interp_frame, file_frame, kLHintsExpandX | kLHintsExpandY);

    m_interpreter.push_back(new ExtensionInterpreter(m_import, m_interp_frame));

    button = new TGTextButton(file_frame, "Add Extension");
    m_tracker.trackFrame("Add extension button", button, file_frame, kLHintsNormal, 5);
    button->Connect("Clicked", "profit::FileInterpreter", this, "handleAddInterpreter()");
  }

  // Not strictly necessary, but it is important that m_interpreter be cleared *before* m_tracker.
  FileInterpreter::~FileInterpreter() {
    while (!m_interpreter.empty()) {
      ExtensionInterpreter * interpreter = m_interpreter.back();
      m_interpreter.pop_back();
      delete interpreter;
    }
    m_tracker.clear();
  }

  void FileInterpreter::handleAddInterpreter() {
    m_interpreter.push_back(new ExtensionInterpreter(m_import, m_interp_frame));
    m_import->updateDisplay();
  }

  FileImport::FileImport(Gui * gui): m_tracker(), m_interpreter(), m_gui(gui), m_transient_frame(0), m_interp_frame(0) {
  }

  FileImport::~FileImport() {
    while (!m_interpreter.empty()) {
      FileInterpreter * interpreter = m_interpreter.back();
      m_interpreter.pop_back();
      delete interpreter;
    }
    m_tracker.clear();
    delete m_transient_frame;
  }

  void FileImport::runInTestMode() {
    gClient->WaitFor(buildGui());
  }

  void FileImport::handleCloseWindow() {
    m_interpreter.clear();
    m_tracker.clear();
    m_interp_frame = 0;
    m_transient_frame = 0;
  }

  void FileImport::handleAddInterpreter() {
    m_interpreter.push_back(new FileInterpreter(this, m_interp_frame));
    updateDisplay();
  }

  void FileImport::updateDisplay() {
    m_interp_frame->MapSubwindows();
    m_interp_frame->Layout();
    m_interp_frame->MapWindow();
  }

  TGTransientFrame * FileImport::buildGui() {
    m_tracker.clear();

    TGTransientFrame * transient_frame = new TGTransientFrame(gClient->GetRoot(), m_gui->getRootFrame(), 600, 400);
    if (0 == transient_frame) throw std::runtime_error("FileImport::runInTestMode() could not create file import dialog box");
    transient_frame->Connect("CloseWindow", "profit::FileImport", this, "handleCloseWindow()");
    transient_frame->SetMinWidth(300);

    TGCompositeFrame * cf = new TGHorizontalFrame(transient_frame, 600, 22);
    m_tracker.trackFrame("Button frame", cf, 0, kLHintsExpandX);
    transient_frame->AddFrame(cf, m_tracker.back().getHints());

    TGButton * button = new TGTextButton(cf, "OK");
    m_tracker.trackFrame("OK button", button, cf, kLHintsNormal, 5, 0, 5, 0);

    button = new TGTextButton(cf, "Cancel");
    m_tracker.trackFrame("Cancel button", button, cf, kLHintsNormal, 0, 0, 5, 0);

    m_interp_frame = new TGVerticalFrame(transient_frame, 10, 10);
    m_tracker.trackFrame("File interpreter container frame", m_interp_frame, 0, kLHintsExpandX | kLHintsExpandY);
    transient_frame->AddFrame(m_interp_frame, m_tracker.back().getHints());

    m_interpreter.push_back(new FileInterpreter(this, m_interp_frame));

    button = new TGTextButton(transient_frame, "Add File");
    m_tracker.trackFrame("Add file button", button, 0, kLHintsNormal, 5);
    transient_frame->AddFrame(button, m_tracker.back().getHints());
    button->Connect("Clicked", "profit::FileImport", this, "handleAddInterpreter()");

    transient_frame->MapSubwindows();
    transient_frame->Layout();
    transient_frame->MapWindow();

    return transient_frame;
  }

}
