/** \file Gui.cxx
    \brief Implementation of Gui class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <algorithm>
#include <cassert>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <cctype>
#include <iomanip>
#include <iostream>
#include <limits>
#include <list>
#include <memory>
#include <set>
#include <sstream>
#include <stdexcept>

#include "TApplication.h"
#include "TCanvas.h"
#include "TGButton.h"
#include "TGClient.h"
#include "TGFileDialog.h"
#include "TGFrame.h"
#include "TGLabel.h"
#include "TGListBox.h"
#include "TGMenu.h"
#include "TGMsgBox.h"
#include "TGTextView.h"
#include "TH1.h"
#include "TList.h"
#include "TRootEmbeddedCanvas.h"
#include "TStyle.h"
#include "TSystem.h"

#include "profit/AtomicLine.h"
#include "profit/AtomicTransition.h"
#include "profit/DataEntryPanel.h"
#include "profit/DataEntryWidget.h"
#include "profit/Energy.h"
#include "profit/Gui.h"
#include "profit/GuiOStream.h"
#include "profit/Os.h"
#include "profit/PlasmaDb.h"
#include "profit/SpecFile.h"
#include "profit/Spectrum.h"
#include "profit/SpectrumView.h"
#include "profit/StripChart.h"
#include "profit/Table.h"
#include "profit/TearOffMenu.h"

ClassImp(profit::Gui)

namespace {

  const size_t g_file_name_size = 1024;

}

namespace profit {

  void Gui::initRoot() {
    static bool m_done = false;
    if (!m_done) {
      // Reset all Root signal handlers so that Root's interpreter will not catch signals.
      gSystem->ResetSignal(kSigBus);
      gSystem->ResetSignal(kSigSegmentationViolation);
      gSystem->ResetSignal(kSigSystem);
      gSystem->ResetSignal(kSigPipe);
      gSystem->ResetSignal(kSigIllegalInstruction);
      gSystem->ResetSignal(kSigQuit);
      gSystem->ResetSignal(kSigInterrupt);
      gSystem->ResetSignal(kSigWindowChanged);
      gSystem->ResetSignal(kSigAlarm);
      gSystem->ResetSignal(kSigChild);
      gSystem->ResetSignal(kSigUrgent);
      gSystem->ResetSignal(kSigFloatingException);
      gSystem->ResetSignal(kSigTermination);
      gSystem->ResetSignal(kSigUser1);
      gSystem->ResetSignal(kSigUser2);

      // Create application object if it does not already exist.
      if (0 == gApplication) {
        // When creating application object, use -n switch to prevent user's .rootrc from being called.
        int argc = 2;
        char * argv[] = { "profit", "-n", 0 };
        new TApplication("profit", &argc, argv);
        if (0 == gApplication) throw std::runtime_error("Gui::initRoot: failed to allocate TApplication");
      }

      // Indicate this method was already run to prevent its being called again.
      m_done = true;

      // Make sure the graphical environment is OK. Otherwise, all GUI operations will fail.
      if (0 == gClient) throw std::runtime_error("Gui::initRoot failed to initialize graphical environment");

      // Tweak style.
      gStyle->SetPaperSize(TStyle::kUSLetter);
      gStyle->SetFillStyle(4000);
    }
  }

  /// \brief Construct the GUI singleton.
  Gui * Gui::instance() {
    static Gui * s_gui = new Gui;
    return s_gui;
  }

  // TODO: Remove the following comment when multiple simultaneous databases supported.
  // NOTE: For now, m_use_xstar must always be !m_use_aped, since we don't support both at once!
  Gui::Gui(): m_frame_cont(), m_view_cont(), m_stored_view(), m_spectrum_index(), m_spec_file_cont(), m_matches(),
    m_version(), m_atom_db_file_name(), m_xstar_db_file_name(), m_spec_file_name(), m_spec_dir_name(),
    m_lh_pad(kLHintsNormal, 3, 3, 3, 3), m_lh_expand(kLHintsExpandX | kLHintsExpandY), m_lh_expand_x(kLHintsExpandX),
    m_lh_expand_y(kLHintsExpandY), m_lh_expand_pad(kLHintsExpandX | kLHintsExpandY, 3, 3, 3, 3),
    m_lh_expand_x_pad(kLHintsExpandX, 3, 3, 3, 3), m_emissivity_threshold(1.e-18), m_src_redshift(0.), m_min_velocity(-3000.),
    m_max_velocity(3000.), m_number_plots(5), m_spec_menu_size(0), m_main_frame(0), m_spectrum_summary(0), m_wavelength_button(0),
    m_energy_button(0), m_velocity_button(0), m_atomic_info(0), m_small_canvas(0), m_large_canvas(0), m_text_output(0),
    m_line_box(0), m_store_canvas(0), m_store(0), m_spec_file(0), m_spectrum(0), m_thumbnail_view(0), m_main_view(0),
    m_main_output(0), m_file_menu(0), m_spectrum_menu(0), m_use_aped(true), m_use_xstar(false), m_wavelength_space(true),
    m_const_velocity_bounds(true) {

    // Use ATOMDB to find APED line list.
    // TODO: Add check for accessibility of apec_linelist.fits.
    m_atom_db_file_name = Os::joinPathSegment(Os::getEnv("ATOMDB"), "apec_linelist.fits");

    // Use XSTARDB to find xstar_linelist.dat.
    // TODO: Add check for accessibility of xstar_linelist.fits.
    m_xstar_db_file_name = Os::joinPathSegment(Os::getEnv("XSTARDB"), "xstar_linelist.dat");

    // Global Root initialization.
    Gui::initRoot();

    try {
      // Create and store top-level main frame.
      m_main_frame = new TGMainFrame(gClient->GetRoot(), 675, 600);
      trackFrame("main frame", m_main_frame);

      // Connect signals generated by/for main frame.
      m_main_frame->Connect("CloseWindow()", "profit::Gui", this, "handleCloseWindow()");

      // Display application name at top of main frame.
      m_main_frame->SetWindowName("profit");

      // Construct menu.
      constructMenu();

      // Upper horizontal frame, with spectrum summary box on left and spectrum thumbnail on right.
      ////////////////////////////////////////////////////////////////////////////////////////////////////
      TGHorizontalFrame * hf = new TGHorizontalFrame(m_main_frame, 675, 140);
      trackFrame("upper horizontal frame", hf);
      m_main_frame->AddFrame(hf, &m_lh_expand_x);

      // Spectrum summary text view box, upper left.
      m_spectrum_summary = new TGTextView(hf, 424, 134);
      trackFrame("spectrum information window", m_spectrum_summary);

      // TODO: Does this leak?
      TGFont * myfont = gClient->GetFontPool()->GetFont("7x13bold", kFALSE);
      if (0 != myfont) {
        m_spectrum_summary->SetFont(myfont->GetFontStruct());
      }

      hf->AddFrame(m_spectrum_summary, &m_lh_expand_x_pad);

      // Spectrum thumbnail frame, upper right.
      TGCompositeFrame * thumbnail_frame = new TGVerticalFrame(hf, 239, 134);
      trackFrame("thumbnail frame", thumbnail_frame);
      hf->AddFrame(thumbnail_frame);

      // Spectrum thumbnail canvas, upper right.
      m_small_canvas = new StoredCanvas("small canvas", thumbnail_frame, 239, 134, false);
      trackFrame("thumbnail canvas", m_small_canvas);
      thumbnail_frame->AddFrame(m_small_canvas, &m_lh_pad);

      ////////////////////////////////////////////////////////////////////////////////////////////////////

      constructButtonBars();

      // Bottom horizontal frame, with main spectrum and output windows on left and line ids and stored
      // spectra on right.
      ////////////////////////////////////////////////////////////////////////////////////////////////////
      hf = new TGHorizontalFrame(m_main_frame, 675, 382);
      trackFrame("bottom horizontal frame", hf);
      m_main_frame->AddFrame(hf, &m_lh_expand);

      // Left-hand vertical frame, containing main spectrum view and output window.
      TGVerticalFrame * vf = new TGVerticalFrame(hf, 440, 382);
      trackFrame("bottom left vertical frame", vf);
      hf->AddFrame(vf, &m_lh_expand);

      // Spectrum main view frame, center left.
      TGVerticalFrame * large_frame = new TGVerticalFrame(vf, 440, 270);
      trackFrame("main spectrum view frame", large_frame);
      vf->AddFrame(large_frame, &m_lh_expand);

      // Spectrum main view canvas, center left.
      m_large_canvas = new ActiveCanvas("large canvas", large_frame, 440, 270);
      trackFrame("main spectrum view canvas", m_large_canvas);
      large_frame->AddFrame(m_large_canvas, &m_lh_expand_pad);

      // Output window, bottom left.
      m_text_output = new TGTextView(vf, 440, 112);
      trackFrame("text output box", m_text_output);
      if (0 != myfont) {
        m_text_output->SetFont(myfont->GetFontStruct());
      }
      vf->AddFrame(m_text_output, &m_lh_expand_x_pad);

      // Right-hand vertical frame, containing line ids and stored spectra.
      vf = new TGVerticalFrame(hf, 235, 382);
      trackFrame("bottom right vertical frame", vf);
      hf->AddFrame(vf, &m_lh_expand_y);

      // Line id box.
      m_line_box = new TGListBox(vf);
      trackFrame("line identify box", m_line_box);
      vf->AddFrame(m_line_box, &m_lh_pad);
      m_line_box->Connect("Selected(Int_t, Int_t)", "profit::Gui", this, "handleLineSelect(int, int)");
      m_line_box->Resize(235, 120);

      // Spectra storage canvas (used to produce scrollbars).
      m_store_canvas = new TGCanvas(vf, 235, 262);
      trackFrame("spectra storage canvas", m_store_canvas);
      vf->AddFrame(m_store_canvas, &m_lh_expand_pad);

      // Spectra storage frame, to contain actual spectra.
      m_store = new TGVerticalFrame(m_store_canvas->GetViewPort(), 215, 262);
      trackFrame("spectra storage frame", m_store);
      m_store_canvas->SetContainer(m_store);

      // Setup main output stream.
      m_main_output = new GuiOStream(m_text_output);
      if (0 == m_main_output) throw std::runtime_error("Failed to allocate main output stream.");

    } catch (...) {
      delete m_main_output;
      deleteAllFrames();
      throw;
    }

    // Display whole GUI.
    m_main_frame->MapSubwindows();
    m_main_frame->Layout();
    m_main_frame->MapWindow();
  }

  Gui::~Gui() {
    deleteAllViews();
    delete m_main_output;
    deleteAllFrames();
  }

  void Gui::run() {
    gApplication->Run(kTRUE);
  }

  void Gui::handleCloseWindow() {
    gApplication->Terminate(0);
  }

  class DescendingEmissivity {
    public:
      // Sort in descending order.
      bool operator ()(const AtomicLine & line1, const AtomicLine & line2) const {
        return line1.getPeakEmissivity() > line2.getPeakEmissivity();
      }
  };

  void Gui::handleLineId() {
    // Get current selected line index.
    Int_t selected = m_line_box->GetSelected();

    AtomicLine selected_line;

    // If a line is selected, save it.
    if (0 <= selected) {
      selected_line = m_matches[selected];
    }

    // Erase current set of line identities.
    clearLineId();

    if (0 != m_main_view) {
      std::list<std::string> database;
      if (m_use_aped) database.push_back(m_atom_db_file_name);
      if (m_use_xstar) database.push_back(m_xstar_db_file_name);

      for (std::list<std::string>::iterator itor = database.begin(); itor != database.end(); ++itor) {
        try {
          // Create atomic database accessor.
          AtomDb atom_db(*itor);

          // Apply emissivity cut.
          atom_db.filter(m_emissivity_threshold);

          // Filter database based on current boundaries.
          atom_db.filter(m_main_view->getMinEnergy(), m_main_view->getMaxEnergy());

          // Get matches from this database.
          AtomDb::AtomicLineCont match = atom_db.getMatches();

          // Add current matches at end of container.
          m_matches.insert(m_matches.end(), match.begin(), match.end());
        } catch (const std::exception & x) {
          getMainOutput() << x.what() << std::endl;
        }
      }

      // Sort in ascending order.
      std::sort(m_matches.begin(), m_matches.end(), DescendingEmissivity());

      // Determine in what units to display lines.
      bool wavelength_space = m_wavelength_space;
      if (0 != dynamic_cast<EnergyView *>(m_main_view)) {
        wavelength_space = false;
      } else if (0 != dynamic_cast<WavelengthView *>(m_main_view)) {
        wavelength_space = true;
      }

      // Add current matches to list box.
      Int_t entry = 0;
      for (AtomDb::AtomicLineCont::const_iterator itor = m_matches.begin(); itor != m_matches.end(); ++itor, ++entry) {
        m_line_box->AddEntry(itor->getLabel(m_matches.front(), wavelength_space).c_str(), entry);
      }

      if (0 <= selected) {
        // Find line matching previous selection in current set of identified lines.
        AtomDb::AtomicLineCont::iterator itor = std::find(m_matches.begin(), m_matches.end(), selected_line);
        if (m_matches.end() != itor) {
          // Re-select this line.
          m_line_box->Select(itor - m_matches.begin(), kTRUE);
          handleLineSelect();
        }
      }

      // Why is this needed? May be a Root 4.02 bug.
      m_line_box->Layout();
    }
  }

  void Gui::handleLineSelect(int, int) {
    if (0 != m_main_view) {
      if (kButtonDisabled == m_atomic_info->GetState()) m_atomic_info->SetState(kButtonUp, kFALSE);
      if (kButtonDisabled == m_velocity_button->GetState()) {
        if (0 != dynamic_cast<VelocityView *>(m_main_view)) {
          m_velocity_button->SetState(kButtonDown, kFALSE);
        } else {
          m_velocity_button->SetState(kButtonUp, kFALSE);
        }
      }
    }
  }

  void Gui::handleWavelengthView() {
    // Set other button states properly.
    m_energy_button->SetState(kButtonUp, kFALSE);
    if (kButtonDisabled != m_velocity_button->GetState()) m_velocity_button->SetState(kButtonUp, kFALSE);

    // Change view to wavelength domain if it isn't already.
    WavelengthView * view = dynamic_cast<WavelengthView *>(m_main_view);
    if (0 == view) {
      if (0 != m_main_view) handleSelectView(WavelengthView(*m_main_view));
      else if (0 != m_spectrum) handleSelectView(WavelengthView(m_main_view->getSpectrum(), m_src_redshift));
    }
  }

  void Gui::handleEnergyView() {
    // Set other button states properly.
    m_wavelength_button->SetState(kButtonUp, kFALSE);
    if (kButtonDisabled != m_velocity_button->GetState()) m_velocity_button->SetState(kButtonUp, kFALSE);

    // Change view to energy domain if it isn't already.
    EnergyView * view = dynamic_cast<EnergyView *>(m_main_view);
    if (0 == view) {
      if (0 != m_main_view) handleSelectView(EnergyView(*m_main_view));
      else if (0 != m_spectrum) handleSelectView(EnergyView(m_main_view->getSpectrum(), m_src_redshift));
    }
  }

  void Gui::handleVelocityView() {
    if (0 == m_main_view && 0 == m_spectrum) {
      getMainOutput() << "No spectrum open; cannot convert to velocity view." << std::endl;
      m_atomic_info->SetState(kButtonDisabled, kFALSE);
      m_velocity_button->SetState(kButtonDisabled, kFALSE);
      return;
    }
    // Get current selected line index.
    Int_t selected = m_line_box->GetSelected();

    // If no line selected, cannot go to velocity view.
    if (0 > selected) {
      // This should not ever happen, but if it does, at least display a message which is of some use to users.
      getMainOutput() << "No line selected; cannot convert to velocity view." << std::endl;
      m_atomic_info->SetState(kButtonDisabled, kFALSE);
      m_velocity_button->SetState(kButtonDisabled, kFALSE);
      return;
    }

    // Store currently selected line.
    AtomicLine selected_line = m_matches[selected];

    Wavelength lambda_0(selected_line.getLambda());

    if (0 != m_main_view) {
      handleSelectView(VelocityView(*m_main_view, lambda_0));
    } else if (0 != m_spectrum) {
      handleSelectView(VelocityView(*m_spectrum, lambda_0, m_src_redshift));
    }

    // Enforce limits from preferences.
    if (m_const_velocity_bounds) {
      VelocityView::VelocityToEnergy to_energy(lambda_0);
      if (0 != m_main_view)
        m_main_view->select(Energy(to_energy(m_max_velocity)), Energy(to_energy(m_min_velocity)));
    }

    // Re-do line identification.
    handleLineId();

    if (0 <= selected) {
      // Find line matching previous selection in current set of identified lines.
      AtomDb::AtomicLineCont::iterator itor = std::find(m_matches.begin(), m_matches.end(), selected_line);
      if (m_matches.end() != itor) {
        // Re-select this line.
        m_line_box->Select(itor - m_matches.begin(), kTRUE);
        handleLineSelect();
      }
    }

  }

  void Gui::handleFitLine() {
    if (0 != m_main_view) {
      GuiOStream & os(getMainOutput());
      try {
        m_main_view->fit();
        os.markLastLine();
        m_main_view->fitSummary(os);
        os.gotoMark();
      } catch (const std::exception & x) {
        os.clearMark();
        os << x.what() << std::endl;
      }
    }
  }

  void Gui::handleStoreView() {
    if (0 != m_main_view) {
      SpectrumView * view = m_main_view->clone();
      trackView("stored view", view);

      // Do garbage collection.
      std::list<SpectrumView *> stored_view(m_stored_view);
      for (std::list<SpectrumView *>::iterator itor = stored_view.begin(); itor != stored_view.end(); ++itor) {
        if (!(*itor)->isDisplayed()) {
          deleteView(*itor);
        }
      }
      m_stored_view.push_back(view);
      view->display(m_store);
      m_store_canvas->MapSubwindows();
      m_store_canvas->Layout();
      m_store_canvas->MapWindow();
    }
  }

  void Gui::handleAtomicInfo() {
    Int_t selected = m_line_box->GetSelected();

    // Make sure something was in the list box.
    if (0 > selected) {
      getMainOutput() << "Cannot find atomic line info because no line was selected." << std::endl;
      return;
    }

    // Get selected line.
    AtomicLine & line = m_matches[selected];

    if (line.database() == "APED") {
      try {
        // Look up transition for this line.
        AtomicTransition transition = PlasmaDb::instance().findTransition(line.getElement(), line.getIon(), line.getLowerLevel(),
          line.getUpperLevel());

        // Write a description of the transition to the output window, first adding a blank line.
        getMainOutput() << std::endl;

        // Determine in what units to display atomic info.
        bool wavelength_space = m_wavelength_space;
        if (0 != dynamic_cast<EnergyView *>(m_main_view)) {
          wavelength_space = false;
        } else if (0 != dynamic_cast<WavelengthView *>(m_main_view)) {
          wavelength_space = true;
        }

        getMainOutput().markLastLine();
        transition.write(getMainOutput(), wavelength_space);
        getMainOutput().gotoMark();

        // Add newline afterward.
        getMainOutput() << std::endl;

      } catch (const std::exception & x) {
        getMainOutput().clearMark();
        getMainOutput() << std::endl;
        getMainOutput() << "While accessing ATOMDB, caught exception:\n" << x.what() << std::endl;
      }
    } else {
      getMainOutput() << "Atomic information not available for " <<
        (line.database().empty() ? "this" : "the " + line.database()) << " database." << std::endl;
    }

  }

  void Gui::handleLabelLines() {
    if (0 != m_main_view) {
      handleLineId();

      handleClearLabels();

      // Iterate over current set of lines.
      for (AtomDb::AtomicLineCont::const_iterator itor = m_matches.begin(); itor != m_matches.end(); ++itor) {
        // Only label lines above a threshold of 2 decades in log space of emissivity.
        if (log10(itor->getPeakEmissivity() / m_matches.front().getPeakEmissivity()) < -2.) continue;

        // Add a label to the main view for each line.
        m_main_view->addLabel(Wavelength(itor->getLambda()), itor->getRomanName());
      }

      m_main_view->updateDisplay();
    }
  }

  void Gui::handleLabelPeaks() {
    if (0 != m_main_view) {
      handleLineId();

      handleClearLabels();

      if (!m_matches.empty()) {
        // Find all peak candidates.
        std::vector<Energy> peak_candidate;
        try {
          m_main_view->findPeakCandidates(peak_candidate);
        } catch (const std::exception & x) {
          getMainOutput() << x.what() << std::endl;
          return;
        }

        // A Gaussian will be used to weight line ids for each candidate.
        // Sigma for this is the average bin width * 10.
        double sigma_squared = 5. * (m_main_view->getMaxEnergy().energy() - m_main_view->getMinEnergy().energy()) /
          m_main_view->numSelectedBins();
        sigma_squared *= sigma_squared;

        // Container of pointers to lines which should be labeled. A set is used so that
        // the same line will not receive more than one label.
        std::set<AtomicLine *> label;

        // Iterate over peak candidates.
        for (size_t peak_index = 0; peak_index != peak_candidate.size(); ++peak_index) {
          // Container of scores for each line.
          std::vector<double> score(m_matches.size(), 0.);

          // Iterate over lines.
          for (size_t line_index = 0; line_index != m_matches.size(); ++line_index) {
            Wavelength lambda(m_matches[line_index].getLambda());
            double emissivity(m_matches[line_index].getPeakEmissivity());
            double arg = peak_candidate[peak_index].energy() - lambda.energy();
            score[line_index] = emissivity * exp(-arg * arg / (2. * sigma_squared));
          }

          // Find the highest-scoring line.
          std::vector<double>::iterator max_pos = std::max_element(score.begin(), score.end());

          // Add top-scorers to the container of lines to display.
          for (size_t index = 0; index != score.size(); ++index) {
            if (score[index] / *max_pos > .5)
              label.insert(&m_matches[index]);
          }
        }

        // Now iterate over the container of distinct lines and display them.
        for (std::set<AtomicLine *>::iterator itor = label.begin(); itor != label.end(); ++itor)
          m_main_view->addLabel(Wavelength((*itor)->getLambda()), (*itor)->getRomanName());
      }

      m_main_view->updateDisplay();
    }
  }

  void Gui::handleClearLabels() {
    if (0 != m_main_view) {
      m_main_view->clearLabels();
      m_main_view->updateDisplay();
    }
  }

  void Gui::handleExpandX() {
    if (0 != m_main_view) {
      // Get current limits of the main view.
      double e_min = m_main_view->getMinEnergy().energy();
      double e_max = m_main_view->getMaxEnergy().energy();

      // Find width of main view.
      double width = e_max - e_min;

      // Expand each side by 2% of the total width.
      e_min -= width * .02;
      e_max += width * .02;

      // Re-select the displayed region.
      m_main_view->select(Energy(e_min), Energy(e_max));
    }
  }

  void Gui::handleMenu(int id) {
    switch (id) {
      case eTearOff:
        TearOffMenu::display("File", this, m_file_menu, eFile);
        break;
      case ePreferences:
        handlePreferences();
        break;
      case eOpen:
        try {
          openSpecFile(m_spec_file_name, m_spec_dir_name, true);
        } catch (const std::exception & x) {
          getMainOutput() << "Could not open spectrum " << m_spec_file_name << ":\n" << x.what() << std::endl;
        }
        break;
      case ePrint:
        printMainCanvas();
        break;
      case eQuit:
        handleCloseWindow();
        break;
      case eAbout:
        handleAboutProfit();
        break;
      default:
        break;
    }
  }

  void Gui::handleSpectrumMenu(int id) {
    if (eTearOff == id) {
      TearOffMenu::display("Spectrum", this, m_spectrum_menu, eSpectrum);
      return;
    }
    if (id < 0 || (SpecFileCont::size_type) id >= m_spec_menu_size) {
      std::ostringstream os;
      os << "Attempt to select menu item " << id << ", which is out of bounds [0, " << m_spec_menu_size << "]";
      throw std::logic_error(os.str());
    }

    // Get spectrum file object.
    SpecFile * spec_file = m_spec_file_cont[id];

    // Use spectrum file to open the spectrum object.
    openSpectrum(*spec_file, m_spectrum_index[id]);
  }

  void Gui::handleSelectView(const SpectrumView & view) {
    if (m_main_view != &view) {
      bool update_summary = true;
      bool update_thumbnail = true;

      const Spectrum & new_spec(view.getSpectrum());
      const SpecFile * new_spec_file(new_spec.getSpecFile());

      if (0 != m_main_view) {
        const Spectrum & old_spec(m_main_view->getSpectrum());
        const SpecFile * old_spec_file(old_spec.getSpecFile());

        if (old_spec_file == new_spec_file) update_summary = false;
        if (&old_spec == &new_spec) update_thumbnail = false;
      }
      if (update_summary) displaySpecFileSummary(*new_spec_file);

      // Update the thumbnail.
      if (update_thumbnail) {
        deleteView(m_thumbnail_view);
        if (m_wavelength_space)
          m_thumbnail_view = new WavelengthView(new_spec, m_src_redshift);
        else
          m_thumbnail_view = new EnergyView(new_spec, m_src_redshift);
        trackView("thumbnail view", m_thumbnail_view);
        m_thumbnail_view->display(m_small_canvas);
      }

      // Clear the currently selected list of identified lines if the energy range of the new view differs from
      // the main view.
#if 0
      if (0 != m_main_view) {
        double epsilon = std::numeric_limits<float>::epsilon();
        double current_min = m_main_view->getMinEnergy().energy();
        double current_max = m_main_view->getMaxEnergy().energy();
        double new_min = view.getMinEnergy().energy();
        double new_max = view.getMaxEnergy().energy();
        if ((epsilon < fabs((current_min - new_min) / current_min)) || (epsilon < fabs((current_max - new_max) / current_max))) {
          clearLineId();
        }
      }
#endif

      deleteView(m_main_view);
      m_main_view = view.clone();
      trackView("main view", m_main_view);
      m_large_canvas->setView(m_main_view);
      m_main_view->display(m_large_canvas, kLHintsExpandX | kLHintsExpandY,
        m_large_canvas->GetWidth(), m_large_canvas->GetHeight());

      if (0 != dynamic_cast<WavelengthView *>(m_main_view)) {
        m_wavelength_button->SetState(kButtonDown);
        m_energy_button->SetState(kButtonUp);
        m_velocity_button->SetState(kButtonUp);
      } else if (0 != dynamic_cast<EnergyView *>(m_main_view)) {
        m_wavelength_button->SetState(kButtonUp);
        m_energy_button->SetState(kButtonDown);
        m_velocity_button->SetState(kButtonUp);
      } else if (0 != dynamic_cast<VelocityView *>(m_main_view)) {
        m_wavelength_button->SetState(kButtonUp);
        m_energy_button->SetState(kButtonUp);
        m_velocity_button->SetState(kButtonDown);
      }
      if (0 > m_line_box->GetSelected()) m_velocity_button->SetState(kButtonDisabled);

      // Do garbage collection.
      std::list<SpectrumView *> stored_view(m_stored_view);
      for (std::list<SpectrumView *>::iterator itor = stored_view.begin(); itor != stored_view.end(); ++itor) {
        if (!(*itor)->isDisplayed()) {
          deleteView(*itor);
        }
      }
      m_main_frame->MapSubwindows();
      m_main_frame->Layout();
      m_main_frame->MapWindow();
    }
  }

  void Gui::handleStripChart() {
    if (0 != m_main_view) {
      StripChart(m_main_view, m_main_frame, 400, 600, m_number_plots);
    }
  }

  void Gui::handleChartChecked() {
    if (m_stored_view.empty()) {
      *m_main_output << "No spectra are stored." << std::endl;
    } else {
      StripChart::view_cont_type highlighted;

      for (std::list<SpectrumView *>::iterator itor = m_stored_view.begin(); itor != m_stored_view.end(); ++itor) {
        if ((*itor)->getHighlight()) highlighted.push_back(*itor);
      }

      if (highlighted.empty()) {
        *m_main_output << "No stored spectra are highlighted." << std::endl;
      } else {
        StripChart(highlighted, m_main_frame, 400, 600);
      }
    }
  }

  void Gui::handleSmooth() {
    if (0 != m_main_view) {
      m_main_view->smooth();
    }
  }

  void Gui::handlePreferences() {
    DataEntryPanel panel("profit preferences", m_main_frame, 250, 292);

    DataEntryWidget * emissivity(panel.addWidget("Minimum Emissivity (erg cm^3/s)"));
    emissivity->set(m_emissivity_threshold);

    DataEntryWidget * src_redshift(panel.addWidget("Source Redshift"));
    src_redshift->set(m_src_redshift);

    DataEntryWidget * chart(panel.addWidget("Number of Plots in Chart"));
    chart->set(m_number_plots);

    DataEntryWidget * v_bounds(panel.addWidget("Use constant bounds for velocity view", "check button"));

    DataEntryWidget * v_min(panel.addWidget("Minimum Velocity (km/s)"));
    v_min->set(m_min_velocity);

    DataEntryWidget * v_max(panel.addWidget("Maximum Velocity (km/s)"));
    v_max->set(m_max_velocity);

    v_bounds->control(v_min);
    v_bounds->control(v_max);
    v_bounds->set(m_const_velocity_bounds ? "on" : "off");

    DataEntryWidget * use_aped(panel.addWidget("Use ATOMDB", "check button"));
    use_aped->set(m_use_aped ? "on" : "off");

    DataEntryWidget * use_xstar(panel.addWidget("Use Xstar Database", "check button"));
    use_xstar->set(m_use_xstar ? "on" : "off");

    DataEntryWidget * wavelength_space(panel.addWidget("Display abscissa in wavelength space", "check button"));
    wavelength_space->set(m_wavelength_space ? "on" : "off");

    DataEntryWidget * energy_space(panel.addWidget("Display abscissa in energy space", "check button"));
    energy_space->set(m_wavelength_space ? "off" : "on");

    // Tie buttons together so that they are mutually exclusive (this is temporary).
    use_aped->tie(use_xstar);
    use_xstar->tie(use_aped);

    // Tie buttons together so that they are mutually exclusive.
    wavelength_space->tie(energy_space);
    energy_space->tie(wavelength_space);

    bool panel_done = false;
    while (!panel_done) {
      panel.run();

      panel_done = true;
      if (panel.accept()) {
        // WARNING: Redshift must be handled first, otherwise other changes will be applied
        // even if the redshift is canceled!
        try {
          double new_redshift = 0.;
          src_redshift->get(new_redshift);
          if (m_src_redshift != new_redshift) {
            m_src_redshift = new_redshift;
            for (std::list<SpectrumView *>::iterator itor = m_view_cont.begin(); itor != m_view_cont.end(); ++itor) {
              (*itor)->applySrcRedshift(m_src_redshift);
            }
            if (0 != m_main_view) displaySpecFileSummary(*m_main_view->getSpectrum().getSpecFile());
          }
        } catch (const std::exception & x) {
          *m_main_output << x.what() << std::endl;
        }

        // The try-catch block above may have reset panel_done, signaling to return to the data entry panel.
        if (!panel_done) continue;

        // Try to get remaining data from data entry panel.
        try {
          emissivity->get(m_emissivity_threshold);
        } catch (const std::exception & x) {
          *m_main_output << x.what() << std::endl;
        }
        try {
          chart->get(m_number_plots);
        } catch (const std::exception & x) {
          *m_main_output << x.what() << std::endl;
        }
        try {
          std::string use;
          v_bounds->get(use);
          m_const_velocity_bounds = use == "on" ? true : false;
        } catch (const std::exception & x) {
          *m_main_output << x.what() << std::endl;
        }
        try {
          v_min->get(m_min_velocity);
        } catch (const std::exception & x) {
          *m_main_output << x.what() << std::endl;
        }
        try {
          v_max->get(m_max_velocity);
        } catch (const std::exception & x) {
          *m_main_output << x.what() << std::endl;
        }
        try {
          std::string use;
          use_aped->get(use);
          m_use_aped = use == "on" ? true : false;
        } catch (const std::exception & x) {
          *m_main_output << x.what() << std::endl;
        }
        try {
          std::string use;
          use_xstar->get(use);
          m_use_xstar = use == "on" ? true : false;
        } catch (const std::exception & x) {
          *m_main_output << x.what() << std::endl;
        }
        try {
          std::string use;
          wavelength_space->get(use);
          bool wavelength_space = use == "on" ? true : false;
          if (m_wavelength_space != wavelength_space) {
            m_wavelength_space = wavelength_space;
            if (0 != m_main_view) {
              if (m_main_view->isVelocity()) {
                // Do not recast velocity view.
              } else if (m_wavelength_space) {
                clearLineId();
                handleSelectView(WavelengthView(*m_main_view));
              } else {
                clearLineId();
                handleSelectView(EnergyView(*m_main_view));
              }
            }
          }
        } catch (const std::exception & x) {
          *m_main_output << x.what() << std::endl;
        }
      }
    }
  }

  void Gui::handleAboutProfit() {
    std::string m_banner;
    if (m_version.empty() || m_version == "develop") {
      m_banner = "This is Profit, development version.\n";
    } else {
      m_banner = "This is Profit, version " + m_version + ".\n";
    }
    m_banner += "For more information, please see:\n"
                "http://heasarc.gsfc.nasa.gov/docs/software/profit";
    new TGMsgBox(gClient->GetRoot(), m_main_frame, "About Profit", m_banner.c_str(), 0, kMBOk);
  }

  void Gui::openSpecFile(std::string spec_file_name, std::string spec_dir_name, bool force_dialog) {
    // Allowed file types for spectrum.
    static const char * spec_file_types[] = {
      "All files", "*.*",
      "Spectrum", "*.pha",
      "Spectrum", "*.pi",
      "Spectrum", "*.fits",
      "Xspec fakeit spectrum", "*.fak",
      "User text file spectrum", "*.txt",
      0, 0
    };

    // Use a file dialog box, if necessary, to get the name of the spectrum file.
    if (spec_file_name.empty() || force_dialog)
      spec_file_name = fileDialog(spec_file_types, spec_file_name, spec_dir_name);

    // If user hit cancel, do nothing further at this point.
    if (spec_file_name.empty()) return;

    // See if file is already open and go no further if it is.
    SpecFileCont::iterator spec_file_itor = m_spec_file_cont.begin();
    for (; spec_file_itor != m_spec_file_cont.end(); ++spec_file_itor) {
      if (spec_file_name == (*spec_file_itor)->getFileName()) {
        bool update_summary = true;
        if (0 != m_main_view) {
          const SpecFile * old_spec_file(m_main_view->getSpectrum().getSpecFile());
          if (old_spec_file == *spec_file_itor) update_summary = false;
        }
        if (update_summary) {
          displaySpecFileSummary(*(*spec_file_itor));
        }
        return;
      }
    }

    // File not already open, so open it.
    SpecFile * spec_file = new SpecFile(spec_file_name);

    // Find out how many spectra are contained in this file.
    SpecFile::size_type num_spectra = spec_file->getNumSpectra();

    // Make entries in spec file container, so that all spectra can be referred to by number.
    m_spec_file_cont.resize(m_spec_file_cont.size() + num_spectra, spec_file);

    // Assign this spectrum file pointer to the current spectrum file pointer.
    m_spec_file = spec_file;

    // Assign the name of this spectrum to the current spectrum file name.
    m_spec_file_name = spec_file_name;

    // Display the file part only of the file name.
    std::string::size_type slash_pos = m_spec_file_name.find_last_of("/\\");
    if (std::string::npos != slash_pos) {
      m_spec_dir_name = m_spec_file_name.substr(0, slash_pos);
      m_spec_file_name = m_spec_file_name.substr(slash_pos + 1);
    } else {
      m_spec_dir_name = "";
    }

    // For single spectrum files add a simple entry to the spectrum popup;
    // for multi-spectrum files, a nested popup.
    // Handle single or multiple spectra.
    if (1 == num_spectra) {
      m_spectrum_menu->AddEntry(m_spec_file_name.c_str(), m_spec_menu_size++);
      m_spectrum_index.push_back(0);

      openSpectrum(*spec_file, 0);
    } else {
      // Multiple spectrum case; set up menu for selecting spectra, then stop.
      TGPopupMenu * multi_spec = new TGPopupMenu(gClient->GetRoot(), 10, 10);
      trackFrame("spectrum nested pop-up menu", multi_spec);
      m_spectrum_menu->AddPopup(m_spec_file_name.c_str(), multi_spec);
      for (SpecFile::size_type index = 0; index != num_spectra; ++index) {
        std::ostringstream os;
        os << index + 1;
        multi_spec->AddEntry(os.str().c_str(), m_spec_menu_size++);
        // Keep track of which spectrum in the file this entry corresponds to.
        m_spectrum_index.push_back(index);
      }
      m_spectrum = 0;
      deleteView(m_main_view); m_main_view = 0;
      deleteView(m_thumbnail_view); m_thumbnail_view = 0;
      new TGMsgBox(gClient->GetRoot(), m_main_frame,
        "Choose Spectrum", "This file contains multiple spectra; select Spectrum to open using Spectrum menu", 0, kMBOk);
      displaySpecFileSummary(*spec_file);
    }
  }

  void Gui::openSpectrum(SpecFile & spec_file, int spec_id) {
    Spectrum * new_spectrum(0);
    // Try to load the spectrum using just the id.
    try {
      new_spectrum = spec_file.getSpectrum(spec_id);
    } catch (const std::exception &) {}

    std::string ebounds_file_name;
    if (0 == new_spectrum) {
      // Could not automatically find ebounds; ask the user for them.
      static const char * resp_file_types[] = {
        "All files", "*.*",
        "Response matrix file", "*.rmf",
        "Response file", "*.rsp",
        0, 0
      };
      Int_t user_click = 0;
      Int_t buttons = kMBOk | kMBCancel;
      std::string default_ebounds = spec_file.getDetectorEboundsFileName();
      if (!default_ebounds.empty()) buttons |= kMBApply;

      // Set up context-sensitive message for message box.
      std::string msg = "Please tell profit where to find columns with E_MIN and E_MAX for this spectrum.\n";
      if (!default_ebounds.empty()) msg += "Click \"Apply\" button to use the default values for this mission/instrument.\n";
      msg += "Click \"OK\" button to enter the location of these columns. Typically they are\n"
        "available in an instrument response file (often with suffix .rmf or .rsp).";

      new TGMsgBox(gClient->GetRoot(), m_main_frame, "Cannot find energy bin specification", msg.c_str(), 0, buttons, &user_click);

      if (kMBOk == user_click) {
        // Could not find ebounds, so ask user to locate it.
        ebounds_file_name = fileDialog(resp_file_types, "", m_spec_dir_name);
      } else if (kMBApply == user_click) {
        ebounds_file_name = default_ebounds;
      } else if (kMBCancel == user_click) {
        return;
      }

      // If user clicks cancel, just stop here without error.
      if (ebounds_file_name.empty()) return;

      try {
        new_spectrum = spec_file.getSpectrum(ebounds_file_name, spec_id);
      } catch (const std::exception & x) {
        getMainOutput() << "Error loading spectrum:" << std::endl;
        getMainOutput() << x.what() << std::endl;
      }
    }

    if (0 == new_spectrum) {
      std::ostringstream os;
      if (1 == spec_file.getNumSpectra()) {
        os << "Unable to open spectrum in file " << m_spec_file_name;
      } else {
        os << "Unable to open spectrum number " << spec_id << " in file " << m_spec_file_name;
      }
      getMainOutput() << os.str() << std::endl;
      return;
    }

    // Clear out current list of identified lines.
    clearLineId();

    if (m_wavelength_space)
      handleSelectView(WavelengthView(*new_spectrum, m_src_redshift));
    else
      handleSelectView(EnergyView(*new_spectrum, m_src_redshift));
  }

  void Gui::displaySpecFileSummary(const SpecFile & spec_file) {
    std::vector<std::string> summary;
    spec_file.getSummary(summary);
    m_spectrum_summary->Clear();
    for (std::vector<std::string>::iterator itor = summary.begin(); itor != summary.end(); ++itor) {
      m_spectrum_summary->AddLine(itor->c_str());
    }
    std::ostringstream os;
    os << std::setprecision(7) << "Redshift: " << m_src_redshift;
    m_spectrum_summary->AddLine(os.str().c_str());
  }

  void Gui::printMainCanvas() {
    printCanvas(m_large_canvas);
  }

  void Gui::printCanvas(const TRootEmbeddedCanvas * canvas) const {
    // Allowed file types.
    const char * file_types[] = {
      "Postscript", "*.ps",
      "Encapsulated postscript", "*.eps",
      "PDF", "*.pdf",
      "GIF", "*.gif",
      "JPeg", "*.jpg",
      "PNG", "*.png",
      "TIFF", "*.tiff",
      "XML", "*.xml",
      "Root", "*.root",
      "SVG", "*.svg",
      "XPM", "*.xpm",
      0, 0
    };

    std::string file_name = fileDialog(file_types, "", m_spec_dir_name);

    // Print into the resulting file name.
    if (!file_name.empty()) {
      canvas->GetCanvas()->SetAttFillPS(kWhite, 4000);
      canvas->GetCanvas()->SetFillStyle(4000);
      canvas->GetCanvas()->Print(file_name.c_str());
    }
  }

  void Gui::clearLineId() {
    // Empty current set of identified lines from list box.
    Int_t num_entries = m_line_box->GetNumberOfEntries();
    if (0 < num_entries) m_line_box->RemoveEntries(0, num_entries - 1);
    m_atomic_info->SetState(kButtonDisabled, kFALSE);
    m_velocity_button->SetState(kButtonDisabled, kFALSE);
#if 0
    if (0 != m_main_view) {
      if (m_main_view->isVelocity()) {
        EButtonState state = m_velocity_button->GetState();
        if (kButtonDisabled == state || kButtonUp == state) m_velocity_button->SetState(kButtonDown, kFALSE);
        m_atomic_info->SetState(kButtonUp, kFALSE);
      } else {
        m_velocity_button->SetState(kButtonDisabled, kFALSE);
        m_atomic_info->SetState(kButtonDisabled, kFALSE);
      }
    }
#endif

    // Clear out container of atomic lines.
    m_matches.clear();

    m_line_box->Layout();
  }

  GuiOStream & Gui::getMainOutput() { return *m_main_output; }

  void Gui::setVersion(const std::string & version) { m_version = version; }

  void Gui::deleteAllFrames() {
    m_frame_cont.clear();
  }

  void Gui::deleteAllViews() {
    m_thumbnail_view = 0;
    m_main_view = 0;
    // Delete all views in reverse order of creation.
    while (!m_view_cont.empty()) {
      // First get the pointer in question.
      SpectrumView * view = m_view_cont.back();
      // Remove this pointer from the container of all views.
      m_view_cont.pop_back();
      // Destroy the view.
      delete view;
    }
    m_stored_view.clear();

    m_store_canvas->Layout();
    m_large_canvas->GetCanvas()->Update();
    m_small_canvas->GetCanvas()->Update();
  }

  void Gui::trackFrame(const std::string & context, TGFrame * frame, TGCompositeFrame * parent, ULong_t hints,
    Int_t pad_left, Int_t pad_right, Int_t pad_top, Int_t pad_bottom) {
    if (0 == frame) throw std::runtime_error("Allocation of " + context + " failed");
    std::list<RootFrame>::iterator itor = std::find(m_frame_cont.begin(), m_frame_cont.end(), frame);
    if (itor == m_frame_cont.end())
      m_frame_cont.push_back(RootFrame(frame, parent, hints, pad_left, pad_right, pad_top, pad_bottom));
  }

  void Gui::deleteFrame(TGFrame * frame) {
    std::list<RootFrame>::iterator itor = std::find(m_frame_cont.begin(), m_frame_cont.end(), frame);
    if (m_frame_cont.end() != itor) m_frame_cont.erase(itor);
  }

  void Gui::trackView(const std::string & context, SpectrumView * view) {
    if (0 == view) throw std::runtime_error("Allocation of " + context + " failed");
    // Make sure view is not already tracked.
    std::list<SpectrumView *>::iterator itor = std::find(m_view_cont.begin(), m_view_cont.end(), view);
    if (m_view_cont.end() == itor) {
      m_view_cont.push_back(view);
    }
  }

  void Gui::deleteView(SpectrumView * view) {
    if (0 == view) return;

    bool delete_view = false;

    // Remove from the container of all views.
    std::list<SpectrumView *>::iterator itor = std::find(m_view_cont.begin(), m_view_cont.end(), view);
    if (m_view_cont.end() != itor) {
      m_view_cont.erase(itor);
      delete_view = true;
    }

    // Remove from the container of views in the storage area.
    itor = std::find(m_stored_view.begin(), m_stored_view.end(), view);
    if (m_stored_view.end() != itor) m_view_cont.erase(itor);

    if (delete_view) delete view;

    m_store_canvas->Layout();
    m_large_canvas->GetCanvas()->Update();
    m_small_canvas->GetCanvas()->Update();
  }

  void Gui::constructMenu() {
    TGMenuBar * menu_bar = new TGMenuBar(m_main_frame);
    trackFrame("main frame menu bar", menu_bar);
    m_main_frame->AddFrame(menu_bar);

    TGPopupMenu * file = menu_bar->AddPopup("&File");
    file->AddEntry("--------------", eTearOff);
    file->AddEntry("About", eAbout);
    file->AddEntry("Preferences", ePreferences);
    file->AddSeparator();
    file->AddEntry("&Open", eOpen);
//    file->AddEntry("&Save", eSave);
//    file->AddEntry("&Export to ASCII", eExportAscii);
    file->AddSeparator();
    file->AddEntry("&Print", ePrint);
//    file->AddEntry("P&age Setup", ePageSetup);
    file->AddSeparator();
    file->AddEntry("&Quit", eQuit);
    file->Connect("Activated(Int_t)", "profit::Gui", this, "handleMenu(int)");

    m_file_menu = file;

    m_spectrum_menu = menu_bar->AddPopup("&Spectrum");
    m_spectrum_menu->AddEntry("--------------", eTearOff);
    m_spectrum_menu->Connect("Activated(Int_t)", "profit::Gui", this, "handleSpectrumMenu(int)");

  }

  void Gui::constructButtonBars() {
    // First row of buttons.
    TGHorizontalFrame * button_group = new TGHorizontalFrame(m_main_frame, 675, 10);
    trackFrame("button frame", button_group, m_main_frame, kLHintsNormal, 5, 5);

    TGTextButton * line_id = new TGTextButton(button_group, "Line ID");
    trackFrame("Line ID button", line_id);
    button_group->AddFrame(line_id);
    line_id->Connect("Clicked()", "profit::Gui", this, "handleLineId()");
    line_id->SetToolTipText("List lines/features in main view's displayed spectrum.");

    TGTextButton * fit_line = new TGTextButton(button_group, "Fit Line");
    trackFrame("Fit Line button", fit_line);
    button_group->AddFrame(fit_line);
    fit_line->Connect("Clicked()", "profit::Gui", this, "handleFitLine()");
    fit_line->SetToolTipText("Fit displayed spectrum with line plus Gaussians model.");

    TGTextButton * store = new TGTextButton(button_group, "Store");
    trackFrame("Store button", store);
    button_group->AddFrame(store);
    store->Connect("Clicked()", "profit::Gui", this, "handleStoreView()");
    store->SetToolTipText("Store spectrum as currently displayed.");

    m_atomic_info = new TGTextButton(button_group, "Info");
    trackFrame("Atomic Info button", m_atomic_info);
    button_group->AddFrame(m_atomic_info);
    m_atomic_info->SetState(kButtonDisabled, kFALSE);
    m_atomic_info->Connect("Clicked()", "profit::Gui", this, "handleAtomicInfo()");
    m_atomic_info->SetToolTipText("Retrieve details of selected transition and print in output window.");

    TGTextButton * label_lines = new TGTextButton(button_group, "Label Lines");
    trackFrame("Label Lines button", label_lines);
    button_group->AddFrame(label_lines);
    label_lines->Connect("Clicked()", "profit::Gui", this, "handleLabelLines()");
    label_lines->SetToolTipText("Mark line/feature positions on displayed spectrum.");

    TGTextButton * label_peaks = new TGTextButton(button_group, "Label Peaks");
    trackFrame("Label Peaks button", label_peaks);
    button_group->AddFrame(label_peaks);
    label_peaks->Connect("Clicked()", "profit::Gui", this, "handleLabelPeaks()");
    label_peaks->SetToolTipText("Mark line/feature positions closest to strong peaks.");

    TGTextButton * clear_line_labels = new TGTextButton(button_group, "Clear");
    trackFrame("Clear Line Labels button", clear_line_labels);
    button_group->AddFrame(clear_line_labels);
    clear_line_labels->Connect("Clicked()", "profit::Gui", this, "handleClearLabels()");
    clear_line_labels->SetToolTipText("Remove all markers from displayed spectrum.");

    TGTextButton * expand_x = new TGTextButton(button_group, "Zoom Out");
    trackFrame("Zoom Out button", expand_x);
    button_group->AddFrame(expand_x);
    expand_x->Connect("Clicked()", "profit::Gui", this, "handleExpandX()");
    expand_x->SetToolTipText("Zoom out main view by 5%.");

    TGTextButton * chart = new TGTextButton(button_group, "Chart");
    trackFrame("Strip chart button", chart);
    button_group->AddFrame(chart);
    chart->Connect("Clicked()", "profit::Gui", this, "handleStripChart()");
    chart->SetToolTipText("Create new window containing displayed spectrum split into multiple spectra.");

    TGTextButton * button = new TGTextButton(button_group, "Chart Checked");
    button->SetToolTipText("Create a chart from checked stored spectra.");
    trackFrame("Chart checked button", button);
    button_group->AddFrame(button);
    button->Connect("Clicked()", "profit::Gui", this, "handleChartChecked()");
    button->SetToolTipText("Create new window containing all checked (selected) stored spectra.");

    button = new TGTextButton(button_group, "Smooth");
    button->SetToolTipText("Smooth main view.");
    trackFrame("Smooth button", button);
    button_group->AddFrame(button);
    button->Connect("Clicked()", "profit::Gui", this, "handleSmooth()");
    button->SetToolTipText("Apply 5 bin median smoothing to displayed spectrum.");

    // Second row of buttons.
    button_group = new TGHorizontalFrame(m_main_frame, 675, 10);
    trackFrame("button frame", button_group, m_main_frame, kLHintsNormal, 8, 8, 1);

    TGLabel * label = new TGLabel(button_group, "Select Domain: ");
    trackFrame("Domain label", label);
    button_group->AddFrame(label);

    m_wavelength_button = new TGRadioButton(button_group, "Wavelength");
    trackFrame("Wavelength button", m_wavelength_button);
    button_group->AddFrame(m_wavelength_button);
    m_wavelength_button->SetState(m_wavelength_space ? kButtonDown : kButtonUp, kFALSE);
    m_wavelength_button->Connect("Clicked()", "profit::Gui", this, "handleWavelengthView()");
    m_wavelength_button->SetToolTipText("Convert displayed spectrum into wavelength units.");

    m_energy_button = new TGRadioButton(button_group, "Energy");
    trackFrame("Energy button", m_energy_button);
    button_group->AddFrame(m_energy_button);
    m_energy_button->SetState(m_wavelength_space ? kButtonUp : kButtonDown, kFALSE);
    m_energy_button->Connect("Clicked()", "profit::Gui", this, "handleEnergyView()");
    m_energy_button->SetToolTipText("Convert displayed spectrum into energy units.");

    m_velocity_button = new TGRadioButton(button_group, "Velocity");
    trackFrame("Velocity button", m_velocity_button);
    button_group->AddFrame(m_velocity_button);
    m_velocity_button->SetState(kButtonDisabled, kFALSE);
    m_velocity_button->Connect("Clicked()", "profit::Gui", this, "handleVelocityView()");
    m_velocity_button->SetToolTipText("Convert displayed spectrum into velocity units using selected line as origin.");

  }

  std::string Gui::fileDialog(const char ** file_types, const std::string & init_file_name,
    const std::string & init_dir_name) const {
    std::string returned_file_name;
    char * dir_name = 0;
    char * file_name = 0;
    TGFileInfo file_info;
    memset(&file_info, '\0', sizeof(TGFileInfo));
    try {
      // Buffer to store directory name.
      dir_name = new char[g_file_name_size];
      if (0 == dir_name) throw std::runtime_error("Gui::fileDialog failed to allocate dir_name");
      strncpy(dir_name, init_dir_name.empty() ? "." : init_dir_name.c_str(), g_file_name_size - 1);
      dir_name[g_file_name_size - 1] = '\0';

      // Buffer to store file name.
      file_name = new char[g_file_name_size];
      if (0 == file_name) throw std::runtime_error("Gui::fileDialog failed to allocate file_name");
      strncpy(file_name, init_file_name.c_str(), g_file_name_size - 1);
      file_name[g_file_name_size - 1] = '\0';

      // File types for dialog box display.
      // Note that TGFileInfo mostly deletes strings it owns, except if the file name is changed
      // from the initial name (which is almost always is).
      file_info.fIniDir = dir_name; dir_name = 0;
      file_info.fFilename = file_name;
      file_info.fFileTypes = file_types;

      // This auto-destructs when it is finished.
      new TGFileDialog(0, 0, kFDOpen, &file_info);

      if (0 != file_info.fFilename) returned_file_name = file_info.fFilename;
    } catch (...) {
      // Overcome memory leak in Root.
      if (file_name != file_info.fFilename) delete [] file_name;
      throw;
    }
    file_info.fFilename = 0;
    file_info.fIniDir = 0;
    // Overcome memory leak in Root.
    if (file_name != file_info.fFilename) delete [] file_name;
    return returned_file_name;
  }

  ActiveCanvas::ActiveCanvas(const char * name, const TGWindow * p, UInt_t w, UInt_t h):
    TRootEmbeddedCanvas(name, p, w, h), m_view(0), m_x(0), m_button_3_down(false) {
  }

  Bool_t ActiveCanvas::HandleContainerButton(Event_t * event) {
    Bool_t status = kTRUE;
    bool root_handles_event = true;
    if (0 != event) {
      // Determine if this event is inside the plot.
      double x_min, x_max, y_min, y_max;
      fCanvas->GetRangeAxis(x_min, y_min, x_max, y_max);
      double event_x, event_y;
      fCanvas->AbsPixeltoXY(event->fX, event->fY, event_x, event_y);
      bool in_plot = x_min <= event_x && event_x <= x_max && y_min <= event_y && event_y <= y_max;

      switch (event->fCode) {
        case kButton1:
          if (in_plot) {
            // These events are used to zoom: remap them to look like they come from the X axis.
            // Add 1 pixel to make the event be slightly below the axis, and 1 more to prevent
            // round-off errors resulting from the conversion from floating point graph coordinates
            // to pixels.
            event->fY = fCanvas->YtoAbsPixel(y_min) + 2;
          }
          break;
        case kButton3:
          // Button 3 events inside the plot window are used to start drag/drop view.
          if (in_plot) {
            root_handles_event = false;
            if (kButtonPress == event->fType) {
              m_x = event->fX;
              m_button_3_down = true;
            }
          }

          // Any button 3 events stops drag/drop view, no matter where the event occurs.
          if (kButtonPress != event->fType) {
            m_button_3_down = false;
          }

          // Be careful about button 3 press events.
          if (kButtonPress == event->fType) {
            // Check which object will handle event.
            TObjLink * obj_link = 0;
            fCanvas->Pick(event->fX, event->fY, obj_link);
  
            // Disallow all events which will not be handled by the histogram (i.e. axes).
            TObject * obj = 0;
            if (0 != obj_link && 0 != (obj = obj_link->GetObject())) {
              std::string class_name = obj->ClassName();
              if (class_name != "TH1D") root_handles_event = false;
            }
          }
          break;
        default:
          break;
      }
    }
    if (root_handles_event) {
      status = TRootEmbeddedCanvas::HandleContainerButton(event);
    }
    return status;
  }

  Bool_t ActiveCanvas::HandleContainerMotion(Event_t * event) {
    // Edit events which are near X axis, but not too near Y axis.
    if (0 != event) {
      // Handle scrolling with right-click.
      if (0 != m_view && m_button_3_down) {
        m_view->scroll((event->fX - m_x) / GetWidth());
        m_x = event->fX;
        return kTRUE;
      }

      // Determine if this event is inside the frame.
      double x_min, x_max, y_min, y_max;
      fCanvas->GetRangeAxis(x_min, y_min, x_max, y_max);
      double event_x, event_y;
      fCanvas->AbsPixeltoXY(event->fX, event->fY, event_x, event_y);
      bool in_plot = x_min <= event_x && event_x <= x_max && y_min <= event_y && event_y <= y_max;

      if (in_plot) {
        // These events are used to zoom: remap them to look like they come from the X axis.
        // Add 1 pixel to make the event be slightly below the axis, and 1 more to prevent
        // round-off errors resulting from the conversion from graph coordinate to pixels.
        event->fY = fCanvas->YtoAbsPixel(y_min) + 2;
      }
    }

    return TRootEmbeddedCanvas::HandleContainerMotion(event);
  }

  void ActiveCanvas::setView(SpectrumView * view) { m_view = view; }

  StoredCanvas::StoredCanvas(const char * name, const TGWindow * p, UInt_t w, UInt_t h, bool selectable):
    TRootEmbeddedCanvas(name, p, w, h), m_view(0), m_selectable(selectable) {
      // Prevent tampering with this canvas.
      GetCanvas()->SetEditable(kFALSE);
  }

  Bool_t StoredCanvas::HandleContainerButton(Event_t * event) {
    if (!m_selectable) return kTRUE;
// TODO Can this be made to work? Double clicking produces 2 single clicks before a double click
// event leading to at best the "select-unselect" flicker and at worst ignoring
// double click (if slow refresh). Is there some way to tell the future and see that
// no double click follows? One workaround would be to use the right click for selecting,
// but we're not doing that for now because we may want right click to do something else.
//    Int_t button = event->fCode;
//    if (kButtonRelease == event->fType && kButton1 == button) {
//      m_view->setHighlight(!m_view->getHighlight());
//    }

// For now don't handle any single click, so don't deliver this event.
//    return TRootEmbeddedCanvas::HandleContainerButton(event);
    return kTRUE;
  }

  Bool_t StoredCanvas::HandleContainerDoubleClick(Event_t *) {
    if (0 != m_view) {
      Gui * gui(Gui::instance());
      gui->clearLineId();
      gui->handleSelectView(*m_view);
    }
    return kTRUE;
  }

  void StoredCanvas::setView(SpectrumView * view) { m_view = view; }
}
