/** \file GuiOStream.cxx
    \brief Implementation of GuiOStream class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include "Rtypes.h"
#include "TGScrollBar.h"
#include "TGTextView.h"
#include "TIterator.h"
#include "TList.h"

#include "profit/GuiOStream.h"

#include <cassert>
#include <iostream>

namespace profit {

  GuiOStream::GuiOStream(TGTextView * text_view): OStream(false), m_os(), m_text_view(text_view), m_scroll_bar(0), m_position(-1) {
    // Find a pointer to the scroll bar. Look at subframes of the text view.
    TIter next(text_view->GetList());
    for (TObject * obj = next(); 0 != obj && 0 == m_scroll_bar; obj = next()) {
      TGFrameElement * el = dynamic_cast<TGFrameElement *>(obj);
      if (0 != el) {
        m_scroll_bar = dynamic_cast<TGVScrollBar *>(el->fFrame);
      }
    }

    // Connect the string stream which is used for formatting the output for the text view.
    connect(m_os);
  }

  void GuiOStream::markLastLine() { m_position = m_text_view->ReturnLineCount(); }

  void GuiOStream::gotoMark() { scrollTo(m_position); clearMark(); }

  void GuiOStream::postProcess() {
    // Iterate over string contents, displaying only complete lines in the text view.
    std::string::size_type begin = 0;
    std::string::size_type end = m_os.str().find_first_of("\n");

    // Flag to indicate whether there is anything actually to write.
    bool something_to_write = std::string::npos != end;

    // Record marked position or else current position.
    Long_t line_count = m_text_view->ReturnLineCount();

    // Address a bug on Windows in Root 4.04.02. The output gets mixed up though it gets corrected
    // after a redraw. This problem goes away if we explicitly force the scrollbar to the bottom
    // before adding text to the output.
    if (something_to_write && 0 != m_scroll_bar && kTRUE == m_scroll_bar->IsMapped()) scrollTo(line_count);

    for (; std::string::npos != end; end = m_os.str().find_first_of("\n", begin)) {
      // Add the current token to the text view.
      std::string line = m_os.str().substr(begin, end - begin);
      m_text_view->AddLine(line.c_str());
      // Start looking for the next token immediately after the newline which terminated this string.
      begin = end + 1;
    }

    // Remove everthing which was just displayed from the buffer.
    if (0 != begin) m_os.str(m_os.str().substr(begin));

    // Reset the line count to point to the marked position if there *is* a marked position.
    if (0 <= m_position) line_count = m_position;

    // Scroll to the marked position.
    if (something_to_write && 0 != m_scroll_bar && kTRUE == m_scroll_bar->IsMapped()) scrollTo(line_count);
  }

  void GuiOStream::clearMark() { m_position = -1; }

  void GuiOStream::scrollTo(long position) {
    if (0 <= position && 0 != m_scroll_bar && kTRUE == m_scroll_bar->IsMapped()) m_text_view->SetVsbPosition(position);
  }
}
