/** \file Spectrum.cxx
    \brief Implementation of Spectrum class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include "profit/SpecFile.h"
#include "profit/Spectrum.h"

#include <algorithm>
#include <cctype>
#include <cmath>
#include <map>
#include <sstream>
#include <stdexcept>
#include <utility>

namespace profit {

  Spectrum::Spectrum(const SpecFile & spec_file, int spec_index): m_spec_file(&spec_file),
    m_e_min(spec_file.getInputTable()->getNumRecords()), m_e_max(spec_file.getInputTable()->getNumRecords()),
    m_counts(spec_file.getInputTable()->getNumRecords()), m_counts_error(spec_file.getInputTable()->getNumRecords()),
    m_quality(spec_file.getInputTable()->getNumRecords()), m_spec_index(spec_index) {
    load(*spec_file.getInputTable());
  }

  Spectrum::Spectrum(const SpecFile & spec_file, const InputTable & ebounds_table): m_spec_file(&spec_file),
    m_e_min(spec_file.getInputTable()->getNumRecords()), m_e_max(spec_file.getInputTable()->getNumRecords()),
    m_counts(spec_file.getInputTable()->getNumRecords()), m_counts_error(spec_file.getInputTable()->getNumRecords()),
    m_quality(spec_file.getInputTable()->getNumRecords()), m_spec_index(-1) {
    load(*spec_file.getInputTable(), ebounds_table);
  }

  Spectrum::Spectrum(const SpecFile & spec_file, const InputTable & ebounds_table, int spec_index): m_spec_file(&spec_file), 
    m_e_min(spec_file.getInputTable()->getNumRecords()), m_e_max(spec_file.getInputTable()->getNumRecords()),
    m_counts(), m_counts_error(), m_quality(), m_spec_index(spec_index) {
    load(*spec_file.getInputTable(), ebounds_table);
  }

  void Spectrum::select(double min_energy, double max_energy) {
    std::vector<double>::size_type ii = 0;
    // Iterate over minimum energies, skipping energies < min_energy.
    for (ii = 0; ii < m_e_min.size() && m_e_min[ii] < min_energy; ++ii) {}
    // ii holds the position of the first bin whose minimum energy is >= min_energy.
    std::vector<double>::size_type low_cut = ii;

    // Find last energy bin whose maximum is <= the maximum value.
    for (ii = low_cut; ii < m_e_max.size() && m_e_max[ii] <= max_energy; ++ii) {}
    // ii holds the position of one bin past the last bin whose maximum energy is <= max_energy.
    std::vector<double>::size_type high_cut = ii;

    // Apply cuts to all arrays.
    m_e_min = std::vector<double>(m_e_min.begin() + low_cut, m_e_min.begin() + high_cut);
    m_e_max = std::vector<double>(m_e_max.begin() + low_cut, m_e_max.begin() + high_cut);
    m_counts = std::vector<double>(m_counts.begin() + low_cut, m_counts.begin() + high_cut);
    m_counts_error = std::vector<double>(m_counts_error.begin() + low_cut, m_counts_error.begin() + high_cut);
    m_quality = std::vector<double>(m_quality.begin() + low_cut, m_quality.begin() + high_cut);
  }

  const std::vector<double> & Spectrum::minEnergy() const { return m_e_min; }

  const std::vector<double> & Spectrum::maxEnergy() const { return m_e_max; }

  const std::vector<double> & Spectrum::counts() const { return m_counts; }

  const std::vector<double> & Spectrum::countsError() const { return m_counts_error; }

  void Spectrum::load(const InputTable & spectrum_table) {
    // Check for existence of columns E_MIN and/or E_MAX, right in the spectrum table.
    bool spec_has_energy = false;
    try {
      spectrum_table.getFieldDim("e_min");
      spec_has_energy = true;
    } catch (const std::exception &) {}
    if (!spec_has_energy) {
      try {
        spectrum_table.getFieldDim("e_max");
        spec_has_energy = true;
      } catch (const std::exception &) {}
    }

    // If spectrum has energy column(s), create Spectrum using spectrum table for ebounds.
    if (spec_has_energy) {
      load(spectrum_table, spectrum_table);
      return;
    }

    // Next try getting EBOUNDS extension from the same file.
    try {
      // Reconstruct full file name from spectrum file name.
      std::string ebounds_file_name = spectrum_table.getDirectoryName();
      if (ebounds_file_name.empty()) ebounds_file_name = spectrum_table.getFileName();
      else ebounds_file_name += "/" + spectrum_table.getFileName();

      std::auto_ptr<InputTable> ebounds_table(new InputTable(ebounds_file_name, "EBOUNDS"));
      load(spectrum_table, *ebounds_table);
      return;
    } catch (const std::exception &) {}

    // Next try reading RESPFILE keyword and looking for that file.
    std::string ebounds_file_name;
    try {
      ebounds_file_name = spectrum_table.getKeyword("RESPFILE");
    } catch (const std::exception &) {}

    // If this is a vector-valued spectrum, see if there is a RESPFILE column.
    if (ebounds_file_name.empty() && 0 <= m_spec_index && spectrum_table.getNumRecords() > (unsigned long) m_spec_index) {
      try {
        spectrum_table.getRecord(m_spec_index).getField("RESPFILE", ebounds_file_name);
      } catch (const std::exception &) {}
    }

    if (!ebounds_file_name.empty()) {
      try {
        std::auto_ptr<InputTable> ebounds_table(new InputTable(ebounds_file_name, "EBOUNDS"));
        load(spectrum_table, *ebounds_table);
        return;
      } catch (const std::exception &) {}
      try {
        // Try looking in the same directory where the spectrum was found.
        std::string ebounds_dir_name = spectrum_table.getDirectoryName();
        if (!ebounds_dir_name.empty()) {
          ebounds_file_name = ebounds_dir_name + "/" + ebounds_file_name;
          std::auto_ptr<InputTable> ebounds_table(new InputTable(ebounds_file_name, "EBOUNDS"));
          load(spectrum_table, *ebounds_table);
          return;
        }
      } catch (const std::exception &) {}
    }

    throw std::runtime_error("Spectrum::load: Could not find EBOUNDS extension for file " + spectrum_table.getFileName());
  }

  void Spectrum::load(const InputTable & spectrum_table, const InputTable & ebounds_table) {
    if (0 == spectrum_table.getNumRecords()) {
      throw std::runtime_error("Spectrum is empty");
    } else if (1 != spectrum_table.getFieldDim("counts")) {
      if (0 > m_spec_index)
        throw std::logic_error("Spectrum::load: Attempt to open vector-valued spectrum file as a scalar-valued spectrum");
    } else { // 1 == spectrum_table.getFieldDim.
      if (0 <= m_spec_index)
        throw std::logic_error("Spectrum::load: Attempt to open scalar-valued spectrum file as a vector-valued spectrum");
    }

    // Counts are from column in spectrum table.
    loadField(spectrum_table, "counts", m_spec_index, m_counts);

    try {
      // Errors are optional.
      loadField(spectrum_table, "stat_err", m_spec_index, m_counts_error);
    } catch (const std::exception &) {
      m_counts_error.resize(m_counts.size());
      for (std::vector<double>::size_type idx = 0; idx != m_counts.size(); ++idx) {
        // Set errors on counts, using Chi-Gehrels statistic.
        m_counts_error[idx] = (1. + sqrt(m_counts[idx] + .75));
      }
    }

    try {
      // Quality flags are optional.
      loadField(spectrum_table, "quality", m_spec_index, m_quality);
    } catch (const std::exception &) {
      m_quality.resize(m_counts.size(), 0.);
    }

    // Get energy bounds from ebounds table, taking into account the contents of the spectrum file.
    loadEnergyBounds(spectrum_table, ebounds_table);

    // Check for error conditions.
    if (m_counts.size() != m_e_min.size() || m_counts.size() != m_e_max.size() ||
      m_counts.size() != m_counts_error.size() || m_counts.size() != m_quality.size()) {
      std::ostringstream os;
      os << "Array sizes don't match: counts: " << m_counts.size() <<
        ", counts error: " << m_counts_error.size() <<
        ", e_min: " << m_e_min.size() << ", e_max: " << m_e_max.size() <<
        ", quality: " << m_quality.size();
      throw std::runtime_error(os.str());
    }

    for (std::vector<double>::size_type idx = 0; idx < m_counts.size(); ++idx) {
      if (.5 < m_quality[idx]) {
        m_counts[idx] = 0.;
        m_counts_error[idx] = 0.; // Smith hack.
      }
    }

  }

  std::string Spectrum::getFileName() const { return m_spec_file->getFileName(); }

  const SpecFile * Spectrum::getSpecFile() const { return m_spec_file; }

  void Spectrum::loadEnergyBounds(const InputTable & spectrum_table, const InputTable & ebounds_table) {
    InputTable::size_type field_dim = ebounds_table.getFieldDim("e_min");
    Record record;
    if (1 == field_dim) {
      // Minimum energy bounds must be present in ebounds table, in this case as a scalar.
      ebounds_table.getColumn("e_min", m_e_min);
    } else {
      record = findBestEbounds(spectrum_table, ebounds_table);
      record.getField("e_min", m_e_min);
    }

    // Put counts into a map indexed on e_min, so that the order can be restored after e_min is sorted.
    std::map<double, double> count_map;
    std::map<double, double> count_error_map;
    std::map<double, double> quality_map;
    for (std::vector<double>::size_type idx = 0; idx < m_counts.size(); ++idx) {
      count_map.insert(std::make_pair(m_e_min[idx], m_counts[idx]));
      count_error_map.insert(std::make_pair(m_e_min[idx], m_counts_error[idx]));
      quality_map.insert(std::make_pair(m_e_min[idx], m_quality[idx]));
    }

    try {
      // Maximum energy bounds may or may not be present.
      if (1 == field_dim) {
        ebounds_table.getColumn("e_max", m_e_max);
      } else {
        record.getField("e_max", m_e_max);
      }

      // Make sure each interval individually has width, and that min < max.
      for (std::vector<double>::size_type idx = 0; idx != m_e_max.size(); ++idx) {
        if (m_e_min[idx] >= m_e_max[idx])
          throw std::runtime_error("E_MIN must be < E_MAX for all rows of spectrum file");
      }
    } catch (const std::exception &) {
      // TODO: write a warning about e_max not being found and discarding the last point in the spectrum.
    }

    // Sort minimum energies into ascending order.
    std::sort(m_e_min.begin(), m_e_min.end());

    // Restore counts to correct order.
    for (std::vector<double>::size_type idx = 0; idx < m_counts.size(); ++idx) {
      m_counts[idx] = count_map[m_e_min[idx]];
      m_counts_error[idx] = count_error_map[m_e_min[idx]];
      m_quality[idx] = quality_map[m_e_min[idx]];
    }

    if (m_e_max.empty()) {
      // No e_max field: use e_min from next row for each e_max.
      if (m_e_min.size() > 1) {
        m_e_max.assign(m_e_min.begin() + 1, m_e_min.end());
        m_e_min.resize(m_e_min.size() - 1);
        m_counts.resize(m_counts.size() - 1);
        m_counts_error.resize(m_counts_error.size() - 1);
        m_quality.resize(m_quality.size() - 1);
      } else {
        throw std::runtime_error("If E_MAX field is not present, E_MIN must contain at least 2 bins");
      }
    } else {
      // Maximum energies were found in file, so sort them into ascending order also.
      std::sort(m_e_max.begin(), m_e_max.end());
    }

    // Make sure no two intervals start or end with same value.
    for (std::vector<double>::size_type idx = 1; idx < m_e_max.size(); ++idx) {
      if (m_e_min[idx - 1] == m_e_min[idx])
        throw std::runtime_error("E_MIN values must be unique");
      if (m_e_max[idx - 1] == m_e_max[idx])
        throw std::runtime_error("E_MAX values must be unique");
    }

  }

  Record Spectrum::findBestEbounds(const InputTable & spectrum_table, const InputTable & ebounds_table) {
    Record ebounds_record;
    std::vector<std::string> field_name;

    // Get names of all columns in the ebounds.
    ebounds_table.getColumnNames(field_name);

    Record spec_rec;
    std::vector<std::string> spec_data(field_name.size());
    bool use_keywords = false;
    if (0 > m_spec_index) {
      // One spectrum in file; look for keyword values matching column names.
      use_keywords = true;
    } else if (InputTable::size_type(m_spec_index) < spectrum_table.getNumRecords()) {
      spec_rec = spectrum_table.getRecord(m_spec_index);
    } else {
      std::ostringstream os;
      os << "Unable to look up spectrum number " << m_spec_index << " in file " << spectrum_table.getFileName() <<
        ", which has " << spectrum_table.getNumRecords() << " spectra." << std::endl;
      throw std::logic_error(os.str());
    }

    // With the exception of channel, e_min, e_max, use the columns as keys to seek data in the spectrum table.
    for (std::vector<std::string>::size_type idx = 0; idx != field_name.size(); ++idx) {
      // Convert to lowercase.
      for (std::string::iterator itor = field_name[idx].begin(); itor != field_name[idx].end(); ++itor)
        *itor = std::tolower(*itor);

      // Skip columns which are part of ebounds definition.
      if (field_name[idx] == "channel" || field_name[idx] == "e_min" || field_name[idx] == "e_max") continue;

      // Extract values from keyword or column for comparison later on with corresponding values from ebounds.
      if (use_keywords) spec_data[idx] = spectrum_table.getKeyword(field_name[idx]);
      else spec_rec.getField(field_name[idx], spec_data[idx]);
    }

    InputTable::size_type rec_idx = 0;
    for (rec_idx = 0; rec_idx != ebounds_table.getNumRecords(); ++rec_idx) {
      Record record = ebounds_table.getRecord(rec_idx);
      std::vector<std::string>::size_type idx = 0;
      for (idx = 0; idx != field_name.size(); ++idx) {
        // Skip columns which are part of ebounds definition.
        if (field_name[idx] == "channel" || field_name[idx] == "e_min" || field_name[idx] == "e_max") continue;
        // See if this record matches the values of the record in spectrum file.
        std::string value;
        record.getField(field_name[idx], value);
        if (value != spec_data[idx]) break;
      }
      // Success if loop finished without finding any inequalities.
      if (idx == field_name.size()) {
        ebounds_record = record;
        break;
      }
    }
    // Failure if loop exited upon reaching the end of the table.
    if (rec_idx == ebounds_table.getNumRecords()) {
      throw std::runtime_error("Unable to find ebounds to match " + spectrum_table.getFileName() +
        " in " + ebounds_table.getFileName());
    }

    return ebounds_record;
  }

  void Spectrum::loadField(const InputTable & table, const std::string & field_name, int spec_index,
    std::vector<double> & value) const {
    if (0 > spec_index) {
      table.getColumn(field_name, value);
    } else {
      table.getRecord(spec_index).getField(field_name, value);
    }
  }

}
