/** \file TestDatum.cxx
    \brief Implementation of Datum class-specific parts of the unit test.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include "profit/DataHandler.h"
#include "profit/DataView.h"
#include "profit/Datum.h"
#include "profit/UnitTest.h"

#include <cstdlib>
#include <vector>

namespace {
    using namespace profit;

    // Test interaction between Datum and DataView.
    class TestView : public DataView {
      public:
        TestView(Datum & datum): m_datum(&datum), m_value(datum.getValue()) {
          datum.addView(*this);
        }

        virtual void update(const const_data_cont_type & data_cont) {
          if (data_cont.end() != data_cont.find(m_datum)) m_value = m_datum->getValue();
        }

        std::string getValue() const { return m_value; }

      private:
        const Datum * m_datum;
        std::string m_value;
    };

    // Test DataHandler.
    class TestHandler : public DataHandler {
      public:
        TestHandler(Datum & datum): m_datum(&datum) {}

        virtual void handle(const Message & message, const Datum & datum, const_data_cont_type & data_cont) {
          if (message == "select" && &datum == m_datum) {
            m_datum->setValue("select");
            data_cont.insert(m_datum);
          }
        }

      private:
        Datum * m_datum;
    };

}

namespace profit {

  void UnitTest::testDatum() {
    std::string method("testDatum");

    typedef std::pair<std::string, std::string> expected_datum_type;
    typedef std::vector<expected_datum_type> expected_child_type;

    class CheckDatum {
      public:
        CheckDatum(const std::string & context, const Datum & datum, const expected_datum_type & expected_datum) {
          std::string method("CheckDatum");
          // Confirm name is correct.
          if (expected_datum.first != datum.getName()) {
            error(method) << context << ": datum.getName() returned \"" <<
              datum.getName() << "\", not \"" << expected_datum.first << "\" as expected." << std::endl;
          }

          // Confirm value is correct.
          if (expected_datum.second != datum.getValue()) {
            error(method) << context << ": datum.getValue() returned \"" <<
              datum.getValue() << "\", not \"" << expected_datum.second << "\" as expected." << std::endl;
          }
        }
    };

    // Create an object representing the domain of a spectrum.
    std::string name_string("domain");
    std::string value_string("value");

    Datum domain(name_string, value_string);

    expected_datum_type expected(name_string, value_string);
    std::string description = "After Datum domain(\"" + name_string + "\", \"" + value_string + ") ";
    CheckDatum(description, domain, expected);

    // Assign new value.
    value_string = "original value";
    expected.second = value_string;
    domain.setValue(value_string);
    CheckDatum("After domain.setValue(\"" + value_string + "\")", domain, expected);

    // Insert a child.
    name_string = "minimum";
    value_string = "100 keV";
    domain.insertChild(name_string, value_string);

    expected.first = name_string;
    expected.second = value_string;
    Datum::const_iterator itor = domain.begin();

    description = "After domain.insertChild(\"" + name_string + "\", \"" + value_string + ") ";
    // Obfuscated cast is for benefit of solaris.
    if ((const_cast<const Datum &>(domain)).end() == itor) {
      error(method) << description << "domain has no children." << std::endl;
    } else {
      CheckDatum(description, itor->second, expected);
    }

    try {
      // Inserting a duplicate record should throw an exception.
      description = "Calling a second time domain.insertChild(\"" + name_string + "\", \"" + value_string + ") ";
      domain.insertChild(name_string, value_string);
      error(method) << description << " did not throw an exception." << std::endl;
    } catch (const std::exception &) {
    }

    name_string = "dimension";
    value_string = "energy";
    domain.insertChild(name_string, value_string);

    expected.first = name_string;
    expected.second = value_string;
    itor = domain.begin();

    description = "After domain.insertChild(\"" + name_string + "\", \"" + value_string + ") ";
    // Obfuscated cast is for benefit of solaris.
    if ((const_cast<const Datum &>(domain)).end() == itor) {
      error(method) << description << "domain still has no children." << std::endl;
    } else {
      CheckDatum(description, itor->second, expected);
      ++itor;
      expected.first = "minimum";
      expected.second = "100 keV";
      // Obfuscated cast is for benefit of solaris.
      if ((const_cast<const Datum &>(domain)).end() == itor) {
        error(method) << description << "domain has only one child." << std::endl;
      } else {
        CheckDatum(description, itor->second, expected);
      }
    }

    // Make sure domain doesn't yet have any affected views.
    Datum::view_cont_type affected_view;

    description = "Before creating any views, domain.getAffectedViews(affected_view) ";
    domain.getAffectedViews(affected_view);
    if (!affected_view.empty()) {
      error(method) << description << "returned a non-empty container." << std::endl;
    }

    // Create a test view object.
    TestView view(domain);

    // Creating this test view should have added the test view to domain's container of views.
    description = "After creating one view, domain.getAffectedViews(affected_view) ";
    domain.getAffectedViews(affected_view);
    if (affected_view.empty()) {
      error(method) << description << "returned an empty container." << std::endl;
    }

    description = "After TestView view(domain) ";
    std::string orig_domain = domain.getValue();

    if (view.getValue() != orig_domain) {
      error(method) << description << "view.getValue() returned \"" << view.getValue() << "\" not \"" <<
        orig_domain << "\", as expected." << std::endl;
    }

    value_string = "new value";
    description = "After domain.setValue(\"" + value_string + "\") ";
    domain.setValue(value_string);
    if (view.getValue() != orig_domain) {
      error(method) << description << "view.getValue() returned \"" << view.getValue() << "\" not \"" <<
        orig_domain << "\", as expected." << std::endl;
    }

    description = "After domain.updateAllViews() ";
    domain.updateAllViews();
    if (view.getValue() != domain.getValue()) {
      error(method) << description << "view.getValue() returned \"" << view.getValue() << "\" not \"" <<
        domain.getValue() << "\", as expected." << std::endl;
    }

    TestHandler handler(domain);

    value_string = "select";
    description = "After handler.handleMessage(\"" + value_string + "\", domain) ";
    handler.handleMessage(value_string, domain);
    if (domain.getValue() != value_string) {
      error(method) << description << "domain.getValue() returned \"" << domain.getValue() << "\" not \"" <<
        value_string << "\", as expected." << std::endl;
    }
    if (view.getValue() != value_string) {
      error(method) << description << "view.getValue() returned \"" << view.getValue() << "\" not \"" <<
        value_string << "\", as expected." << std::endl;
    }

    value_string = "not-a-command";
    description = "Calling handler.handleMessage(\"" + value_string + "\", domain) ";
    handler.handleMessage(value_string, domain);
    if (domain.getValue() == value_string) {
      error(method) << description << "domain.getValue() unexpectedly changed domain value to \"" << value_string <<
        "\"." << std::endl;
    }
    if (view.getValue() == value_string) {
      error(method) << description << "view.getValue() unexpectedly changed view value to \"" << value_string <<
        "\"." << std::endl;
    }

  }

}
