/** \file TestEnergy.cxx
    \brief Implementation of Energy class-specific parts of the unit test.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <cstdio>

#include "profit/Energy.h"
#include "profit/UnitTest.h"

namespace profit {

  void UnitTest::testEnergy() {
    std::string method("testEnergy()");

    double expected[] = {
      // Energy, wavelength computed independently from formula E = 12.3984 / wavelength.
      123.45678901234567890, .10042704090384336822,
      // Special case: energy = 0. -> wavelength = infinity -> 12.3984e+15 to avoid floating exceptions.
      0., 12.3984e+15,
      // Special case: wavelength = 0. -> energy = infinity -> 1.e+15 to avoid floating exceptions.
      1.e+15, 0.,
      // Large negative energy should be truncated to minimum.
      -123.45678901234567890, 12.3984e+15,
      // Large negative wavelength should be truncated to minimum.
      1.e+15, -123.45678901234567890,
      // Small negative energy should be truncated to minimum.
      -1.e-15 / (1. + s_double_epsilon), 12.3984e+15,
      // Small negative wavelength should be truncated to minimum.
      1.e+15, -12.3984e-15 / (1. + s_double_epsilon),
      // Smallest convertible energy == largest convertible wavelength.
      1.e-15 * (1. + s_double_epsilon), 12.3984e+15 / (1. + s_double_epsilon),
      // Smallest convertible wavelength == largest convertible energy.
      1.e+15 / (1. + s_double_epsilon), 12.3984e-15 * (1. + s_double_epsilon),
      // Energy too small.
      1.e-15 / (1. + s_double_epsilon), 12.3984e+15,
      // Wavelength too small.
      1.e+15, 12.3984e-15 / (1. + s_double_epsilon),
      // Energy too large.
      1.e+15 * (1. + s_double_epsilon), 12.3984e-15,
      // Wavelength too large.
      1.e-15, 12.3984e+15 * (1. + s_double_epsilon)
    };

    for (size_t ii = 0; ii != sizeof(expected)/sizeof(double) / 2; ++ii) {
      info(method) << "Calling testEnergy(" << expected[2 * ii] << ", " << expected[2 * ii + 1] << ")" << std::endl;
      testEnergy(expected[2 * ii], expected[2 * ii + 1]);
      info(method) << std::endl;
    }
  }

  void UnitTest::testEnergy(double correct_energy, double correct_wavelength) {
    std::string method("testEnergy(double, double)");

    // Create an Energy object.
    Energy energy1(correct_energy);

    // Confirm its energy and approximate wavelength.
    testEnergy("after Energy energy1(correct_energy)", energy1, correct_energy, correct_wavelength);

    // Create a Wavelength object.
    Wavelength wavelength1(correct_wavelength);

    // Confirm its wavelength and approximate energy.
    testEnergy("after Wavelength wavelength1(correct_wavelength)", wavelength1, correct_energy, correct_wavelength);

    // Test Energy copy constructor.
    Energy energy2(energy1);
    testEnergy("after Energy energy2(energy1)", energy2, energy1.energy(), energy1.wavelength());

    // Test Wavelength copy constructor.
    Wavelength wavelength2(wavelength1);
    testEnergy("after Energy wavelength2(wavelength1)", wavelength2, wavelength1.energy(), wavelength1.wavelength());

    // Test construction of Energy from Wavelength ("copy" construction from const IEnergy &).
    Energy energy3(wavelength1);
    testEnergy("after Energy energy3(wavelength1)", energy3, wavelength1.energy(), wavelength1.wavelength());

    // Test construction of Wavelength from Energy ("copy" construction from const IEnergy &).
    Wavelength wavelength3(energy1);
    testEnergy("after Wavelength wavelength3(energy1)", wavelength3, energy1.energy(), energy1.wavelength());

    // Test Energy assignment operator. First create Energy object which should have distinct values from the original
    // ones.
    Energy energy4(.9 * correct_energy);
    // Assign from that to another test object, which should change the destination object.
    energy2 = energy4;
    testEnergy("after energy2 = energy4", energy2, energy4.energy(), energy4.wavelength());

    // Test Wavelength assignment operator. First create Wavelength object which should have distinct values from the original
    // ones.
    Wavelength wavelength4(.9 * correct_wavelength);
    // Assign from that to another test object, which should change the destination object.
    wavelength2 = wavelength4;
    testEnergy("after wavelength2 = wavelength4", wavelength2, wavelength4.energy(), wavelength4.wavelength());

    // Test cross-class assignments, that is, operator =(const IEnergy &).
    energy3 = wavelength4;
    testEnergy("after energy3 = wavelength4", energy3, wavelength4.energy(), wavelength4.wavelength());
    wavelength3 = energy4;
    testEnergy("after wavelength3 = energy4", wavelength3, energy4.energy(), energy4.wavelength());
  }

  void UnitTest::testEnergy(const std::string & hint, const IEnergy & ienergy, double correct_energy, double correct_wavelength)
    const {
    std::string method("testEnergy(hint, ...)");

    // Create energy and wavelength comparison values which have a small positive limit near 0., to reflect
    // approximation used by IEnergy class.
    double compare_energy = correct_energy < 1.e-15 ? 1.e-15 : correct_energy;
    compare_energy = compare_energy > 1.e15 ? 1.e15 : compare_energy;
    double compare_wavelength = correct_wavelength < 12.3984e-15 ? 12.3984e-15 : correct_wavelength;
    compare_wavelength = compare_wavelength > 12.3984e15 ? 12.3984e15 : compare_wavelength;

    // Test for correctness of energy method.
    if (!isEqual(ienergy.energy(), compare_energy, 0., s_double_epsilon))
      error(method) << hint << ", called with correct_energy == " << correct_energy <<
        ", correct_wavelength == " << correct_wavelength <<
        ", " << ", energy() returned " << ienergy.energy() << ", not " << compare_energy <<
        ", as expected." << std::endl;

    // Test for correctness of wavelength method.
    if (!isEqual(ienergy.wavelength(), compare_wavelength, 0., s_double_epsilon))
      error(method) << hint << ", called with correct_energy == " << correct_energy <<
        ", correct_wavelength == " << correct_wavelength <<
        ", " << ", wavelength() returned " << ienergy.wavelength() << ", not " << compare_wavelength <<
        ", as expected." << std::endl;
  }

}
