/** \file TestOs.cxx
    \brief Implementation of Os class-specific parts of the unit test.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include "profit/Os.h"
#include "profit/Stream.h"
#include "profit/UnitTest.h"

#include <fstream>
#include <iostream>
#include <string>

namespace profit {

  void UnitTest::testOs() {
    std::string method("testOs");

    try {
      Os::getEnv("SPUD");
      error(method) << "Os::getEnv(\"SPUD\") did not throw exception before SPUD was set." << std::endl;
    } catch (const std::exception & x) {
      // Expected.
      info(method) << "Os::getEnv(\"SPUD\") threw exception: " << x.what() << ", as expected." << std::endl;
    }

    Os::setEnv("SPUD", "spud");

    try {
      std::string value = Os::getEnv("SPUD");
      if ("spud" != value) error(method) << "After SPUD was set, Os::getEnv(\"SPUD\") returned \"" << value <<
        "\", not \"spud\", as expected." << std::endl;
    } catch (const std::exception & x) {
      error(method) << "After SPUD was set, Os::getEnv(\"SPUD\") threw exception: " << x.what() << std::endl;
    }

#ifdef WIN32
    std::system("echo The word spud should be repeated after this colon: \%SPUD\%");
#else
    std::system("echo The word spud should be repeated after this colon: $SPUD");
#endif

    // Write a valid configuration file.
    {
      std::ofstream fs("valid_config");
      fs << "	 # comment" << std::endl;
      fs << " 	" << std::endl;
      fs << std::endl;
      fs << " SPUD		= bozo " << std::endl;
      fs << " SPUD2		= ${SPUD}/2/${SPUD}/${SPUD" << std::endl;
      fs << " BLANK		= blank" << std::endl;
    }

    // Confirm that valid configuration file may be read.
    try {
      bool failed = false;
      Os::readEnv("valid_config");

      // Check effects of setting this configuration.
      std::string value = Os::getEnv("SPUD");
      if ("bozo" != value) {
        error(method) << "After Os::readEnv(\"valid_config\"), Os::getEnv(\"SPUD\") returned \"" << value <<
          "\", not \"bozo\", as expected." << std::endl;
        failed = true;
      }

      value = Os::getEnv("BLANK");
      if ("blank" != value) {
        error(method) << "After Os::readEnv(\"valid_config\"), Os::getEnv(\"BLANK\") returned \"" << value <<
          "\", not \"blank\", as expected." << std::endl;
        failed = true;
      }

      value = Os::getEnv("SPUD2");
      if ("bozo/2/bozo/${SPUD" != value) {
        error(method) << "After Os::readEnv(\"valid_config\"), Os::getEnv(\"SPUD2\") returned \"" << value <<
          "\", not \"bozo/2/bozo/${SPUD\", as expected." << std::endl;
        failed = true;
      }

      if (!failed) info(method) << "Os::readEnv(\"valid_config\") set environment variables as expected." << std::endl;
    } catch (const std::exception & x) {
      error(method) << "Os::readEnv(\"valid_config\") threw exception: " << x.what() << std::endl;
    }

    // Write an invalid configuration file.
    {
      std::ofstream fs("invalid_config");
      fs << "	 # comment" << std::endl;
      fs << " 	no equals " << std::endl;
    }

    // Confirm that invalid configuration file throws.
    try {
      Os::readEnv("invalid_config");
      error(method) << "Os::readEnv(\"invalid_config\") with line \"no equals\" did not throw exception." << std::endl;
    } catch (const std::exception & x) {
      info(method) << "Os::readEnv(\"invalid_config\") threw exception: " << x.what() << ", as expected." << std::endl;
    }

    // Write another invalid configuration file.
    {
      std::ofstream fs("invalid_config");
      fs << "	 # comment" << std::endl;
      fs << " 	= value only" << std::endl;
    }

    // Confirm that invalid configuration file throws.
    try {
      Os::readEnv("invalid_config");
      error(method) << "Os::readEnv(\"invalid_config\") with line \"no equals\" did not throw exception." << std::endl;
    } catch (const std::exception & x) {
      info(method) << "Os::readEnv(\"invalid_config\") threw exception: " << x.what() << ", as expected." << std::endl;
    }

  }

}
