/** \file TestSpecFile.cxx
    \brief Implementation of SpecFile class-specific parts of the unit test.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include "profit/SpecFile.h"
#include "profit/Spectrum.h"
#include "profit/UnitTest.h"

#include <memory>
#include <stdexcept>

namespace profit {

  void UnitTest::testSpecFile() {
    std::string method("testSpecFile()");

    info(method) << "Creating SpecFile object for xrs.fak." << std::endl;
    SpecFile xrs_spec_file("xrs.fak");
    info(method) << std::endl;

    info(method) << "Calling SpecFile::getSpectrum(); this should work." << std::endl;
    try {
      xrs_spec_file.getSpectrum();
    } catch (const std::exception & x) {
      error(method) << x.what() << std::endl;
    }
    info(method) << std::endl;

    info(method) << "Calling SpecFile::getSpectrum(0); this should work." << std::endl;
    try {
      xrs_spec_file.getSpectrum(0);
    } catch (const std::exception & x) {
      error(method) << x.what() << std::endl;
    }
    info(method) << std::endl;

    info(method) << "Calling SpecFile::getSpectrum(-1); this should fail." << std::endl;
    try {
      xrs_spec_file.getSpectrum(-1);
      error(method) << "Did not fail!" << std::endl;
    } catch (const std::exception & x) {
      info(method) << "Expected: " << x.what() << std::endl;
    }
    info(method) << std::endl;

    info(method) << "Calling SpecFile::getSpectrum(1); this should fail." << std::endl;
    try {
      xrs_spec_file.getSpectrum(1);
      error(method) << "Did not fail!" << std::endl;
    } catch (const std::exception & x) {
      info(method) << "Expected: " << x.what() << std::endl;
    }
    info(method) << std::endl;

    info(method) << "Confirming that SpecFile::getNumSpectra() returns expected value." << std::endl;
    SpecFile::size_type num_spec = xrs_spec_file.getNumSpectra();
    if (1 != num_spec) {
      error(method) <<  "SpecFile::getNumSpectra() returned " << num_spec << ", not 1 as expected." << std::endl;
    }
    info(method) << std::endl;

    info(method) << "Creating SpecFile object for capella_pha2.fits." << std::endl;
    SpecFile chandra_spec_file("capella_pha2.fits");
    info(method) << std::endl;

    info(method) << "Calling SpecFile::getSpectrum(2); this should fail because ebounds are not found." << std::endl;
    try {
      chandra_spec_file.getSpectrum(2);
      error(method) << "Did not fail!" << std::endl;
    } catch (const std::exception & x) {
      info(method) << "Expected: " << x.what() << std::endl;
    }
    info(method) << std::endl;

    info(method) << "Calling SpecFile::getSpectrum(\"capella_row3.pha\", 2); this should work." << std::endl;
    try {
      chandra_spec_file.getSpectrum("capella_row3.rmf", 2);
    } catch (const std::exception & x) {
      error(method) << x.what() << std::endl;
    }
    info(method) << std::endl;

  }

}
