/** \file TestSpectrum.cxx
    \brief Implementation of Spectrum class-specific parts of the unit test.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <cstddef>
#include <cstdio>
#include <stdexcept>

#include "profit/SpecFile.h"
#include "profit/Spectrum.h"
#include "profit/Table.h"
#include "profit/UnitTest.h"

namespace profit {

  void UnitTest::testSpectrum() {
    std::string method("testSpectrum()");

    info(method) << "Opening table \"SPECTRUM\" in file \"xrs.fak\"." << std::endl;
    SpecFile spec_file("xrs.fak");
    info(method) << std::endl;

    info(method) << "Opening table \"EBOUNDS\" in file \"xrs.rmf\"." << std::endl;
    InputTable ebounds_table("xrs.rmf", "EBOUNDS");
    info(method) << std::endl;

    info(method) << "Creating Spectrum object from input data tables." << std::endl;
    Spectrum spectrum(spec_file, ebounds_table);
    info(method) << std::endl;

    info(method) << "Confirming data in spectrum after construction." << std::endl;
    testSpectrum("Just after constructing spectrum", spectrum, 16384);
    info(method) << std::endl;

    info(method) << "Performing energy cut on spectrum with range (6., 7.)." << std::endl;
    spectrum.select(6., 7.);
    info(method) << std::endl;

    info(method) << "Confirming data in spectrum after energy cut with range (6., 7.)." << std::endl;
    testSpectrum("After spectrum.select(6., 7.)", spectrum, 999);
    if (6. > spectrum.minEnergy().front())
      error(method) << "After spectrum.select(6., 7.), first bin in spectrum has minimum energy == " <<
        spectrum.minEnergy().front() << ", which is not >= 6. as expected." << std::endl;
    if (7. < spectrum.maxEnergy().back())
      error(method) << "After spectrum.select(6., 7.), last bin in spectrum has maximum energy == " <<
        spectrum.maxEnergy().back() << ", which is not <= 7. as expected." << std::endl;
    info(method) << std::endl;

    {
      info(method) << "Creating a copy of spectrum after energy cut with range (6., 7.)." << std::endl;
      Spectrum spectrum_copy = spectrum;
      info(method) << std::endl;

      info(method) << "Confirming data in copy of spectrum after energy cut with range (6., 7.)." << std::endl;
      testSpectrum("After Spectrum spectrum_copy = spectrum", spectrum_copy, 999);
      info(method) << std::endl;

      info(method) << "Performing energy cut on copy of spectrum with range (6.5, 6.5) which should exclude all bins." << std::endl;
      spectrum_copy.select(6.5, 6.5);
      info(method) << std::endl;

      info(method) << "Confirming data in copy of spectrum after energy cut with range (6.5, 6.5)." << std::endl;
      testSpectrum("After spectrum_copy.select(6.5, 6.5)", spectrum_copy, 0);
      info(method) << std::endl;
    }

    {
      Spectrum spectrum_copy = spectrum;

      info(method) << "Performing energy cut on copy of spectrum with range (5., 6.)." << std::endl;
      spectrum_copy.select(5., 6.);
      info(method) << std::endl;

      info(method) << "Confirming data in copy of spectrum after energy cut with range (5., 6.)." << std::endl;
      testSpectrum("After Spectrum spectrum_copy.select(5., 6.)", spectrum_copy, 0);
      info(method) << std::endl;
    }

    {
      Spectrum spectrum_copy = spectrum;

      info(method) << "Performing energy cut on copy of spectrum with range (7., 8.)." << std::endl;
      spectrum_copy.select(7., 8.);
      info(method) << std::endl;

      info(method) << "Confirming data in copy of spectrum after energy cut with range (7., 8.)." << std::endl;
      testSpectrum("After spectrum_copy.select(7., 8.)", spectrum_copy, 0);
      info(method) << std::endl;
    }

    info(method) << "Attempting to create a spectrum object from mismatched \"SPECTRUM\" and \"EBOUNDS\" tables." << std::endl;
    // Note: breaking ebounds table here.
    ebounds_table.filter("#row<16000");
    try {
      Spectrum bad_spectrum(spec_file, ebounds_table);
      error(method) << "No exception was thrown when creating a Spectrum from \"SPECTRUM\" and \"EBOUNDS\" tables " <<
        "with different sizes." << std::endl;
    } catch (const std::exception & x) {
      info(method) << x.what() << std::endl;
    }
    info(method) << std::endl;

    testTextFile(m_text_spectrum_file_name);
    testTextFile(m_unordered_text_file_name);

    info(method) << "Opening table \"SPECTRUM\" in file \"r1_o1_pha.fits\"." << std::endl;
    SpecFile xmm_spec_file("r1_o1_pha.fits");
    info(method) << std::endl;

    info(method) << "Opening table \"EBOUNDS\" in file \"r1_o1.rmf\"." << std::endl;
    InputTable xmm_ebounds_table("r1_o1.rmf", "EBOUNDS");
    info(method) << std::endl;

    info(method) << "Creating Spectrum object from input data tables." << std::endl;
    Spectrum xmm_spectrum(xmm_spec_file, xmm_ebounds_table);
    info(method) << std::endl;

    info(method) << "Confirming data in spectrum after construction." << std::endl;
    testSpectrum("Just after constructing spectrum", xmm_spectrum, 3400);
    info(method) << std::endl;

  }

  void UnitTest::testSpectrum(const std::string & hint, const Spectrum & spectrum, size_t num_elements) const {
    std::string method("testSpectrum(hint, ...)");

    if (num_elements != spectrum.minEnergy().size())
      error(method) << hint << ", minimum energy container in Spectrum object has " << spectrum.minEnergy().size() <<
        " elements, not " << num_elements << " as expected." << std::endl;
    if (num_elements != spectrum.maxEnergy().size())
      error(method) << hint << ", maximum energy container in Spectrum object has " << spectrum.minEnergy().size() <<
        " elements, not " << num_elements << " as expected." << std::endl;
    if (num_elements != spectrum.counts().size())
      error(method) << hint << ", counts container in Spectrum object has " << spectrum.counts().size() <<
        " elements, not " << num_elements << " as expected." << std::endl;
  }

  void UnitTest::testTextFile(const std::string & text_file_name) const {
    std::string method("testTextFile(text_file_name)");
    try {
      info(method) << "Opening text file containing spectrum with columns e_min and counts." << std::endl;
      SpecFile text_file(text_file_name);
      info(method) << std::endl;

      info(method) << "Creating spectrum from text file with columns e_min and counts." << std::endl;
      Spectrum text_spectrum(text_file);
      info(method) << std::endl;

      double expected_min_energy[] = { 0., 1., 2., 3., 4., 5., 6., 7., 8., 9., 10., 11. };
      double expected_max_energy[] = { 1., 2., 3., 4., 5., 6., 7., 8., 9., 10., 11., 12. };
      double expected_counts[] = { 0., 1., 3., 1., 6., 4., 2., 5., 8., 3.,  2.,  1. };
      size_t expected_size = sizeof(expected_counts) / sizeof(double);

      const std::vector<double> & min_energy(text_spectrum.minEnergy());
      const std::vector<double> & max_energy(text_spectrum.maxEnergy());
      const std::vector<double> & counts(text_spectrum.counts());

      info(method) << "Confirming contents of vectors in spectrum from text file with columns e_min and counts." << std::endl;
      if (min_energy.size() != expected_size) {
        error(method) << "In spectrum created from text file, size of min_energy vector is " << min_energy.size() <<
          ", not " << expected_size << ", as expected." << std::endl;
      } else {
        for (size_t index = 0; index != expected_size; ++index) {
          if (min_energy[index] != expected_min_energy[index])
            error(method) << "In spectrum created from text file, min_energy[" << index << "] was " << min_energy[index] <<
              ", not " << expected_min_energy[index] << ", as expected." << std::endl;
        }
      }
      if (max_energy.size() != expected_size) {
        error(method) << "In spectrum created from text file, size of max_energy vector is " << max_energy.size() <<
          ", not " << expected_size << ", as expected." << std::endl;
      } else {
        for (size_t index = 0; index != expected_size; ++index) {
          if (max_energy[index] != expected_max_energy[index])
            error(method) << "In spectrum created from text file, max_energy[" << index << "] was " << max_energy[index] <<
              ", not " << expected_max_energy[index] << ", as expected." << std::endl;
        }
      }
      if (counts.size() != expected_size) {
        error(method) << "In spectrum created from text file, size of counts vector is " << counts.size() <<
          ", not " << expected_size << ", as expected." << std::endl;
      } else {
        for (size_t index = 0; index != expected_size; ++index) {
          if (counts[index] != expected_counts[index])
            error(method) << "In spectrum created from text file, counts[" << index << "] was " << counts[index] <<
              ", not " << expected_counts[index] << ", as expected." << std::endl;
        }
      }
      info(method) << std::endl;

      testSpectrum("After creating spectrum from text file with columns e_min and counts", text_spectrum, expected_size);
      
    } catch (const std::exception & x) {
      error(method) << "While creating a spectrum from a text file which includes e_min and counts, caught exception: " <<
        x.what() << std::endl;
    }
  }
}
