/** \file TestStream.cxx
    \brief Implementation of OStream class-specific parts of the unit test.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <fstream>
#include <iostream>
#include <limits>
#include <string>

#include "profit/Gui.h"
#include "profit/Stream.h"
#include "profit/GuiOStream.h"
#include "profit/UnitTestGui.h"

namespace profit {

  void UnitTest::testStream() {
    std::string method("UnitTest::testStream()");

    info(method) << "Testing non-GUI aspects of streams only." << std::endl;
    info(method) << std::endl;

    // Before initializing standard streams, write to pout. This should have no effect.
    pout << "This was written before initializing standard streams, so THIS SHOULD NOT APPEAR on any stream!" << std::endl;

    // Set max_chat used for tests.
    unsigned int max_chat = 3;

    // Initialize standard streams with maximum chatter of max_chat.
    InitStdStreams("test_profit", max_chat, true);

    // Initialize standard streams again with max_chat == 0 -- should have no effect.
    InitStdStreams("wrong_executable", 0, false);

    std::cout << "After this, there should be exactly one line on each of perr, plog and pout." << std::endl;

    // Test direct writes to standard streams.
    perr << "This should appear on std::cerr." << std::endl;
    plog << "This should appear on std::clog." << std::endl;
    pout << "This should appear on std::cout." << std::endl;

    // Create a file-based STL stream.
    std::string out_file = "test_profit-out";
    std::ofstream std_os(out_file.c_str());

    // Verify that the file was opened correctly.
    if (std_os.bad()) {
      error(method) << "ERROR: could not open output file " << out_file << std::endl;
      return;
    }

    // Connect all standard streams to the output file.
    perr.connect(std_os);
    plog.connect(std_os);
    pout.connect(std_os);

    // Also disconnect them from standard streams to prevent any further notices.
    perr.disconnect(std::cerr);
    plog.disconnect(std::clog);
    pout.disconnect(std::cout);

    // Write something directly to each stream.
    std_os << "Three lines should now be written to this file with no prefix, one via each stream:" << std::endl;
    perr << Chat(max_chat + 1) << "Despite Chat(" << max_chat + 1 << ") this was written to perr." << std::endl;
    plog << Chat(max_chat + 1) << "Despite Chat(" << max_chat + 1 << ") this was written to plog." << std::endl;
    pout << Chat(max_chat + 1) << "Despite Chat(" << max_chat + 1 << ") this was written to pout." << std::endl;

    // Create a stream which forwards to pout, but uses chatter, which can be used to suppress some output.
    OStream my_out(true);
    my_out.connect(pout);

    // Test that this works correctly with different chat levels.
    std_os << "Three lines should now be written to this file via my_out:" << std::endl;
    my_out << "This was written to my_out with default chat level." << std::endl;
    my_out << Chat(max_chat - 1) << "This was written to my_out with maximum chatter set to " << max_chat - 1 << std::endl;
    my_out << Chat(max_chat) << "This was written to my_out with maximum chatter set to " << max_chat << std::endl;
    my_out << Chat(max_chat + 1) <<
      "THIS SHOULD NOT APPEAR! This was written to my_out with maximum chatter set to " << max_chat + 1 << std::endl;
    my_out << "THIS SHOULD NOT APPEAR! This was written to my_out after maximum chatter was set to " << max_chat + 1 << std::endl;

    // Add the method name to the output stream's prefix.
    my_out.setPrefix("exec_name: ");

    // Make certain the prefix appears.
    std_os << "The next line should begin with a prefix of \"exec_name: \"" << std::endl;

    // Write line with prefix.
    my_out << Chat(max_chat) <<
      "This was written to my_out with explicit Chat(" << max_chat << ") after prefix was set." << std::endl;

    // Test setting stream precision.
    double a_double = 1.23456789012;

    pout << "With default precision, 1.23456789012 is displayed as " << a_double << std::endl;
    pout.precision(12);
    pout << "After precision was set to 12, 1.23456789012 is displayed as " << a_double << std::endl;

    // Test setting format flags.
    pout << "With default format flags, 16 is displayed as " << 16 << std::endl;
    pout.flags(std::ios_base::hex);
    pout << "With format flags hex, 16 is displayed as " << 16 << std::endl;

    // Test field width and fill characters.
    pout << "With default width and fill character, 1.234 shifted twice gives " << 1.234 << 1.234 << std::endl;
    pout.setf(std::ios_base::left, std::ios_base::adjustfield);
    pout.fill('#');
    pout.width(16);
    pout << "With width == 16 and fill character #, 1.234 shifted twice gives ";
    pout.width(16);
    pout << 1.234;
    pout.width(16);
    pout << 1.234;
    pout << std::endl;
  }

  void UnitTestGui::testStream() {
    std::string method("UnitTestGui::testStream()");
    UnitTest::testStream();

    // Create Gui object to use for tests.
    info(method) << "Getting Gui singleton." << std::endl;
    Gui * gui(Gui::instance());
    if (0 == gui) {
      error(method) << "Failed to allocate or retrieve Gui." << std::endl;
      info(method) << std::endl;
      return;
    }
    info(method) << std::endl;

    info(method) << "Testing GUI aspects of streams only." << std::endl;
    info(method) << "Getting GUI output stream." << std::endl;
    GuiOStream & os(gui->getMainOutput());
    info(method) << std::endl;

    info(method) << "Sending output to Gui's main message window." << std::endl;
    os << "This is the " << 1 << "st of \n" << 3 << " lines to be displayed" << std::endl;
    os << "in the Gui's main message window." << std::endl;
    info(method) << std::endl;
  }
}
