/** \file TextTableFilePtr.cxx
    \brief Implementation of TextTableFilePtr class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <cctype>
#include <cerrno>
#include <cstddef>
#include <cstdlib>
#include <limits>
#include <list>
#include <stdexcept>

#include "TextTableFilePtr.h"
#include "profit/Table.h"

namespace {

  class Buffer {
    public:
      Buffer(size_t size): m_size(size), m_buf(new char[size]) {
        if (0 == m_buf) throw std::runtime_error("Cannot allocate buffer");
      }
      ~Buffer() { delete [] m_buf; }
      size_t size() const { return m_size; }
      char * get() { return m_buf; }

    private:
      size_t m_size;
      char * m_buf;
  };

}

namespace profit {

  TextTableFilePtr::TextTableFilePtr(const std::string & file_name):
    TableFilePtr(file_name, ""), m_file_stream(file_name.c_str()), m_keyword(), m_column(), m_field_id() {
    if (!m_file_stream.good()) throw std::runtime_error(("TextTableFilePtr could not open file " + file_name).c_str());
    
    Buffer buf(1024);

    // Flag indicating file is in profit native format.
    bool profit_format = false;

    while (m_file_stream.good()) {
      // Read next line from file.
      m_file_stream.getline(buf.get(), buf.size());

      // Get pointer to the buffer's underlying C buffer, and remove comments.
      const char * cp = buf.get();

      // Make a container to hold a row's worth of information.
      std::list<std::string> row;

      // Skip leading space.
      while (0 != isspace(*cp)) ++cp;

      // Skip this line if first significant character is a bang.
      if ('!' == *cp) continue;

      // Check for keywords.
      if (parseKeyword(cp)) {
        if (!profit_format) {
          keyword_cont_type::iterator fmt = m_keyword.find("FORMAT");
          if (m_keyword.end() != fmt && "profit" == fmt->second) profit_format = true;
        }
        continue;
      }

      // If first line was not FORMAT = profit, read no further.
      if (!profit_format) {
        m_file_stream.close();
        throw std::runtime_error("TextTableFilePtr::TextTableFilePtr(file_name): File " + file_name +
          " is not in PROFIT text file format.");
      }

      // Parse buffer until its end is reached.
      while ('\0' != *cp) {
        // Look for non-space, which signifies a significant value.
        const char * end = cp;
        while (0 == isspace(*end) && '!' != *end && '\0' != *end) ++end;

        // Put this value into the row if it has any significant characters.
        if (cp != end) row.push_back(std::string(cp, end));

        // If a comment was detected, break and do nothing further with this row.
        if ('!' == *end) break;

        // Start after the current pointer.
        cp = end;
  
        // Skip space between values.
        while (0 != isspace(*cp)) ++cp;
      }

      // Skip row if completely blank.
      if (row.empty()) continue;

      // If m_column is completely empty, it means this is the first row which contained a number, so add space
      // for at least one column.
      if (m_column.empty()) m_column.push_back(column_type());

      // If there are too few columns for the size of row, increase the size of m_column with empty vectors as needed.
      while (m_column.size() < row.size()) {
        m_column.push_back(column_type(m_column.front().size()));
      }

      // If there are too few elements in the row for the number of columns, pad out the row with 0.s.
      while (row.size() < m_column.size()) {
        row.push_back("");
      }

      // Now the number of fields in row is the same as the number of columns in m_column.
      // It is therefore safe to iterate over these containers and push each value from row onto the appropriate column.
      column_cont_type::iterator out_itor = m_column.begin();
      for (std::list<std::string>::iterator in_itor = row.begin(); in_itor != row.end(); ++in_itor, ++out_itor) {
        out_itor->push_back(*in_itor);
      }

      // Increment the number of rows which were successfully read.
      ++m_num_rows;
    }

    keyword_cont_type::iterator itor = m_keyword.find("TTYPE");
    if (m_keyword.end() != itor) parseFields(itor->second);

    // Check for errors.
    if (m_file_stream.bad()) throw std::runtime_error("Error while reading file " + file_name);
  }

  TextTableFilePtr::~TextTableFilePtr() {}

  void TextTableFilePtr::getColumn(const std::string & field_name, std::vector<double> & column) const {
    // Remove current content of the column.
    column.clear();
    column.resize(m_column[getColumnId(field_name)].size());
    for (column_type::size_type ii = 0; ii != column.size(); ++ii) {
      column[ii] = getCell(field_name, ii);
    }
  }

  void TextTableFilePtr::getColumn(const std::string & field_name, std::vector<std::string> & column) const {
    column = m_column[getColumnId(field_name)];
  }

  void TextTableFilePtr::getColumnNames(std::vector<std::string> & name) const { name = m_field_name; }

  TableFilePtr::size_type TextTableFilePtr::getColumnId(const std::string & field_name) const {
    // Start off assuming the field was not found.
    size_type field_id = std::numeric_limits<size_type>::max();

    std::string field_name_uc(field_name);
    for (std::string::iterator itor = field_name_uc.begin(); itor != field_name_uc.end(); ++itor) *itor = toupper(*itor);

    // Look for field in associative container.
    field_id_type::const_iterator itor = m_field_id.find(field_name_uc);

    // If field was found, return the lookup of the column identifier for that field.
    if (m_field_id.end() != itor) field_id = itor->second;
    else throw std::runtime_error("TextTableFilePtr::getColumnId(...) could not find field named \"" + field_name +
      "\", in file \"" + m_file_name + "\"");

    return field_id;
  }

  double TextTableFilePtr::getCell(const std::string & field_name, TableFilePtr::size_type record_index) const {
    return strToDouble(m_column[getColumnId(field_name)][record_index]);
  }

  void TextTableFilePtr::getCell(const std::string & field_name, TableFilePtr::size_type record_index, std::string & field_value)
    const {
    field_value = m_column[getColumnId(field_name)][record_index];
  }

  void TextTableFilePtr::getCell(const std::string &, TableFilePtr::size_type, std::vector<double> &) const {
    throw std::logic_error("TextTableFilePtr::getCell(string, size_type, vector<double> &) not implemented");
  }

  void TextTableFilePtr::filter(const std::string & expression) {
    throw std::logic_error("TextTableFilePtr::filter not implemented");
  }

  void TextTableFilePtr::filter(const std::string & field_name, double min_value, double max_value) {
    // Get the column used for filtering.
    column_type & column(m_column[getColumnId(field_name)]);

    // Container of indices which pass the filter.
    std::list<column_type::size_type> save_index;

    // Iterate over the column, and add to the save_index all those indices which pass the filter.
    for (column_type::size_type index = 0; index != column.size(); ++index) {
      double column_value = strToDouble(column[index]);
      if (min_value <= column_value && column_value <= max_value) save_index.push_back(index);
    }

    // Now go through all the columns and repopulate them.
    for (column_cont_type::iterator col_itor = m_column.begin(); col_itor != m_column.end(); ++col_itor) {
      column_type & orig_col(*col_itor);
      column_type new_col(save_index.size());

      // Iterate over save_index, and put each saved original column value into the new column.
      column_type::size_type index = 0;
      for (std::list<column_type::size_type>::iterator itor = save_index.begin(); itor != save_index.end(); ++itor) {
        new_col[index++] = orig_col[*itor];
      }

      // Replace original column.
      orig_col = new_col;
    }

    // Reset the number of rows based on the filtering performed.
    m_num_rows = save_index.size();
  }

  InputTable::size_type TextTableFilePtr::getFieldDim(const std::string & field_name) const { return 1; }

  std::string TextTableFilePtr::getKeyword(const std::string & key_name) const {
    // Convert key name to uppercase.
    std::string key_name_uc(key_name);
    for (std::string::iterator itor = key_name_uc.begin(); itor != key_name_uc.end(); ++itor) *itor = toupper(*itor);

    // Define string to hold key value.
    std::string keyword_value;

    // Look for keyword in container.
    keyword_cont_type::const_iterator itor = m_keyword.find(key_name_uc);

    // Return keyword value if it was found. Otherwise throw an exception.
    if (m_keyword.end() != itor) keyword_value = itor->second;
    else throw std::runtime_error("Cannot read keyword \"" + key_name + "\" in text file " + m_file_name);

    return keyword_value;
  }

  bool TextTableFilePtr::parseKeyword(const char * cp) {
    bool keyword_found = false;
    const char * begin = cp;
    
    // First look for white space or = to terminate the name of the keyword.
    while ('\0' != *cp && 0 == isspace(*cp) && '=' != *cp) ++cp;

    const char * end = cp;

    // Next look for trailing white space.
    while (0 != isspace(*cp)) ++cp;

    // If an = was found, continue with process. Otherwise, this is not a keyword.
    if ('=' == *cp) {
      // Begin/end delimit the keyword name.
      std::string key_name(begin, end);

      // First, convert name of keyword to all upper case.
      for (std::string::iterator itor = key_name.begin(); itor != key_name.end(); ++itor) *itor = toupper(*itor);

      // Skip past the equals sign and any space which follows it.
      while (0 != isspace(*++cp)) {}

      // cp now points to the beginning of the value.
      begin = cp;

      // Find end of value.
      while ('\0' != *cp) ++cp;

      // Back up to exclude any trailing white space.
      while (cp > begin && 0 != isspace(*(cp - 1))) --cp;

      // cp now points to the last space or terminating 0.
      end = cp;

      // Put the key-value pair in the associative container.
      m_keyword[key_name].assign(begin, end);

      // Success, so flag keyword as found.
      keyword_found = true;
    }

    return keyword_found;
  }

  void TextTableFilePtr::parseFields(const std::string & field_name) {
    // Convert field name to uppercase.
    std::string field_name_uc(field_name);
    for (std::string::iterator itor = field_name_uc.begin(); itor != field_name_uc.end(); ++itor) *itor = toupper(*itor);

    // Split up string into words which give names of fields.
    size_type index = 0;
    std::string::const_iterator begin = field_name_uc.begin();
    while (begin != field_name_uc.end()) {
      // Find end of this token (whitespace or end of string).
      std::string::const_iterator end = begin;
      while (end != field_name_uc.end() && 0 == isspace(*end)) ++end;

      // Add token to container of fields.
      m_field_id[std::string(begin, end)] = index++;

      // Add token to container of field names.
      m_field_name.push_back(std::string(begin, end));

      // Start looking for the next token after the current position.
      begin = end;
      while (begin != field_name_uc.end() && 0 != isspace(*begin)) ++begin;
    }
  }

  double TextTableFilePtr::strToDouble(const std::string & double_string) const {
    char * suffix = 0;
    errno = 0;
    double double_value = 0.;
    if (!double_string.empty()) {
      double_value = strtod(double_string.c_str(), &suffix);
      if (0 != errno || (0 != suffix && '\0' != *suffix))
        throw std::runtime_error("TextTableFilePtr::strToDouble could not convert value \"" + double_string + "\"");
    }
    return double_value;
  }
}
