/** \file TextTableFilePtr.h
    \brief Declaration of TextTableFilePtr class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_TextTableFilePtr_h
#define profit_TextTableFilePtr_h

#include <fstream>
#include <list>
#include <map>
#include <string>

#include "TableFilePtr.h"

namespace profit {

  /** \class TextTableFilePtr
      \brief Internal base class for concrete implementations of Table interface for different file formats.
  */
  class TextTableFilePtr : public TableFilePtr {
    public:
      /** \brief Create an file accessor object.
          \param file_name The name of the file.
      */
      TextTableFilePtr(const std::string & file_name);

      /// \brief Close the file being accessed, and free all associated resources.
      virtual ~TextTableFilePtr();

      /** \brief Get the contents of a table column. Column values must all be convertible to double.
          \param field_name The name of the field/column.
          \param column The destination array to hold the column contents.
      */
      virtual void getColumn(const std::string & field_name, std::vector<double> & column) const;

      /** \brief Get the contents of a table column as strings.
          \param field_name The name of the field/column.
          \param column The destination array to hold the column contents.
      */
      virtual void getColumn(const std::string & field_name, std::vector<std::string> & column) const;

      /** \brief Get the names of table columns, in order as a vector of strings.
          \param name The destination array to hold the column identifiers.
      */
      virtual void getColumnNames(std::vector<std::string> & name) const;

      /** \brief Get the identifier for the given column.
          \param field_name The name of the field/column.
      */
      virtual size_type getColumnId(const std::string & field_name) const;

      /** \brief Get the cell identified by the field/column and record/row number.
          \param field_name The name of the field/column.
          \param record_index The index (0-based row number) in the file.
      */
      virtual double getCell(const std::string & field_name, size_type record_index) const;

      /** \brief Get the cell identified by the field/column and record/row number.
          \param field_name The name of the field/column.
          \param record_index The index (0-based row number) in the file.
          \param field_value The value of the field/column.
      */
      virtual void getCell(const std::string & field_name, size_type record_index, std::string & field_value) const;

      /** \brief Get the cell identified by the field/column and record/row number.
          \param field_name The name of the field/column.
          \param record_index The index (0-based row number) in the file.
          \param field_value The value of the field/column.
      */
      virtual void getCell(const std::string & field_name, size_type record_index, std::vector<double> & field_value) const;

      /** \brief Filter the current table contents using the given expression.
          \param expression The expression (Cfitsio row filtering syntax.)
      */
      virtual void filter(const std::string & expression);

      /** \brief Filter the current table contents using the given expression.
          \param field_name The name of the field on which to filter.
          \param min_value The minimum value of the filter range.
          \param max_value The maximum value of the filter range.
      */
      virtual void filter(const std::string & field_name, double min_value, double max_value);

      /** \brief Get the dimensionality of the given field.
          \param field_name The name of the field/column.
      */
      virtual size_type getFieldDim(const std::string & field_name) const;

      /** \brief Return the content of a keyword, as a string.
          \param key_name The name of the keyword to get.
      */
      virtual std::string getKeyword(const std::string & key_name) const;

    protected:
      typedef std::vector<std::string> column_type;
      typedef std::vector<column_type> column_cont_type;
      typedef std::map<std::string, column_cont_type::size_type> field_id_type;
      typedef std::map<std::string, std::string> keyword_cont_type;
      typedef std::vector<std::string> column_name_cont_type;

      /** \brief Interpret and store a keyword, assuming the input string has the form keyword = value.
          Return true if the string was successfully parsed.
          \param cp The input string to parse.
      */
      virtual bool parseKeyword(const char * cp);

      /** \brief Interpret the given string as white space delimited list of fields in this file.
          \param field_name The string containing the names.
      */
      virtual void parseFields(const std::string & field_name);

      /** \brief Convert a string which contains a number into a double value.
          \param double_value The string to convert.
      */
      double strToDouble(const std::string & double_value) const;

      std::ifstream m_file_stream;
      keyword_cont_type m_keyword;
      column_cont_type m_column;
      field_id_type m_field_id;
  };

}

#endif
