/** \file profit_main.cxx
    \brief Profit top-level application, which sets up the environment and then
    runs the *real* profit application.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include "profit/Os.h"

#include <iostream>
#include <stdexcept>
#include <string>

#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

using namespace profit;

namespace {

#ifdef WIN32
  const char * s_delim = "/\\";
  const char * s_dir_join = "\\";
  const char * s_path_delim = ";";
  std::string s_pfiles_delim = "|";
#else
  const char * s_delim = "/";
  const char * s_dir_join = "/";
  const char * s_path_delim = ":";
  std::string s_pfiles_delim = ";";
#endif

  void ProcessCmdLine(int & argc, char ** & argv) {
#ifdef WIN32
    static char debug_cmd[] = "devenv";
    // TODO: Pick up the version number automatically instead of hard-wired 2.0.
    static char debug_arg[] = "\"%PROFIT_TOP%\\profit\\2.0\\vs7\\profit\\profit.sln\"";
    static char profit_binary[] = "profit-binary.exe";
#else
    static char debug_cmd[] = "gdb";
    static char debug_arg[] = "profit-binary";
    static char profit_binary[] = "profit-binary";
#endif
    bool debug = false;

    // Detect and enable debug mode, if possible.
    if (1 < argc) {
      if ("-d" == std::string(argv[1]) && '\0' != *debug_cmd) {
        debug = true;
      }
    }

    if (debug) {
      // Reset argv to run gdb on the binary with no arguments.
      argv[0] = debug_cmd;
      argv[1] = debug_arg;
      argc = 2;
    } else {
      // Reset argv[0] to be the name of the new binary.
      argv[0] = profit_binary;
    }
  }

  std::string CatCmdLine(int argc, char ** argv) {
    std::string cmd_line;
    int ii = 0;
    for (ii = 0; ii < argc; ++ii) {
      cmd_line += argv[ii];
      cmd_line += " ";
    }

    // Chop off trailing space from command line.
    if (!cmd_line.empty()) cmd_line = cmd_line.substr(0, cmd_line.size() - 1);

    return cmd_line;
  }

  std::string TruncateLastSegment(const std::string & path) {
    std::string truncated_path;

    // Discard all dots at the end of the path before truncating it.
    std::string clean_path = path;
    std::string::size_type last_slash_pos = clean_path.find_last_of(s_delim);
    while (std::string::npos != last_slash_pos) {
      if ("." == clean_path.substr(last_slash_pos + 1)) {
        clean_path = clean_path.substr(0, last_slash_pos);
        last_slash_pos = clean_path.find_last_of(s_delim);
      } else {
        break;
      }
    }

    // Find the last slash or backslash in the file name; if present, chop off everything after it.
    last_slash_pos = clean_path.find_last_of(s_delim);
    if (std::string::npos != last_slash_pos)
      truncated_path = clean_path.substr(0, last_slash_pos);

    return truncated_path;
  }

#ifndef WIN32
  std::string SearchPath(const std::string & binary_name) {
    std::string dir;

    // Read the path from the OS.
    const char * c_path = ::getenv("PATH");
    if (0 == c_path) throw std::runtime_error("PATH variable is not set");
    std::string path = c_path;

    // Scan the path, check for existence of binary in each part.
    std::string::size_type prev_delim_pos = 0;
    std::string::size_type next_delim_pos = std::string::npos;

    do {
      next_delim_pos = path.find_first_of(s_path_delim, prev_delim_pos);

      // Path segment is everything from the previous path delimiter to the next path delimiter.
      std::string path_seg = path.substr(prev_delim_pos, next_delim_pos - prev_delim_pos);

      // Check for the existence of an executable file within this path segment.
      std::string file_name = path_seg + s_dir_join + binary_name;

      if (0 == access(file_name.c_str(), F_OK | X_OK)) {
        // Unfortunately, access returns 0 if the file is a directory as well.
        // Do not consider this a match if this is the case. The easy way to check
        // is to call access again for "/path/to/file/.". If that succeeds, it is
        // a directory.
        file_name += s_dir_join;
        file_name += ".";
        if (0 != access(file_name.c_str(), F_OK | X_OK)) {
          dir = path_seg;
          break;
        }
      }

      // Look for next segment after the current segment.
      prev_delim_pos = next_delim_pos + 1;
    } while (std::string::npos != next_delim_pos);

    return dir;
  }
#endif

  std::string FindBinaryDirectory(const std::string & binary_name) {
    std::string dir;
#ifdef WIN32
    LPCTSTR pFileName = binary_name.c_str();
    char lpBuffer[FILENAME_MAX] = "";
    DWORD nBufferLength = sizeof(lpBuffer) / sizeof(lpBuffer[0]);
    LPTSTR lpFilePart;
    LPCTSTR lpPath = 0;
    LPCTSTR lpFileName = binary_name.c_str();
    LPCTSTR lpExtension = 0;

    // Search the path to find the given binary.
    DWORD num_char = SearchPath(lpPath, lpFileName, lpExtension, nBufferLength, lpBuffer, &lpFilePart);

    // Find the last slash or backslash in the file name.
    dir = TruncateLastSegment(lpBuffer);
#else
    // Find the last slash or backslash in the file name.
    dir = TruncateLastSegment(binary_name);

    if (dir.empty()) {
      // The binary name included no path; search the path.
      dir = SearchPath(binary_name);
    } else if ('/' != dir[0]) {
      // Binary location is relative to the current directory, so prepend the current directory to the binary's directory.
      char cwd[FILENAME_MAX + 1] = "";
      if (0 == ::getcwd(cwd, FILENAME_MAX))
        throw std::runtime_error("Unable to determine the directory containing the profit binary");
      if (dir.empty()) dir = cwd;
      else dir = Os::joinPathSegment(cwd, dir);
    }
    // TODO Collapse ../ and ./ to make a clean path.
#endif
    return dir;
  }

  void SetUpEnvironment(const std::string & binary_name) {
    // Get the name of the directory in which this binary resides.
    std::string bin_dir = FindBinaryDirectory(binary_name);

    // Top of the profit tree is one directory level "up" from the binary directory.
    std::string top_dir = bin_dir + s_dir_join + "..";

    // Set the PROFIT variable.
    Os::setEnv("PROFIT", top_dir);

    // Set the PKG_TOP variable (redundant with PROFIT).
    Os::setEnv("PKG_TOP", top_dir);

    // Set environment variables derived from the package top directory.
    Os::readEnv(Os::joinPathSegment(bin_dir, "profit-config"));

    // Get variables which affect paths.
    std::string apec = Os::getEnv("APEC");
    std::string rootsys = Os::getEnv("ROOTSYS");

    // Prepend entries to paths.
    std::string path;
    try {
      path = Os::getEnv("DYLD_LIBRARY_PATH");
      path = Os::appendPath(Os::joinPathSegment(apec, "lib"), path);
    } catch (const std::exception &) {
      path = Os::joinPathSegment(apec, "lib");
    }
    path = Os::appendPath(Os::joinPathSegment(rootsys, "lib"), path);
    path = Os::appendPath(Os::joinPathSegment(top_dir, "lib"), path);
    Os::setEnv("DYLD_LIBRARY_PATH", path);

    try {
      path = Os::getEnv("LD_LIBRARY_PATH");
      path = Os::appendPath(Os::joinPathSegment(apec, "lib"), path);
    } catch (const std::exception &) {
      path = Os::joinPathSegment(apec, "lib");
    }
    path = Os::appendPath(Os::joinPathSegment(rootsys, "lib"), path);
    path = Os::appendPath(Os::joinPathSegment(top_dir, "lib"), path);
    Os::setEnv("LD_LIBRARY_PATH", path);

    try {
      path = Os::getEnv("PATH");
      path = Os::appendPath(Os::joinPathSegment(apec, "bin"), path);
    } catch (const std::exception &) {
      path = Os::joinPathSegment(apec, "bin");
    }
    path = Os::appendPath(Os::joinPathSegment(rootsys, "bin"), path);
    path = Os::appendPath(Os::joinPathSegment(top_dir, "bin"), path);
    if (bin_dir != Os::joinPathSegment(top_dir, "bin")) path = Os::appendPath(bin_dir, path);
    Os::setEnv("PATH", path);

    // Set up PFILES.
    Os::setEnv("PFILES", Os::expandEnv(Os::joinPathSegment(s_pfiles_delim + "${APEC}", "pfiles")));
  }

}

int main(int argc, char ** argv) {
  int status = 1;

  try {
    // Set up environment to have all necessary variables to run the real profit.
    SetUpEnvironment(argv[0]);

    // Handle command line switches that are relevant to this program.
    ProcessCmdLine(argc, argv);

    // Concatenate arguments to form new command line.
    std::string cmd_line = CatCmdLine(argc, argv);

    // Run the command and return its status.
    status = std::system(cmd_line.c_str());
  } catch (const std::exception & x) {
    std::cerr << "Profit failed to start: " << x.what() << std::endl;
  }

  // Report final status.
  return status != 0 ? 1 : 0;
}
